(function () {
    'use strict';

    /*! *****************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */

    function __awaiter(thisArg, _arguments, P, generator) {
        function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
        return new (P || (P = Promise))(function (resolve, reject) {
            function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
            function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
            function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
            step((generator = generator.apply(thisArg, _arguments || [])).next());
        });
    }

    /*
     This file is part of GNU Taler
     (C) 2019 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    /**
     * Number of fractional units that one value unit represents.
     */
    const fractionalBase = 1e8;
    /**
     * How many digits behind the comma are required to represent the
     * fractional value in human readable decimal format?  Must match
     * lg(fractionalBase)
     */
    const fractionalLength = 8;
    /**
     * Maximum allowed value field of an amount.
     */
    const maxAmountValue = Math.pow(2, 52);
    /**
     * Get an amount that represents zero units of a currency.
     */
    function getZero(currency) {
        return {
            currency,
            fraction: 0,
            value: 0,
        };
    }
    function jsonifyAmount(amt) {
        if (typeof amt === "string") {
            return parseOrThrow(amt);
        }
        return amt;
    }
    function sum(amounts) {
        if (amounts.length <= 0) {
            throw Error("can't sum zero amounts");
        }
        const jsonAmounts = amounts.map((x) => jsonifyAmount(x));
        return add(jsonAmounts[0], ...jsonAmounts.slice(1));
    }
    /**
     * Add two amounts.  Return the result and whether
     * the addition overflowed.  The overflow is always handled
     * by saturating and never by wrapping.
     *
     * Throws when currencies don't match.
     */
    function add(first, ...rest) {
        const currency = first.currency;
        let value = first.value + Math.floor(first.fraction / fractionalBase);
        if (value > maxAmountValue) {
            return {
                amount: { currency, value: maxAmountValue, fraction: fractionalBase - 1 },
                saturated: true,
            };
        }
        let fraction = first.fraction % fractionalBase;
        for (const x of rest) {
            if (x.currency !== currency) {
                throw Error(`Mismatched currency: ${x.currency} and ${currency}`);
            }
            value =
                value + x.value + Math.floor((fraction + x.fraction) / fractionalBase);
            fraction = Math.floor((fraction + x.fraction) % fractionalBase);
            if (value > maxAmountValue) {
                return {
                    amount: {
                        currency,
                        value: maxAmountValue,
                        fraction: fractionalBase - 1,
                    },
                    saturated: true,
                };
            }
        }
        return { amount: { currency, value, fraction }, saturated: false };
    }
    /**
     * Subtract two amounts.  Return the result and whether
     * the subtraction overflowed.  The overflow is always handled
     * by saturating and never by wrapping.
     *
     * Throws when currencies don't match.
     */
    function sub(a, ...rest) {
        const currency = a.currency;
        let value = a.value;
        let fraction = a.fraction;
        for (const b of rest) {
            if (b.currency !== currency) {
                throw Error(`Mismatched currency: ${b.currency} and ${currency}`);
            }
            if (fraction < b.fraction) {
                if (value < 1) {
                    return { amount: { currency, value: 0, fraction: 0 }, saturated: true };
                }
                value--;
                fraction += fractionalBase;
            }
            console.assert(fraction >= b.fraction);
            fraction -= b.fraction;
            if (value < b.value) {
                return { amount: { currency, value: 0, fraction: 0 }, saturated: true };
            }
            value -= b.value;
        }
        return { amount: { currency, value, fraction }, saturated: false };
    }
    /**
     * Compare two amounts.  Returns 0 when equal, -1 when a < b
     * and +1 when a > b.  Throws when currencies don't match.
     */
    function cmp(a, b) {
        a = jsonifyAmount(a);
        b = jsonifyAmount(b);
        if (a.currency !== b.currency) {
            throw Error(`Mismatched currency: ${a.currency} and ${b.currency}`);
        }
        const av = a.value + Math.floor(a.fraction / fractionalBase);
        const af = a.fraction % fractionalBase;
        const bv = b.value + Math.floor(b.fraction / fractionalBase);
        const bf = b.fraction % fractionalBase;
        switch (true) {
            case av < bv:
                return -1;
            case av > bv:
                return 1;
            case af < bf:
                return -1;
            case af > bf:
                return 1;
            case af === bf:
                return 0;
            default:
                throw Error("assertion failed");
        }
    }
    /**
     * Create a copy of an amount.
     */
    function copy(a) {
        return {
            currency: a.currency,
            fraction: a.fraction,
            value: a.value,
        };
    }
    /**
     * Divide an amount.  Throws on division by zero.
     */
    function divide(a, n) {
        if (n === 0) {
            throw Error(`Division by 0`);
        }
        if (n === 1) {
            return { value: a.value, fraction: a.fraction, currency: a.currency };
        }
        const r = a.value % n;
        return {
            currency: a.currency,
            fraction: Math.floor((r * fractionalBase + a.fraction) / n),
            value: Math.floor(a.value / n),
        };
    }
    function isZero(a) {
        a = jsonifyAmount(a);
        return a.value === 0 && a.fraction === 0;
    }
    /**
     * Parse an amount like 'EUR:20.5' for 20 Euros and 50 ct.
     */
    function parse(s) {
        const res = s.match(/^([a-zA-Z0-9_*-]+):([0-9]+)([.][0-9]+)?$/);
        if (!res) {
            return undefined;
        }
        const tail = res[3] || ".0";
        if (tail.length > fractionalLength + 1) {
            return undefined;
        }
        const value = Number.parseInt(res[2]);
        if (value > maxAmountValue) {
            return undefined;
        }
        return {
            currency: res[1],
            fraction: Math.round(fractionalBase * Number.parseFloat(tail)),
            value,
        };
    }
    /**
     * Parse amount in standard string form (like 'EUR:20.5'),
     * throw if the input is not a valid amount.
     */
    function parseOrThrow(s) {
        const res = parse(s);
        if (!res) {
            throw Error(`Can't parse amount: "${s}"`);
        }
        return res;
    }
    /**
     * Convert a float to a Taler amount.
     * Loss of precision possible.
     */
    function fromFloat(floatVal, currency) {
        return {
            currency,
            fraction: Math.floor((floatVal - Math.floor(floatVal)) * fractionalBase),
            value: Math.floor(floatVal),
        };
    }
    /**
     * Convert to standard human-readable string representation that's
     * also used in JSON formats.
     */
    function stringify(a) {
        a = jsonifyAmount(a);
        const av = a.value + Math.floor(a.fraction / fractionalBase);
        const af = a.fraction % fractionalBase;
        let s = av.toString();
        if (af) {
            s = s + ".";
            let n = af;
            for (let i = 0; i < fractionalLength; i++) {
                if (!n) {
                    break;
                }
                s = s + Math.floor((n / fractionalBase) * 10).toString();
                n = (n * 10) % fractionalBase;
            }
        }
        return `${a.currency}:${s}`;
    }
    /**
     * Check if the argument is a valid amount in string form.
     */
    function check(a) {
        if (typeof a !== "string") {
            return false;
        }
        try {
            const parsedAmount = parse(a);
            return !!parsedAmount;
        }
        catch (_a) {
            return false;
        }
    }
    function mult(a, n) {
        if (!Number.isInteger(n)) {
            throw Error("amount can only be multipied by an integer");
        }
        if (n < 0) {
            throw Error("amount can only be multiplied by a positive integer");
        }
        if (n == 0) {
            return { amount: getZero(a.currency), saturated: false };
        }
        let x = a;
        let acc = getZero(a.currency);
        while (n > 1) {
            if (n % 2 == 0) {
                n = n / 2;
            }
            else {
                n = (n - 1) / 2;
                const r2 = add(acc, x);
                if (r2.saturated) {
                    return r2;
                }
                acc = r2.amount;
            }
            const r2 = add(x, x);
            if (r2.saturated) {
                return r2;
            }
            x = r2.amount;
        }
        return add(acc, x);
    }
    // Export all amount-related functions here for better IDE experience.
    const Amounts = {
        stringify: stringify,
        parse: parse,
        parseOrThrow: parseOrThrow,
        cmp: cmp,
        add: add,
        sum: sum,
        sub: sub,
        mult: mult,
        check: check,
        getZero: getZero,
        isZero: isZero,
        maxAmountValue: maxAmountValue,
        fromFloat: fromFloat,
        copy: copy,
        fractionalBase: fractionalBase,
        divide: divide,
    };

    /*
     This file is part of TALER
     (C) 2019 GNUnet e.V.

     TALER is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     TALER is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    /**
     * Check if we are running under nodejs.
     */
    const isNode = typeof process !== "undefined" && process.release.name === "node";
    function writeNodeLog(message, tag, level, args) {
        try {
            process.stderr.write(`${new Date().toISOString()} ${tag} ${level} `);
            process.stderr.write(`${message}`);
            if (args.length != 0) {
                process.stderr.write(" ");
                process.stderr.write(JSON.stringify(args, undefined, 2));
            }
            process.stderr.write("\n");
        }
        catch (e) {
            // This can happen when we're trying to log something that doesn't want to be
            // converted to a string.
            process.stderr.write(`${new Date().toISOString()} (logger) FATAL `);
            if (e instanceof Error) {
                process.stderr.write("failed to write log: ");
                process.stderr.write(e.message);
            }
            process.stderr.write("\n");
        }
    }
    /**
     * Logger that writes to stderr when running under node,
     * and uses the corresponding console.* method to log in the browser.
     */
    class Logger {
        constructor(tag) {
            this.tag = tag;
        }
        info(message, ...args) {
            if (isNode) {
                writeNodeLog(message, this.tag, "INFO", args);
            }
            else {
                console.info(`${new Date().toISOString()} ${this.tag} INFO ` + message, ...args);
            }
        }
        warn(message, ...args) {
            if (isNode) {
                writeNodeLog(message, this.tag, "WARN", args);
            }
            else {
                console.warn(`${new Date().toISOString()} ${this.tag} INFO ` + message, ...args);
            }
        }
        error(message, ...args) {
            if (isNode) {
                writeNodeLog(message, this.tag, "ERROR", args);
            }
            else {
                console.info(`${new Date().toISOString()} ${this.tag} ERROR ` + message, ...args);
            }
        }
        trace(message, ...args) {
            if (isNode) {
                writeNodeLog(message, this.tag, "TRACE", args);
            }
            else {
                console.info(`${new Date().toISOString()} ${this.tag} TRACE ` + message, ...args);
            }
        }
    }

    /*
     This file is part of TALER
     (C) 2016 GNUnet e.V.

     TALER is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     TALER is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    /**
     * Definition of an object store.
     */
    class Store {
        constructor(name, storeParams) {
            this.name = name;
            this.storeParams = storeParams;
        }
    }
    /**
     * Definition of an index.
     */
    class Index {
        constructor(s, indexName, keyPath, options) {
            this.indexName = indexName;
            this.keyPath = keyPath;
            const defaultOptions = {
                multiEntry: false,
            };
            this.options = Object.assign(Object.assign({}, defaultOptions), (options || {}));
            this.storeName = s.name;
        }
    }

    /*
     This file is part of GNU Taler
     (C) 2018-2020 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var ReserveRecordStatus;
    (function (ReserveRecordStatus) {
        /**
         * Reserve must be registered with the bank.
         */
        ReserveRecordStatus["REGISTERING_BANK"] = "registering-bank";
        /**
         * We've registered reserve's information with the bank
         * and are now waiting for the user to confirm the withdraw
         * with the bank (typically 2nd factor auth).
         */
        ReserveRecordStatus["WAIT_CONFIRM_BANK"] = "wait-confirm-bank";
        /**
         * Querying reserve status with the exchange.
         */
        ReserveRecordStatus["QUERYING_STATUS"] = "querying-status";
        /**
         * The corresponding withdraw record has been created.
         * No further processing is done, unless explicitly requested
         * by the user.
         */
        ReserveRecordStatus["DORMANT"] = "dormant";
        /**
         * The bank aborted the withdrawal.
         */
        ReserveRecordStatus["BANK_ABORTED"] = "bank-aborted";
    })(ReserveRecordStatus || (ReserveRecordStatus = {}));
    /**
     * Status of a denomination.
     */
    var DenominationStatus;
    (function (DenominationStatus) {
        /**
         * Verification was delayed.
         */
        DenominationStatus[DenominationStatus["Unverified"] = 0] = "Unverified";
        /**
         * Verified as valid.
         */
        DenominationStatus[DenominationStatus["VerifiedGood"] = 1] = "VerifiedGood";
        /**
         * Verified as invalid.
         */
        DenominationStatus[DenominationStatus["VerifiedBad"] = 2] = "VerifiedBad";
    })(DenominationStatus || (DenominationStatus = {}));
    var ExchangeUpdateStatus;
    (function (ExchangeUpdateStatus) {
        ExchangeUpdateStatus["FetchKeys"] = "fetch-keys";
        ExchangeUpdateStatus["FetchWire"] = "fetch-wire";
        ExchangeUpdateStatus["FetchTerms"] = "fetch-terms";
        ExchangeUpdateStatus["FinalizeUpdate"] = "finalize-update";
        ExchangeUpdateStatus["Finished"] = "finished";
    })(ExchangeUpdateStatus || (ExchangeUpdateStatus = {}));
    var ExchangeUpdateReason;
    (function (ExchangeUpdateReason) {
        ExchangeUpdateReason["Initial"] = "initial";
        ExchangeUpdateReason["Forced"] = "forced";
        ExchangeUpdateReason["Scheduled"] = "scheduled";
    })(ExchangeUpdateReason || (ExchangeUpdateReason = {}));
    /**
     * Status of a coin.
     */
    var CoinStatus;
    (function (CoinStatus) {
        /**
         * Withdrawn and never shown to anybody.
         */
        CoinStatus["Fresh"] = "fresh";
        /**
         * A coin that has been spent and refreshed.
         */
        CoinStatus["Dormant"] = "dormant";
    })(CoinStatus || (CoinStatus = {}));
    var CoinSourceType;
    (function (CoinSourceType) {
        CoinSourceType["Withdraw"] = "withdraw";
        CoinSourceType["Refresh"] = "refresh";
        CoinSourceType["Tip"] = "tip";
    })(CoinSourceType || (CoinSourceType = {}));
    var ProposalStatus;
    (function (ProposalStatus) {
        /**
         * Not downloaded yet.
         */
        ProposalStatus["DOWNLOADING"] = "downloading";
        /**
         * Proposal downloaded, but the user needs to accept/reject it.
         */
        ProposalStatus["PROPOSED"] = "proposed";
        /**
         * The user has accepted the proposal.
         */
        ProposalStatus["ACCEPTED"] = "accepted";
        /**
         * The user has rejected the proposal.
         */
        ProposalStatus["REFUSED"] = "refused";
        /**
         * Downloading or processing the proposal has failed permanently.
         */
        ProposalStatus["PERMANENTLY_FAILED"] = "permanently-failed";
        /**
         * Downloaded proposal was detected as a re-purchase.
         */
        ProposalStatus["REPURCHASE"] = "repurchase";
    })(ProposalStatus || (ProposalStatus = {}));
    var RefundState;
    (function (RefundState) {
        RefundState["Failed"] = "failed";
        RefundState["Applied"] = "applied";
        RefundState["Pending"] = "pending";
    })(RefundState || (RefundState = {}));
    var RefundReason;
    (function (RefundReason) {
        /**
         * Normal refund given by the merchant.
         */
        RefundReason["NormalRefund"] = "normal-refund";
        /**
         * Refund from an aborted payment.
         */
        RefundReason["AbortRefund"] = "abort-pay-refund";
    })(RefundReason || (RefundReason = {}));
    var AbortStatus;
    (function (AbortStatus) {
        AbortStatus["None"] = "none";
        AbortStatus["AbortRefund"] = "abort-refund";
        AbortStatus["AbortFinished"] = "abort-finished";
    })(AbortStatus || (AbortStatus = {}));
    var ImportPayloadType;
    (function (ImportPayloadType) {
        ImportPayloadType["CoreSchema"] = "core-schema";
    })(ImportPayloadType || (ImportPayloadType = {}));
    class ExchangesStore extends Store {
        constructor() {
            super("exchanges", { keyPath: "baseUrl" });
        }
    }
    class CoinsStore extends Store {
        constructor() {
            super("coins", { keyPath: "coinPub" });
            this.exchangeBaseUrlIndex = new Index(this, "exchangeBaseUrl", "exchangeBaseUrl");
            this.denomPubHashIndex = new Index(this, "denomPubHashIndex", "denomPubHash");
            this.coinEvHashIndex = new Index(this, "coinEvHashIndex", "coinEvHash");
        }
    }
    class ProposalsStore extends Store {
        constructor() {
            super("proposals", { keyPath: "proposalId" });
            this.urlAndOrderIdIndex = new Index(this, "urlIndex", ["merchantBaseUrl", "orderId"]);
        }
    }
    class PurchasesStore extends Store {
        constructor() {
            super("purchases", { keyPath: "proposalId" });
            this.fulfillmentUrlIndex = new Index(this, "fulfillmentUrlIndex", "download.contractData.fulfillmentUrl");
            this.orderIdIndex = new Index(this, "orderIdIndex", ["download.contractData.merchantBaseUrl", "download.contractData.orderId"]);
        }
    }
    class DenominationsStore extends Store {
        constructor() {
            // cast needed because of bug in type annotations
            super("denominations", {
                keyPath: ["exchangeBaseUrl", "denomPubHash"],
            });
            this.exchangeBaseUrlIndex = new Index(this, "exchangeBaseUrlIndex", "exchangeBaseUrl");
        }
    }
    class CurrenciesStore extends Store {
        constructor() {
            super("currencies", { keyPath: "name" });
        }
    }
    class ConfigStore extends Store {
        constructor() {
            super("config", { keyPath: "key" });
        }
    }
    class ReservesStore extends Store {
        constructor() {
            super("reserves", { keyPath: "reservePub" });
        }
    }
    class TipsStore extends Store {
        constructor() {
            super("tips", { keyPath: "walletTipId" });
            // Added in version 2
            this.byMerchantTipIdAndBaseUrl = new Index(this, "tipsByMerchantTipIdAndOriginIndex", ["merchantTipId", "merchantBaseUrl"], {
                versionAdded: 2,
            });
        }
    }
    class WithdrawalGroupsStore extends Store {
        constructor() {
            super("withdrawals", { keyPath: "withdrawalGroupId" });
            this.byReservePub = new Index(this, "withdrawalsByReserveIndex", "reservePub");
        }
    }
    class PlanchetsStore extends Store {
        constructor() {
            super("planchets", { keyPath: "coinPub" });
            this.byGroupAndIndex = new Index(this, "withdrawalGroupAndCoinIdxIndex", ["withdrawalGroupId", "coinIdx"]);
            this.byGroup = new Index(this, "withdrawalGroupIndex", "withdrawalGroupId");
            this.coinEvHashIndex = new Index(this, "coinEvHashIndex", "coinEvHash");
        }
    }
    /**
     * This store is effectively a materialized index for
     * reserve records that are for a bank-integrated withdrawal.
     */
    class BankWithdrawUrisStore extends Store {
        constructor() {
            super("bankWithdrawUris", { keyPath: "talerWithdrawUri" });
        }
    }
    /**
     */
    class BackupProvidersStore extends Store {
        constructor() {
            super("backupProviders", { keyPath: "baseUrl", versionAdded: 3 });
        }
    }
    /**
     * The stores and indices for the wallet database.
     */
    const Stores = {
        coins: new CoinsStore(),
        config: new ConfigStore(),
        currencies: new CurrenciesStore(),
        denominations: new DenominationsStore(),
        exchanges: new ExchangesStore(),
        proposals: new ProposalsStore(),
        refreshGroups: new Store("refreshGroups", {
            keyPath: "refreshGroupId",
        }),
        recoupGroups: new Store("recoupGroups", {
            keyPath: "recoupGroupId",
        }),
        reserves: new ReservesStore(),
        purchases: new PurchasesStore(),
        tips: new TipsStore(),
        withdrawalGroups: new WithdrawalGroupsStore(),
        planchets: new PlanchetsStore(),
        bankWithdrawUris: new BankWithdrawUrisStore(),
        backupProviders: new BackupProvidersStore(),
    };

    /**
     * Truncate a timestamp so that that it represents a multiple
     * of seconds.  The timestamp is always rounded down.
     */
    function timestampTruncateToSecond(t1) {
        if (t1.t_ms === "never") {
            return { t_ms: "never" };
        }
        return {
            t_ms: Math.floor(t1.t_ms / 1000) * 1000,
        };
    }

    /*
     This file is part of GNU Taler
     (C) 2019 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var TransactionType;
    (function (TransactionType) {
        TransactionType["Withdrawal"] = "withdrawal";
        TransactionType["Payment"] = "payment";
        TransactionType["Refund"] = "refund";
        TransactionType["Refresh"] = "refresh";
        TransactionType["Tip"] = "tip";
    })(TransactionType || (TransactionType = {}));
    var WithdrawalType;
    (function (WithdrawalType) {
        WithdrawalType["TalerBankIntegrationApi"] = "taler-bank-integration-api";
        WithdrawalType["ManualTransfer"] = "manual-transfer";
    })(WithdrawalType || (WithdrawalType = {}));
    var PaymentStatus;
    (function (PaymentStatus) {
        /**
         * Explicitly aborted after timeout / failure
         */
        PaymentStatus["Aborted"] = "aborted";
        /**
         * Payment failed, wallet will auto-retry.
         * User should be given the option to retry now / abort.
         */
        PaymentStatus["Failed"] = "failed";
        /**
         * Paid successfully
         */
        PaymentStatus["Paid"] = "paid";
        /**
         * User accepted, payment is processing.
         */
        PaymentStatus["Accepted"] = "accepted";
    })(PaymentStatus || (PaymentStatus = {}));

    /*
     This file is part of GNU Taler
     (C) 2015-2020 Taler Systems SA

     TALER is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     TALER is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var ConfirmPayResultType;
    (function (ConfirmPayResultType) {
        ConfirmPayResultType["Done"] = "done";
        ConfirmPayResultType["Pending"] = "pending";
    })(ConfirmPayResultType || (ConfirmPayResultType = {}));
    var PreparePayResultType;
    (function (PreparePayResultType) {
        PreparePayResultType["PaymentPossible"] = "payment-possible";
        PreparePayResultType["InsufficientBalance"] = "insufficient-balance";
        PreparePayResultType["AlreadyConfirmed"] = "already-confirmed";
    })(PreparePayResultType || (PreparePayResultType = {}));
    /**
     * Reasons for why a coin is being refreshed.
     */
    var RefreshReason;
    (function (RefreshReason) {
        RefreshReason["Manual"] = "manual";
        RefreshReason["Pay"] = "pay";
        RefreshReason["Refund"] = "refund";
        RefreshReason["AbortPay"] = "abort-pay";
        RefreshReason["Recoup"] = "recoup";
        RefreshReason["BackupRestored"] = "backup-restored";
        RefreshReason["Scheduled"] = "scheduled";
    })(RefreshReason || (RefreshReason = {}));

    /*
     This file is part of GNU Taler
     (C) 2020 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    // @ts-ignore
    const _URL = globalThis.URL;
    if (!_URL) {
        throw Error("FATAL: URL not available");
    }
    // @ts-ignore
    const _URLSearchParams = globalThis.URLSearchParams;
    if (!_URLSearchParams) {
        throw Error("FATAL: URLSearchParams not available");
    }

    /*
     This file is part of GNU Taler
     (C) 2019-2020 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var TalerUriType;
    (function (TalerUriType) {
        TalerUriType["TalerPay"] = "taler-pay";
        TalerUriType["TalerWithdraw"] = "taler-withdraw";
        TalerUriType["TalerTip"] = "taler-tip";
        TalerUriType["TalerRefund"] = "taler-refund";
        TalerUriType["TalerNotifyReserve"] = "taler-notify-reserve";
        TalerUriType["Unknown"] = "unknown";
    })(TalerUriType || (TalerUriType = {}));

    /*
      This file is part of GNU Taler
      Copyright (C) 2012-2020 Taler Systems SA

      GNU Taler is free software: you can redistribute it and/or modify it
      under the terms of the GNU Lesser General Public License as published
      by the Free Software Foundation, either version 3 of the License,
      or (at your option) any later version.

      GNU Taler is distributed in the hope that it will be useful, but
      WITHOUT ANY WARRANTY; without even the implied warranty of
      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
      Lesser General Public License for more details.

      You should have received a copy of the GNU Lesser General Public License
      along with this program.  If not, see <http://www.gnu.org/licenses/>.

      SPDX-License-Identifier: LGPL3.0-or-later

      Note: the LGPL does not apply to all components of GNU Taler,
      but it does apply to this file.
     */
    var TalerErrorCode;
    (function (TalerErrorCode) {
        /**
         * Special code to indicate success (no error).
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["NONE"] = 0] = "NONE";
        /**
         * A non-integer error code was returned in the JSON response.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["INVALID"] = 1] = "INVALID";
        /**
         * The response we got from the server was not even in JSON format.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_INVALID_RESPONSE"] = 10] = "GENERIC_INVALID_RESPONSE";
        /**
         * An operation timed out.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_TIMEOUT"] = 11] = "GENERIC_TIMEOUT";
        /**
         * The version string given does not follow the expected CURRENT:REVISION:AGE Format.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_VERSION_MALFORMED"] = 12] = "GENERIC_VERSION_MALFORMED";
        /**
         * The service responded with a reply that was in JSON but did not satsify the protocol. Note that invalid cryptographic signatures should have signature-specific error codes.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_REPLY_MALFORMED"] = 13] = "GENERIC_REPLY_MALFORMED";
        /**
         * The HTTP method used is invalid for this endpoint.
         * Returned with an HTTP status code of #MHD_HTTP_METHOD_NOT_ALLOWED (405).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_METHOD_INVALID"] = 20] = "GENERIC_METHOD_INVALID";
        /**
         * There is no endpoint defined for the URL provided by the client.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_ENDPOINT_UNKNOWN"] = 21] = "GENERIC_ENDPOINT_UNKNOWN";
        /**
         * The JSON in the client's request was malformed (generic parse error).
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_JSON_INVALID"] = 22] = "GENERIC_JSON_INVALID";
        /**
         * The payto:// URI provided by the client is malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_PAYTO_URI_MALFORMED"] = 24] = "GENERIC_PAYTO_URI_MALFORMED";
        /**
         * A required parameter in the request was missing.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_PARAMETER_MISSING"] = 25] = "GENERIC_PARAMETER_MISSING";
        /**
         * A parameter in the request was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_PARAMETER_MALFORMED"] = 26] = "GENERIC_PARAMETER_MALFORMED";
        /**
         * The currencies involved in the operation do not match.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_CURRENCY_MISMATCH"] = 30] = "GENERIC_CURRENCY_MISMATCH";
        /**
         * The URI is longer than the longest URI the HTTP server is willing to parse.
         * Returned with an HTTP status code of #MHD_HTTP_URI_TOO_LONG (414).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_URI_TOO_LONG"] = 31] = "GENERIC_URI_TOO_LONG";
        /**
         * The body is too large to be permissible for the endpoint.
         * Returned with an HTTP status code of #MHD_HTTP_PAYLOAD_TOO_LARGE (413).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_UPLOAD_EXCEEDS_LIMIT"] = 32] = "GENERIC_UPLOAD_EXCEEDS_LIMIT";
        /**
         * The service failed initialize its connection to the database.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_SETUP_FAILED"] = 50] = "GENERIC_DB_SETUP_FAILED";
        /**
         * The service encountered an error event to just start the database transaction.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_START_FAILED"] = 51] = "GENERIC_DB_START_FAILED";
        /**
         * The service failed to store information in its database.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_STORE_FAILED"] = 52] = "GENERIC_DB_STORE_FAILED";
        /**
         * The service failed to fetch information from its database.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_FETCH_FAILED"] = 53] = "GENERIC_DB_FETCH_FAILED";
        /**
         * The service encountered an error event to commit the database transaction (hard, unrecoverable error).
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_COMMIT_FAILED"] = 54] = "GENERIC_DB_COMMIT_FAILED";
        /**
         * The service encountered an error event to commit the database transaction, even after repeatedly retrying it there was always a conflicting transaction. (This indicates a repeated serialization error; should only happen if some client maliciously tries to create conflicting concurrent transactions.)
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_SOFT_FAILURE"] = 55] = "GENERIC_DB_SOFT_FAILURE";
        /**
         * The service's database is inconsistent and violates service-internal invariants.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_DB_INVARIANT_FAILURE"] = 56] = "GENERIC_DB_INVARIANT_FAILURE";
        /**
         * The HTTP server experienced an internal invariant failure (bug).
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_INTERNAL_INVARIANT_FAILURE"] = 60] = "GENERIC_INTERNAL_INVARIANT_FAILURE";
        /**
         * The service could not compute a cryptographic hash over some JSON value.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_FAILED_COMPUTE_JSON_HASH"] = 61] = "GENERIC_FAILED_COMPUTE_JSON_HASH";
        /**
         * The HTTP server had insufficient memory to parse the request.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_PARSER_OUT_OF_MEMORY"] = 70] = "GENERIC_PARSER_OUT_OF_MEMORY";
        /**
         * The HTTP server failed to allocate memory.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_ALLOCATION_FAILURE"] = 71] = "GENERIC_ALLOCATION_FAILURE";
        /**
         * The HTTP server failed to allocate memory for building JSON reply.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["GENERIC_JSON_ALLOCATION_FAILURE"] = 72] = "GENERIC_JSON_ALLOCATION_FAILURE";
        /**
         * Exchange is badly configured and thus cannot operate.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_BAD_CONFIGURATION"] = 1000] = "EXCHANGE_GENERIC_BAD_CONFIGURATION";
        /**
         * Operation specified unknown for this endpoint.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_OPERATION_UNKNOWN"] = 1001] = "EXCHANGE_GENERIC_OPERATION_UNKNOWN";
        /**
         * The number of segments included in the URI does not match the number of segments expected by the endpoint.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_WRONG_NUMBER_OF_SEGMENTS"] = 1002] = "EXCHANGE_GENERIC_WRONG_NUMBER_OF_SEGMENTS";
        /**
         * The same coin was already used with a different denomination previously.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_COIN_CONFLICTING_DENOMINATION_KEY"] = 1003] = "EXCHANGE_GENERIC_COIN_CONFLICTING_DENOMINATION_KEY";
        /**
         * The public key of given to a "/coins/" endpoint of the exchange was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_COINS_INVALID_COIN_PUB"] = 1004] = "EXCHANGE_GENERIC_COINS_INVALID_COIN_PUB";
        /**
         * The exchange is not aware of the denomination key the wallet requested for the operation.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_DENOMINATION_KEY_UNKNOWN"] = 1005] = "EXCHANGE_GENERIC_DENOMINATION_KEY_UNKNOWN";
        /**
         * The signature of the denomination key over the coin is not valid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DENOMINATION_SIGNATURE_INVALID"] = 1006] = "EXCHANGE_DENOMINATION_SIGNATURE_INVALID";
        /**
         * The exchange failed to perform the operation as it could not find the private keys. This is a problem with the exchange setup, not with the client's request.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_KEYS_MISSING"] = 1007] = "EXCHANGE_GENERIC_KEYS_MISSING";
        /**
         * Validity period of the denomination lies in the future.
         * Returned with an HTTP status code of #MHD_HTTP_PRECONDITION_FAILED (412).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_DENOMINATION_VALIDITY_IN_FUTURE"] = 1008] = "EXCHANGE_GENERIC_DENOMINATION_VALIDITY_IN_FUTURE";
        /**
         * Denomination key of the coin is past its expiration time for the requested operation.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_GENERIC_DENOMINATION_EXPIRED"] = 1009] = "EXCHANGE_GENERIC_DENOMINATION_EXPIRED";
        /**
         * The exchange did not find information about the specified transaction in the database.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_NOT_FOUND"] = 1100] = "EXCHANGE_DEPOSITS_GET_NOT_FOUND";
        /**
         * The wire hash of given to a "/deposits/" handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_INVALID_H_WIRE"] = 1101] = "EXCHANGE_DEPOSITS_GET_INVALID_H_WIRE";
        /**
         * The merchant key of given to a "/deposits/" handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_INVALID_MERCHANT_PUB"] = 1102] = "EXCHANGE_DEPOSITS_GET_INVALID_MERCHANT_PUB";
        /**
         * The hash of the contract terms given to a "/deposits/" handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_INVALID_H_CONTRACT_TERMS"] = 1103] = "EXCHANGE_DEPOSITS_GET_INVALID_H_CONTRACT_TERMS";
        /**
         * The coin public key of given to a "/deposits/" handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_INVALID_COIN_PUB"] = 1104] = "EXCHANGE_DEPOSITS_GET_INVALID_COIN_PUB";
        /**
         * The signature returned by the exchange in a /deposits/ request was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_INVALID_SIGNATURE_BY_EXCHANGE"] = 1105] = "EXCHANGE_DEPOSITS_GET_INVALID_SIGNATURE_BY_EXCHANGE";
        /**
         * The signature of the merchant is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSITS_GET_MERCHANT_SIGNATURE_INVALID"] = 1106] = "EXCHANGE_DEPOSITS_GET_MERCHANT_SIGNATURE_INVALID";
        /**
         * The given reserve does not have sufficient funds to admit the requested withdraw operation at this time.  The response includes the current "balance" of the reserve as well as the transaction "history" that lead to this balance.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_INSUFFICIENT_FUNDS"] = 1150] = "EXCHANGE_WITHDRAW_INSUFFICIENT_FUNDS";
        /**
         * The exchange has no information about the "reserve_pub" that was given.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_RESERVE_UNKNOWN"] = 1151] = "EXCHANGE_WITHDRAW_RESERVE_UNKNOWN";
        /**
         * The amount to withdraw together with the fee exceeds the numeric range for Taler amounts.  This is not a client failure, as the coin value and fees come from the exchange's configuration.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_AMOUNT_FEE_OVERFLOW"] = 1152] = "EXCHANGE_WITHDRAW_AMOUNT_FEE_OVERFLOW";
        /**
         * The exchange failed to create the signature using the denomination key.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_SIGNATURE_FAILED"] = 1153] = "EXCHANGE_WITHDRAW_SIGNATURE_FAILED";
        /**
         * The signature of the reserve is not valid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_RESERVE_SIGNATURE_INVALID"] = 1154] = "EXCHANGE_WITHDRAW_RESERVE_SIGNATURE_INVALID";
        /**
         * When computing the reserve history, we ended up with a negative overall balance, which should be impossible.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_HISTORY_ERROR_INSUFFICIENT_FUNDS"] = 1155] = "EXCHANGE_WITHDRAW_HISTORY_ERROR_INSUFFICIENT_FUNDS";
        /**
         * Withdraw period of the coin to be withdrawn is in the past.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_VALIDITY_IN_PAST"] = 1157] = "EXCHANGE_WITHDRAW_VALIDITY_IN_PAST";
        /**
         * Withdraw period of the coin to be withdrawn is in the past.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_DENOMINATION_KEY_LOST"] = 1158] = "EXCHANGE_WITHDRAW_DENOMINATION_KEY_LOST";
        /**
         * The client failed to unblind the blind signature.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WITHDRAW_UNBLIND_FAILURE"] = 1159] = "EXCHANGE_WITHDRAW_UNBLIND_FAILURE";
        /**
         * The respective coin did not have sufficient residual value for the /deposit operation (i.e. due to double spending). The "history" in the response provides the transaction history of the coin proving this fact.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_INSUFFICIENT_FUNDS"] = 1200] = "EXCHANGE_DEPOSIT_INSUFFICIENT_FUNDS";
        /**
         * The signature made by the coin over the deposit permission is not valid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_COIN_SIGNATURE_INVALID"] = 1205] = "EXCHANGE_DEPOSIT_COIN_SIGNATURE_INVALID";
        /**
         * The stated value of the coin after the deposit fee is subtracted would be negative.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_NEGATIVE_VALUE_AFTER_FEE"] = 1207] = "EXCHANGE_DEPOSIT_NEGATIVE_VALUE_AFTER_FEE";
        /**
         * The stated refund deadline is after the wire deadline.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_REFUND_DEADLINE_AFTER_WIRE_DEADLINE"] = 1208] = "EXCHANGE_DEPOSIT_REFUND_DEADLINE_AFTER_WIRE_DEADLINE";
        /**
         * The exchange failed to canonicalize and hash the given wire format. For example, the merchant failed to provide the "salt" or a valid payto:// URI in the wire details.  Note that while the exchange will do some basic sanity checking on the wire details, it cannot warrant that the banking system will ultimately be able to route to the specified address, even if this check passed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_INVALID_WIRE_FORMAT_JSON"] = 1210] = "EXCHANGE_DEPOSIT_INVALID_WIRE_FORMAT_JSON";
        /**
         * The hash of the given wire address does not match the wire hash specified in the proposal data.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_INVALID_WIRE_FORMAT_CONTRACT_HASH_CONFLICT"] = 1211] = "EXCHANGE_DEPOSIT_INVALID_WIRE_FORMAT_CONTRACT_HASH_CONFLICT";
        /**
         * The signature provided by the exchange is not valid.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_DEPOSIT_INVALID_SIGNATURE_BY_EXCHANGE"] = 1221] = "EXCHANGE_DEPOSIT_INVALID_SIGNATURE_BY_EXCHANGE";
        /**
         * The reserve status was requested using a unknown key.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_RESERVES_GET_STATUS_UNKNOWN"] = 1250] = "EXCHANGE_RESERVES_GET_STATUS_UNKNOWN";
        /**
         * The respective coin did not have sufficient residual value for the /refresh/melt operation.  The "history" in this response provdes the "residual_value" of the coin, which may be less than its "original_value".
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_INSUFFICIENT_FUNDS"] = 1300] = "EXCHANGE_MELT_INSUFFICIENT_FUNDS";
        /**
         * The exchange had an internal error reconstructing the transaction history of the coin that was being melted.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_COIN_HISTORY_COMPUTATION_FAILED"] = 1301] = "EXCHANGE_MELT_COIN_HISTORY_COMPUTATION_FAILED";
        /**
         * The exchange encountered melt fees exceeding the melted coin's contribution.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_FEES_EXCEED_CONTRIBUTION"] = 1302] = "EXCHANGE_MELT_FEES_EXCEED_CONTRIBUTION";
        /**
         * The signature made with the coin to be melted is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_COIN_SIGNATURE_INVALID"] = 1303] = "EXCHANGE_MELT_COIN_SIGNATURE_INVALID";
        /**
         * The exchange failed to obtain the transaction history of the given coin from the database while generating an insufficient funds errors.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_HISTORY_DB_ERROR_INSUFFICIENT_FUNDS"] = 1304] = "EXCHANGE_MELT_HISTORY_DB_ERROR_INSUFFICIENT_FUNDS";
        /**
         * The denomination of the given coin has past its expiration date and it is also not a valid zombie (that is, was not refreshed with the fresh coin being subjected to recoup).
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_COIN_EXPIRED_NO_ZOMBIE"] = 1305] = "EXCHANGE_MELT_COIN_EXPIRED_NO_ZOMBIE";
        /**
         * The signature returned by the exchange in a melt request was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_MELT_INVALID_SIGNATURE_BY_EXCHANGE"] = 1306] = "EXCHANGE_MELT_INVALID_SIGNATURE_BY_EXCHANGE";
        /**
         * The provided transfer keys do not match up with the original commitment.  Information about the original commitment is included in the response.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_COMMITMENT_VIOLATION"] = 1353] = "EXCHANGE_REFRESHES_REVEAL_COMMITMENT_VIOLATION";
        /**
         * Failed to produce the blinded signatures over the coins to be returned.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_SIGNING_ERROR"] = 1354] = "EXCHANGE_REFRESHES_REVEAL_SIGNING_ERROR";
        /**
         * The exchange is unaware of the refresh session specified in the request.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_SESSION_UNKNOWN"] = 1355] = "EXCHANGE_REFRESHES_REVEAL_SESSION_UNKNOWN";
        /**
         * The size of the cut-and-choose dimension of the private transfer keys request does not match #TALER_CNC_KAPPA - 1.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_CNC_TRANSFER_ARRAY_SIZE_INVALID"] = 1356] = "EXCHANGE_REFRESHES_REVEAL_CNC_TRANSFER_ARRAY_SIZE_INVALID";
        /**
         * The number of coins to be created in refresh exceeds the limits of the exchange. private transfer keys request does not match #TALER_CNC_KAPPA - 1.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_NEW_DENOMS_ARRAY_SIZE_EXCESSIVE"] = 1357] = "EXCHANGE_REFRESHES_REVEAL_NEW_DENOMS_ARRAY_SIZE_EXCESSIVE";
        /**
         * The number of envelopes given does not match the number of denomination keys given.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_NEW_DENOMS_ARRAY_SIZE_MISMATCH"] = 1358] = "EXCHANGE_REFRESHES_REVEAL_NEW_DENOMS_ARRAY_SIZE_MISMATCH";
        /**
         * The exchange encountered a numeric overflow totaling up the cost for the refresh operation.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_COST_CALCULATION_OVERFLOW"] = 1359] = "EXCHANGE_REFRESHES_REVEAL_COST_CALCULATION_OVERFLOW";
        /**
         * The exchange's cost calculation shows that the melt amount is below the costs of the transaction.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_AMOUNT_INSUFFICIENT"] = 1360] = "EXCHANGE_REFRESHES_REVEAL_AMOUNT_INSUFFICIENT";
        /**
         * The signature made with the coin over the link data is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_LINK_SIGNATURE_INVALID"] = 1361] = "EXCHANGE_REFRESHES_REVEAL_LINK_SIGNATURE_INVALID";
        /**
         * The refresh session hash given to a /refreshes/ handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_INVALID_RCH"] = 1362] = "EXCHANGE_REFRESHES_REVEAL_INVALID_RCH";
        /**
         * Operation specified invalid for this endpoint.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFRESHES_REVEAL_OPERATION_INVALID"] = 1363] = "EXCHANGE_REFRESHES_REVEAL_OPERATION_INVALID";
        /**
         * The coin specified in the link request is unknown to the exchange.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_LINK_COIN_UNKNOWN"] = 1400] = "EXCHANGE_LINK_COIN_UNKNOWN";
        /**
         * The public key of given to a /transfers/ handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_TRANSFERS_GET_WTID_MALFORMED"] = 1450] = "EXCHANGE_TRANSFERS_GET_WTID_MALFORMED";
        /**
         * The exchange did not find information about the specified wire transfer identifier in the database.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_TRANSFERS_GET_WTID_NOT_FOUND"] = 1451] = "EXCHANGE_TRANSFERS_GET_WTID_NOT_FOUND";
        /**
         * The exchange did not find information about the wire transfer fees it charged.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_TRANSFERS_GET_WIRE_FEE_NOT_FOUND"] = 1452] = "EXCHANGE_TRANSFERS_GET_WIRE_FEE_NOT_FOUND";
        /**
         * The exchange found a wire fee that was above the total transfer value (and thus could not have been charged).
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_TRANSFERS_GET_WIRE_FEE_INCONSISTENT"] = 1453] = "EXCHANGE_TRANSFERS_GET_WIRE_FEE_INCONSISTENT";
        /**
         * The exchange knows literally nothing about the coin we were asked to refund. But without a transaction history, we cannot issue a refund. This is kind-of OK, the owner should just refresh it directly without executing the refund.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_COIN_NOT_FOUND"] = 1500] = "EXCHANGE_REFUND_COIN_NOT_FOUND";
        /**
         * We could not process the refund request as the coin's transaction history does not permit the requested refund because then refunds would exceed the deposit amount.  The "history" in the response proves this.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_CONFLICT_DEPOSIT_INSUFFICIENT"] = 1501] = "EXCHANGE_REFUND_CONFLICT_DEPOSIT_INSUFFICIENT";
        /**
         * The exchange knows about the coin we were asked to refund, but not about the specific /deposit operation.  Hence, we cannot issue a refund (as we do not know if this merchant public key is authorized to do a refund).
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_DEPOSIT_NOT_FOUND"] = 1502] = "EXCHANGE_REFUND_DEPOSIT_NOT_FOUND";
        /**
         * The exchange can no longer refund the customer/coin as the money was already transferred (paid out) to the merchant. (It should be past the refund deadline.)
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_MERCHANT_ALREADY_PAID"] = 1503] = "EXCHANGE_REFUND_MERCHANT_ALREADY_PAID";
        /**
         * The refund fee specified for the request is lower than the refund fee charged by the exchange for the given denomination key of the refunded coin.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_FEE_TOO_LOW"] = 1504] = "EXCHANGE_REFUND_FEE_TOO_LOW";
        /**
         * The refunded amount is smaller than the refund fee, which would result in a negative refund.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_FEE_ABOVE_AMOUNT"] = 1505] = "EXCHANGE_REFUND_FEE_ABOVE_AMOUNT";
        /**
         * The signature of the merchant is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_MERCHANT_SIGNATURE_INVALID"] = 1506] = "EXCHANGE_REFUND_MERCHANT_SIGNATURE_INVALID";
        /**
         * Merchant backend failed to create the refund confirmation signature.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_MERCHANT_SIGNING_FAILED"] = 1507] = "EXCHANGE_REFUND_MERCHANT_SIGNING_FAILED";
        /**
         * The signature returned by the exchange in a refund request was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_INVALID_SIGNATURE_BY_EXCHANGE"] = 1508] = "EXCHANGE_REFUND_INVALID_SIGNATURE_BY_EXCHANGE";
        /**
         * The failure proof returned by the exchange is incorrect.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_INVALID_FAILURE_PROOF_BY_EXCHANGE"] = 1509] = "EXCHANGE_REFUND_INVALID_FAILURE_PROOF_BY_EXCHANGE";
        /**
         * Conflicting refund granted before with different amount but same refund transaction ID.
         * Returned with an HTTP status code of #MHD_HTTP_FAILED_DEPENDENCY (424).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_REFUND_INCONSISTENT_AMOUNT"] = 1510] = "EXCHANGE_REFUND_INCONSISTENT_AMOUNT";
        /**
         * The given coin signature is invalid for the request.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_RECOUP_SIGNATURE_INVALID"] = 1550] = "EXCHANGE_RECOUP_SIGNATURE_INVALID";
        /**
         * The exchange could not find the corresponding withdraw operation. The request is denied.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_RECOUP_WITHDRAW_NOT_FOUND"] = 1551] = "EXCHANGE_RECOUP_WITHDRAW_NOT_FOUND";
        /**
         * The coin's remaining balance is zero.  The request is denied.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_RECOUP_COIN_BALANCE_ZERO"] = 1552] = "EXCHANGE_RECOUP_COIN_BALANCE_ZERO";
        /**
         * The exchange failed to reproduce the coin's blinding.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_RECOUP_BLINDING_FAILED"] = 1553] = "EXCHANGE_RECOUP_BLINDING_FAILED";
        /**
         * The coin's remaining balance is zero.  The request is denied.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_RECOUP_COIN_BALANCE_NEGATIVE"] = 1554] = "EXCHANGE_RECOUP_COIN_BALANCE_NEGATIVE";
        /**
         * This exchange does not allow clients to request /keys for times other than the current (exchange) time.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_KEYS_TIMETRAVEL_FORBIDDEN"] = 1600] = "EXCHANGE_KEYS_TIMETRAVEL_FORBIDDEN";
        /**
         * A signature in the server's response was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["EXCHANGE_WIRE_SIGNATURE_INVALID"] = 1650] = "EXCHANGE_WIRE_SIGNATURE_INVALID";
        /**
         * The backend could not find the merchant instance specified in the request.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_INSTANCE_UNKNOWN"] = 2000] = "MERCHANT_GENERIC_INSTANCE_UNKNOWN";
        /**
         * The start and end-times in the wire fee structure leave a hole. This is not allowed.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_HOLE_IN_WIRE_FEE_STRUCTURE"] = 2001] = "MERCHANT_GENERIC_HOLE_IN_WIRE_FEE_STRUCTURE";
        /**
         * The reserve key of given to a /reserves/ handler was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_RESERVE_PUB_MALFORMED"] = 2002] = "MERCHANT_GENERIC_RESERVE_PUB_MALFORMED";
        /**
         * The backend could not locate a required template to generate an HTML reply.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_ACCEPTABLE (406).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_FAILED_TO_LOAD_TEMPLATE"] = 2003] = "MERCHANT_GENERIC_FAILED_TO_LOAD_TEMPLATE";
        /**
         * The backend could not expand the template to generate an HTML reply.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_FAILED_TO_EXPAND_TEMPLATE"] = 2004] = "MERCHANT_GENERIC_FAILED_TO_EXPAND_TEMPLATE";
        /**
         * The proposal is not known to the backend.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_ORDER_UNKNOWN"] = 2005] = "MERCHANT_GENERIC_ORDER_UNKNOWN";
        /**
         * The order provided to the backend could not be completed, because a product to be completed via inventory data is not actually in our inventory.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_PRODUCT_UNKNOWN"] = 2006] = "MERCHANT_GENERIC_PRODUCT_UNKNOWN";
        /**
         * The tip ID is unknown.  This could happen if the tip has expired.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_TIP_ID_UNKNOWN"] = 2007] = "MERCHANT_GENERIC_TIP_ID_UNKNOWN";
        /**
         * The contract obtained from the merchant backend was malformed.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_DB_CONTRACT_CONTENT_INVALID"] = 2008] = "MERCHANT_GENERIC_DB_CONTRACT_CONTENT_INVALID";
        /**
         * The order we found does not match the provided contract hash.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_CONTRACT_HASH_DOES_NOT_MATCH_ORDER"] = 2009] = "MERCHANT_GENERIC_CONTRACT_HASH_DOES_NOT_MATCH_ORDER";
        /**
         * The exchange failed to provide a valid response to the merchant's /keys request.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_GATEWAY (502).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_EXCHANGE_KEYS_FAILURE"] = 2010] = "MERCHANT_GENERIC_EXCHANGE_KEYS_FAILURE";
        /**
         * The exchange failed to respond to the merchant on time.
         * Returned with an HTTP status code of #MHD_HTTP_GATEWAY_TIMEOUT (504).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_EXCHANGE_TIMEOUT"] = 2011] = "MERCHANT_GENERIC_EXCHANGE_TIMEOUT";
        /**
         * The merchant failed to talk to the exchange.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_EXCHANGE_CONNECT_FAILURE"] = 2012] = "MERCHANT_GENERIC_EXCHANGE_CONNECT_FAILURE";
        /**
         * The exchange returned a maformed response.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_GATEWAY (502).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_EXCHANGE_REPLY_MALFORMED"] = 2013] = "MERCHANT_GENERIC_EXCHANGE_REPLY_MALFORMED";
        /**
         * The exchange returned an unexpected response status.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_GATEWAY (502).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GENERIC_EXCHANGE_UNEXPECTED_STATUS"] = 2014] = "MERCHANT_GENERIC_EXCHANGE_UNEXPECTED_STATUS";
        /**
         * The exchange failed to provide a valid answer to the tracking request, thus those details are not in the response.
         * Returned with an HTTP status code of #MHD_HTTP_OK (200).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GET_ORDERS_EXCHANGE_TRACKING_FAILURE"] = 2100] = "MERCHANT_GET_ORDERS_EXCHANGE_TRACKING_FAILURE";
        /**
         * The merchant backend failed to construct the request for tracking to the exchange, thus tracking details are not in the response.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GET_ORDERS_ID_EXCHANGE_REQUEST_FAILURE"] = 2103] = "MERCHANT_GET_ORDERS_ID_EXCHANGE_REQUEST_FAILURE";
        /**
         * The merchant backend failed trying to contact the exchange for tracking details, thus those details are not in the response.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GET_ORDERS_ID_EXCHANGE_LOOKUP_START_FAILURE"] = 2104] = "MERCHANT_GET_ORDERS_ID_EXCHANGE_LOOKUP_START_FAILURE";
        /**
         * The token used to authenticate the client is invalid for this order.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_GET_ORDERS_ID_INVALID_TOKEN"] = 2105] = "MERCHANT_GET_ORDERS_ID_INVALID_TOKEN";
        /**
         * The exchange responded saying that funds were insufficient (for example, due to double-spending).
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_INSUFFICIENT_FUNDS"] = 2150] = "MERCHANT_POST_ORDERS_ID_PAY_INSUFFICIENT_FUNDS";
        /**
         * The denomination key used for payment is not listed among the denomination keys of the exchange.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_DENOMINATION_KEY_NOT_FOUND"] = 2151] = "MERCHANT_POST_ORDERS_ID_PAY_DENOMINATION_KEY_NOT_FOUND";
        /**
         * The denomination key used for payment is not audited by an auditor approved by the merchant.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_DENOMINATION_KEY_AUDITOR_FAILURE"] = 2152] = "MERCHANT_POST_ORDERS_ID_PAY_DENOMINATION_KEY_AUDITOR_FAILURE";
        /**
         * There was an integer overflow totaling up the amounts or deposit fees in the payment.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_AMOUNT_OVERFLOW"] = 2153] = "MERCHANT_POST_ORDERS_ID_PAY_AMOUNT_OVERFLOW";
        /**
         * The deposit fees exceed the total value of the payment.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_FEES_EXCEED_PAYMENT"] = 2154] = "MERCHANT_POST_ORDERS_ID_PAY_FEES_EXCEED_PAYMENT";
        /**
         * After considering deposit and wire fees, the payment is insufficient to satisfy the required amount for the contract.  The client should revisit the logic used to calculate fees it must cover.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_ACCEPTABLE (406).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_INSUFFICIENT_DUE_TO_FEES"] = 2155] = "MERCHANT_POST_ORDERS_ID_PAY_INSUFFICIENT_DUE_TO_FEES";
        /**
         * Even if we do not consider deposit and wire fees, the payment is insufficient to satisfy the required amount for the contract.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_ACCEPTABLE (406).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_PAYMENT_INSUFFICIENT"] = 2156] = "MERCHANT_POST_ORDERS_ID_PAY_PAYMENT_INSUFFICIENT";
        /**
         * The signature over the contract of one of the coins was invalid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_COIN_SIGNATURE_INVALID"] = 2157] = "MERCHANT_POST_ORDERS_ID_PAY_COIN_SIGNATURE_INVALID";
        /**
         * When we tried to find information about the exchange to issue the deposit, we failed.  This usually only happens if the merchant backend is somehow unable to get its own HTTP client logic to work.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_EXCHANGE_LOOKUP_FAILED"] = 2158] = "MERCHANT_POST_ORDERS_ID_PAY_EXCHANGE_LOOKUP_FAILED";
        /**
         * The refund deadline in the contract is after the transfer deadline.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_REFUND_DEADLINE_PAST_WIRE_TRANSFER_DEADLINE"] = 2159] = "MERCHANT_POST_ORDERS_ID_PAY_REFUND_DEADLINE_PAST_WIRE_TRANSFER_DEADLINE";
        /**
         * The payment is too late, the offer has expired.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_OFFER_EXPIRED"] = 2161] = "MERCHANT_POST_ORDERS_ID_PAY_OFFER_EXPIRED";
        /**
         * The "merchant" field is missing in the proposal data. This is an internal error as the proposal is from the merchant's own database at this point.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_MERCHANT_FIELD_MISSING"] = 2162] = "MERCHANT_POST_ORDERS_ID_PAY_MERCHANT_FIELD_MISSING";
        /**
         * Failed to locate merchant's account information matching the wire hash given in the proposal.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_WIRE_HASH_UNKNOWN"] = 2163] = "MERCHANT_POST_ORDERS_ID_PAY_WIRE_HASH_UNKNOWN";
        /**
         * The deposit time for the denomination has expired.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_DENOMINATION_DEPOSIT_EXPIRED"] = 2165] = "MERCHANT_POST_ORDERS_ID_PAY_DENOMINATION_DEPOSIT_EXPIRED";
        /**
         * The exchange of the deposited coin charges a wire fee that could not be added to the total (total amount too high).
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_EXCHANGE_WIRE_FEE_ADDITION_FAILED"] = 2166] = "MERCHANT_POST_ORDERS_ID_PAY_EXCHANGE_WIRE_FEE_ADDITION_FAILED";
        /**
         * The contract was not fully paid because of refunds. Note that clients MAY treat this as paid if, for example, contracts must be executed despite of refunds.
         * Returned with an HTTP status code of #MHD_HTTP_PAYMENT_REQUIRED (402).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_REFUNDED"] = 2167] = "MERCHANT_POST_ORDERS_ID_PAY_REFUNDED";
        /**
         * According to our database, we have refunded more than we were paid (which should not be possible).
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_REFUNDS_EXCEED_PAYMENTS"] = 2168] = "MERCHANT_POST_ORDERS_ID_PAY_REFUNDS_EXCEED_PAYMENTS";
        /**
         * Legacy stuff. Remove me with protocol v1.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["DEAD_QQQ_PAY_MERCHANT_POST_ORDERS_ID_ABORT_REFUND_REFUSED_PAYMENT_COMPLETE"] = 2169] = "DEAD_QQQ_PAY_MERCHANT_POST_ORDERS_ID_ABORT_REFUND_REFUSED_PAYMENT_COMPLETE";
        /**
         * The payment failed at the exchange.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_GATEWAY (502).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAY_EXCHANGE_FAILED"] = 2170] = "MERCHANT_POST_ORDERS_ID_PAY_EXCHANGE_FAILED";
        /**
         * The contract hash does not match the given order ID.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAID_CONTRACT_HASH_MISMATCH"] = 2200] = "MERCHANT_POST_ORDERS_ID_PAID_CONTRACT_HASH_MISMATCH";
        /**
         * The signature of the merchant is not valid for the given contract hash.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_PAID_COIN_SIGNATURE_INVALID"] = 2201] = "MERCHANT_POST_ORDERS_ID_PAID_COIN_SIGNATURE_INVALID";
        /**
         * The merchant failed to send the exchange the refund request.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_ABORT_EXCHANGE_REFUND_FAILED"] = 2251] = "MERCHANT_POST_ORDERS_ID_ABORT_EXCHANGE_REFUND_FAILED";
        /**
         * The merchant failed to find the exchange to process the lookup.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_ABORT_EXCHANGE_LOOKUP_FAILED"] = 2252] = "MERCHANT_POST_ORDERS_ID_ABORT_EXCHANGE_LOOKUP_FAILED";
        /**
         * The merchant could not find the contract.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_ABORT_CONTRACT_NOT_FOUND"] = 2253] = "MERCHANT_POST_ORDERS_ID_ABORT_CONTRACT_NOT_FOUND";
        /**
         * The payment was already completed and thus cannot be aborted anymore.
         * Returned with an HTTP status code of #MHD_HTTP_PRECONDITION_FAILED (412).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_ABORT_REFUND_REFUSED_PAYMENT_COMPLETE"] = 2254] = "MERCHANT_POST_ORDERS_ID_ABORT_REFUND_REFUSED_PAYMENT_COMPLETE";
        /**
         * The hash provided by the wallet does not match the order.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_ABORT_CONTRACT_HASH_MISSMATCH"] = 2255] = "MERCHANT_POST_ORDERS_ID_ABORT_CONTRACT_HASH_MISSMATCH";
        /**
         * The array of coins cannot be empty.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_ABORT_COINS_ARRAY_EMPTY"] = 2256] = "MERCHANT_POST_ORDERS_ID_ABORT_COINS_ARRAY_EMPTY";
        /**
         * We could not claim the order because the backend is unaware of it.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_CLAIM_NOT_FOUND"] = 2300] = "MERCHANT_POST_ORDERS_ID_CLAIM_NOT_FOUND";
        /**
         * We could not claim the order because someone else claimed it first.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_CLAIM_ALREADY_CLAIMED"] = 2301] = "MERCHANT_POST_ORDERS_ID_CLAIM_ALREADY_CLAIMED";
        /**
         * The client-side experienced an internal failure.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_CLAIM_CLIENT_INTERNAL_FAILURE"] = 2302] = "MERCHANT_POST_ORDERS_ID_CLAIM_CLIENT_INTERNAL_FAILURE";
        /**
         * The backend failed to sign the refund request.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_POST_ORDERS_ID_REFUND_SIGNATURE_FAILED"] = 2350] = "MERCHANT_POST_ORDERS_ID_REFUND_SIGNATURE_FAILED";
        /**
         * The client failed to unblind the signature returned by the merchant.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_TIP_PICKUP_UNBLIND_FAILURE"] = 2400] = "MERCHANT_TIP_PICKUP_UNBLIND_FAILURE";
        /**
         * The exchange returned a failure code for the withdraw operation.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_GATEWAY (502).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_TIP_PICKUP_EXCHANGE_ERROR"] = 2403] = "MERCHANT_TIP_PICKUP_EXCHANGE_ERROR";
        /**
         * The merchant failed to add up the amounts to compute the pick up value.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_TIP_PICKUP_SUMMATION_FAILED"] = 2404] = "MERCHANT_TIP_PICKUP_SUMMATION_FAILED";
        /**
         * The tip expired.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_TIP_PICKUP_HAS_EXPIRED"] = 2405] = "MERCHANT_TIP_PICKUP_HAS_EXPIRED";
        /**
         * The requested withdraw amount exceeds the amount remaining to be picked up.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_TIP_PICKUP_AMOUNT_EXCEEDS_TIP_REMAINING"] = 2406] = "MERCHANT_TIP_PICKUP_AMOUNT_EXCEEDS_TIP_REMAINING";
        /**
         * The merchant did not find the specified denomination key in the exchange's key set.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_TIP_PICKUP_DENOMINATION_UNKNOWN"] = 2407] = "MERCHANT_TIP_PICKUP_DENOMINATION_UNKNOWN";
        /**
         * The backend lacks a wire transfer method configuration option for the given instance. Thus, this instance is unavailable (not findable for creating new orders).
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_INSTANCE_CONFIGURATION_LACKS_WIRE"] = 2500] = "MERCHANT_PRIVATE_POST_ORDERS_INSTANCE_CONFIGURATION_LACKS_WIRE";
        /**
         * The proposal had no timestamp and the backend failed to obtain the local time. Likely to be an internal error.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_NO_LOCALTIME"] = 2501] = "MERCHANT_PRIVATE_POST_ORDERS_NO_LOCALTIME";
        /**
         * The order provided to the backend could not be parsed, some required fields were missing or ill-formed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_PROPOSAL_PARSE_ERROR"] = 2502] = "MERCHANT_PRIVATE_POST_ORDERS_PROPOSAL_PARSE_ERROR";
        /**
         * The backend encountered an error: the proposal already exists.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_ALREADY_EXISTS"] = 2503] = "MERCHANT_PRIVATE_POST_ORDERS_ALREADY_EXISTS";
        /**
         * One of the paths to forget is malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_PATCH_ORDERS_ID_FORGET_PATH_SYNTAX_INCORRECT"] = 2510] = "MERCHANT_PRIVATE_PATCH_ORDERS_ID_FORGET_PATH_SYNTAX_INCORRECT";
        /**
         * One of the paths to forget was not marked as forgettable.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_PATCH_ORDERS_ID_FORGET_PATH_NOT_FORGETTABLE"] = 2511] = "MERCHANT_PRIVATE_PATCH_ORDERS_ID_FORGET_PATH_NOT_FORGETTABLE";
        /**
         * The order provided to the backend could not be deleted, our offer is still valid and awaiting payment.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_DELETE_ORDERS_AWAITING_PAYMENT"] = 2520] = "MERCHANT_PRIVATE_DELETE_ORDERS_AWAITING_PAYMENT";
        /**
         * The amount to be refunded is inconsistent: either is lower than the previous amount being awarded, or it is too big to be paid back. In this second case, the fault stays on the business dept. side.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_ID_REFUND_INCONSISTENT_AMOUNT"] = 2530] = "MERCHANT_PRIVATE_POST_ORDERS_ID_REFUND_INCONSISTENT_AMOUNT";
        /**
         * The frontend gave an unpaid order id to issue the refund to.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_ID_REFUND_ORDER_UNPAID"] = 2531] = "MERCHANT_PRIVATE_POST_ORDERS_ID_REFUND_ORDER_UNPAID";
        /**
         * The refund delay was set to 0 and thus no refunds are allowed for this order.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_ORDERS_ID_REFUND_NOT_ALLOWED_BY_CONTRACT"] = 2532] = "MERCHANT_PRIVATE_POST_ORDERS_ID_REFUND_NOT_ALLOWED_BY_CONTRACT";
        /**
         * We internally failed to execute the /track/transfer request.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TRANSFERS_REQUEST_ERROR"] = 2551] = "MERCHANT_PRIVATE_POST_TRANSFERS_REQUEST_ERROR";
        /**
         * The exchange gave conflicting information about a coin which has been wire transferred.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TRANSFERS_CONFLICTING_REPORTS"] = 2553] = "MERCHANT_PRIVATE_POST_TRANSFERS_CONFLICTING_REPORTS";
        /**
         * The exchange charged a different wire fee than what it originally advertised, and it is higher.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_GATEWAY (502).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TRANSFERS_BAD_WIRE_FEE"] = 2554] = "MERCHANT_PRIVATE_POST_TRANSFERS_BAD_WIRE_FEE";
        /**
         * We did not find the account that the transfer was made to.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TRANSFERS_ACCOUNT_NOT_FOUND"] = 2555] = "MERCHANT_PRIVATE_POST_TRANSFERS_ACCOUNT_NOT_FOUND";
        /**
         * The merchant backend cannot create an instance under the given identifier as one already exists. Use PATCH to modify the existing entry.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_INSTANCES_ALREADY_EXISTS"] = 2600] = "MERCHANT_PRIVATE_POST_INSTANCES_ALREADY_EXISTS";
        /**
         * The product ID exists.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_PRODUCTS_CONFLICT_PRODUCT_EXISTS"] = 2650] = "MERCHANT_PRIVATE_POST_PRODUCTS_CONFLICT_PRODUCT_EXISTS";
        /**
         * The update would have reduced the total amount of product lost, which is not allowed.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_PATCH_PRODUCTS_TOTAL_LOST_REDUCED"] = 2660] = "MERCHANT_PRIVATE_PATCH_PRODUCTS_TOTAL_LOST_REDUCED";
        /**
         * The update would have mean that more stocks were lost than what remains from total inventory after sales, which is not allowed.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_PATCH_PRODUCTS_TOTAL_LOST_EXCEEDS_STOCKS"] = 2661] = "MERCHANT_PRIVATE_PATCH_PRODUCTS_TOTAL_LOST_EXCEEDS_STOCKS";
        /**
         * The update would have reduced the total amount of product in stock, which is not allowed.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_PATCH_PRODUCTS_TOTAL_STOCKED_REDUCED"] = 2662] = "MERCHANT_PRIVATE_PATCH_PRODUCTS_TOTAL_STOCKED_REDUCED";
        /**
         * The lock request is for more products than we have left (unlocked) in stock.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_PRODUCTS_LOCK_INSUFFICIENT_STOCKS"] = 2670] = "MERCHANT_PRIVATE_POST_PRODUCTS_LOCK_INSUFFICIENT_STOCKS";
        /**
         * The deletion request is for a product that is locked.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_DELETE_PRODUCTS_CONFLICTING_LOCK"] = 2680] = "MERCHANT_PRIVATE_DELETE_PRODUCTS_CONFLICTING_LOCK";
        /**
         * The requested wire method is not supported by the exchange.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_RESERVES_UNSUPPORTED_WIRE_METHOD"] = 2700] = "MERCHANT_PRIVATE_POST_RESERVES_UNSUPPORTED_WIRE_METHOD";
        /**
         * The reserve could not be deleted because it is unknown.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_DELETE_RESERVES_NO_SUCH_RESERVE"] = 2710] = "MERCHANT_PRIVATE_DELETE_RESERVES_NO_SUCH_RESERVE";
        /**
         * The reserve that was used to fund the tips has expired.
         * Returned with an HTTP status code of #MHD_HTTP_GONE (410).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_EXPIRED"] = 2750] = "MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_EXPIRED";
        /**
         * The reserve that was used to fund the tips was not found in the DB.
         * Returned with an HTTP status code of #MHD_HTTP_SERVICE_UNAVAILABLE (503).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_UNKNOWN"] = 2751] = "MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_UNKNOWN";
        /**
         * The backend knows the instance that was supposed to support the tip, and it was configured for tipping. However, the funds remaining are insufficient to cover the tip, and the merchant should top up the reserve.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_INSUFFICIENT_FUNDS"] = 2752] = "MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_INSUFFICIENT_FUNDS";
        /**
         * The backend failed to find a reserve needed to authorize the tip.
         * Returned with an HTTP status code of #MHD_HTTP_SERVICE_UNAVAILABLE (503).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_NOT_FOUND"] = 2753] = "MERCHANT_PRIVATE_POST_TIP_AUTHORIZE_RESERVE_NOT_FOUND";
        /**
         * The merchant backend encountered a failure in computing the deposit total.
         * Returned with an HTTP status code of #MHD_HTTP_OK (200).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["MERCHANT_PRIVATE_GET_ORDERS_ID_AMOUNT_ARITHMETIC_FAILURE"] = 2800] = "MERCHANT_PRIVATE_GET_ORDERS_ID_AMOUNT_ARITHMETIC_FAILURE";
        /**
         * The signature from the exchange on the deposit confirmation is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["AUDITOR_DEPOSIT_CONFIRMATION_SIGNATURE_INVALID"] = 3100] = "AUDITOR_DEPOSIT_CONFIRMATION_SIGNATURE_INVALID";
        /**
         * Wire transfer attempted with credit and debit party being the same bank account.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_SAME_ACCOUNT"] = 5101] = "BANK_SAME_ACCOUNT";
        /**
         * Wire transfer impossible, due to financial limitation of the party that attempted the payment.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_UNALLOWED_DEBIT"] = 5102] = "BANK_UNALLOWED_DEBIT";
        /**
         * Negative number was used (as value and/or fraction) to initiate a Amount object.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_NEGATIVE_NUMBER_AMOUNT"] = 5103] = "BANK_NEGATIVE_NUMBER_AMOUNT";
        /**
         * A number too big was used (as value and/or fraction) to initiate a amount object.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_NUMBER_TOO_BIG"] = 5104] = "BANK_NUMBER_TOO_BIG";
        /**
         * Could not login for the requested operation.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_LOGIN_FAILED"] = 5105] = "BANK_LOGIN_FAILED";
        /**
         * The bank account referenced in the requested operation was not found. Returned along "400 Not found".
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_UNKNOWN_ACCOUNT"] = 5106] = "BANK_UNKNOWN_ACCOUNT";
        /**
         * The transaction referenced in the requested operation (typically a reject operation), was not found.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_TRANSACTION_NOT_FOUND"] = 5107] = "BANK_TRANSACTION_NOT_FOUND";
        /**
         * Bank received a malformed amount string.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_BAD_FORMAT_AMOUNT"] = 5108] = "BANK_BAD_FORMAT_AMOUNT";
        /**
         * The client does not own the account credited by the transaction which is to be rejected, so it has no rights do reject it.  To be returned along HTTP 403 Forbidden.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_REJECT_NO_RIGHTS"] = 5109] = "BANK_REJECT_NO_RIGHTS";
        /**
         * This error code is returned when no known exception types captured the exception, and comes along with a 500 Internal Server Error.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_UNMANAGED_EXCEPTION"] = 5110] = "BANK_UNMANAGED_EXCEPTION";
        /**
         * This error code is used for all those exceptions that do not really need a specific error code to return to the client, but need to signal the middleware that the bank is not responding with 500 Internal Server Error.  Used for example when a client is trying to register with a unavailable username.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_SOFT_EXCEPTION"] = 5111] = "BANK_SOFT_EXCEPTION";
        /**
         * The request UID for a request to transfer funds has already been used, but with different details for the transfer.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_TRANSFER_REQUEST_UID_REUSED"] = 5112] = "BANK_TRANSFER_REQUEST_UID_REUSED";
        /**
         * The withdrawal operation already has a reserve selected.  The current request conflicts with the existing selection.
         * Returned with an HTTP status code of #MHD_HTTP_CONFLICT (409).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["BANK_WITHDRAWAL_OPERATION_RESERVE_SELECTION_CONFLICT"] = 5113] = "BANK_WITHDRAWAL_OPERATION_RESERVE_SELECTION_CONFLICT";
        /**
         * The sync service failed find the account in its database.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_ACCOUNT_UNKNOWN"] = 6100] = "SYNC_ACCOUNT_UNKNOWN";
        /**
         * The SHA-512 hash provided in the If-None-Match header is malformed.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_BAD_IF_NONE_MATCH"] = 6101] = "SYNC_BAD_IF_NONE_MATCH";
        /**
         * The SHA-512 hash provided in the If-Match header is malformed or missing.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_BAD_IF_MATCH"] = 6102] = "SYNC_BAD_IF_MATCH";
        /**
         * The signature provided in the "Sync-Signature" header is malformed or missing.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_BAD_SYNC_SIGNATURE"] = 6103] = "SYNC_BAD_SYNC_SIGNATURE";
        /**
         * The signature provided in the "Sync-Signature" header does not match the account, old or new Etags.
         * Returned with an HTTP status code of #MHD_HTTP_FORBIDDEN (403).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_INVALID_SIGNATURE"] = 6104] = "SYNC_INVALID_SIGNATURE";
        /**
         * The "Content-length" field for the upload is not a number.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_MALFORMED_CONTENT_LENGTH"] = 6105] = "SYNC_MALFORMED_CONTENT_LENGTH";
        /**
         * The "Content-length" field for the upload is too big based on the server's terms of service.
         * Returned with an HTTP status code of #MHD_HTTP_PAYLOAD_TOO_LARGE (413).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_EXCESSIVE_CONTENT_LENGTH"] = 6106] = "SYNC_EXCESSIVE_CONTENT_LENGTH";
        /**
         * The server is out of memory to handle the upload. Trying again later may succeed.
         * Returned with an HTTP status code of #MHD_HTTP_PAYLOAD_TOO_LARGE (413).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_OUT_OF_MEMORY_ON_CONTENT_LENGTH"] = 6107] = "SYNC_OUT_OF_MEMORY_ON_CONTENT_LENGTH";
        /**
         * The uploaded data does not match the Etag.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_INVALID_UPLOAD"] = 6108] = "SYNC_INVALID_UPLOAD";
        /**
         * HTTP server experienced a timeout while awaiting promised payment.
         * Returned with an HTTP status code of #MHD_HTTP_REQUEST_TIMEOUT (408).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_PAYMENT_GENERIC_TIMEOUT"] = 6109] = "SYNC_PAYMENT_GENERIC_TIMEOUT";
        /**
         * Sync could not setup the payment request with its own backend.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_PAYMENT_CREATE_BACKEND_ERROR"] = 6110] = "SYNC_PAYMENT_CREATE_BACKEND_ERROR";
        /**
         * The sync service failed find the backup to be updated in its database.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_PREVIOUS_BACKUP_UNKNOWN"] = 6111] = "SYNC_PREVIOUS_BACKUP_UNKNOWN";
        /**
         * The "Content-length" field for the upload is missing.
         * Returned with an HTTP status code of #MHD_HTTP_BAD_REQUEST (400).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["SYNC_MISSING_CONTENT_LENGTH"] = 6112] = "SYNC_MISSING_CONTENT_LENGTH";
        /**
         * The wallet does not implement a version of the exchange protocol that is compatible with the protocol version of the exchange.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_IMPLEMENTED (501).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_EXCHANGE_PROTOCOL_VERSION_INCOMPATIBLE"] = 7000] = "WALLET_EXCHANGE_PROTOCOL_VERSION_INCOMPATIBLE";
        /**
         * The wallet encountered an unexpected exception.  This is likely a bug in the wallet implementation.
         * Returned with an HTTP status code of #MHD_HTTP_INTERNAL_SERVER_ERROR (500).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_UNEXPECTED_EXCEPTION"] = 7001] = "WALLET_UNEXPECTED_EXCEPTION";
        /**
         * The wallet received a response from a server, but the response can't be parsed.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_RECEIVED_MALFORMED_RESPONSE"] = 7002] = "WALLET_RECEIVED_MALFORMED_RESPONSE";
        /**
         * The wallet tried to make a network request, but it received no response.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_NETWORK_ERROR"] = 7003] = "WALLET_NETWORK_ERROR";
        /**
         * The wallet tried to make a network request, but it was throttled.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_HTTP_REQUEST_THROTTLED"] = 7004] = "WALLET_HTTP_REQUEST_THROTTLED";
        /**
         * The wallet made a request to a service, but received an error response it does not know how to handle.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_UNEXPECTED_REQUEST_ERROR"] = 7005] = "WALLET_UNEXPECTED_REQUEST_ERROR";
        /**
         * The denominations offered by the exchange are insufficient.  Likely the exchange is badly configured or not maintained.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_EXCHANGE_DENOMINATIONS_INSUFFICIENT"] = 7006] = "WALLET_EXCHANGE_DENOMINATIONS_INSUFFICIENT";
        /**
         * The wallet does not support the operation requested by a client.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_CORE_API_OPERATION_UNKNOWN"] = 7007] = "WALLET_CORE_API_OPERATION_UNKNOWN";
        /**
         * The given taler://pay URI is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_INVALID_TALER_PAY_URI"] = 7008] = "WALLET_INVALID_TALER_PAY_URI";
        /**
         * The signature on a coin by the exchange's denomination key is invalid after unblinding it.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_EXCHANGE_COIN_SIGNATURE_INVALID"] = 7009] = "WALLET_EXCHANGE_COIN_SIGNATURE_INVALID";
        /**
         * The exchange does not know about the reserve (yet), and thus withdrawal can't progress.
         * Returned with an HTTP status code of #MHD_HTTP_NOT_FOUND (404).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_EXCHANGE_WITHDRAW_RESERVE_UNKNOWN_AT_EXCHANGE"] = 7010] = "WALLET_EXCHANGE_WITHDRAW_RESERVE_UNKNOWN_AT_EXCHANGE";
        /**
         * The wallet core service is not available.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_CORE_NOT_AVAILABLE"] = 7011] = "WALLET_CORE_NOT_AVAILABLE";
        /**
         * The bank has aborted a withdrawal operation, and thus a withdrawal can't complete.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_WITHDRAWAL_OPERATION_ABORTED_BY_BANK"] = 7012] = "WALLET_WITHDRAWAL_OPERATION_ABORTED_BY_BANK";
        /**
         * An HTTP request made by the wallet timed out.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_HTTP_REQUEST_GENERIC_TIMEOUT"] = 7013] = "WALLET_HTTP_REQUEST_GENERIC_TIMEOUT";
        /**
         * The order has already been claimed by another wallet.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_ORDER_ALREADY_CLAIMED"] = 7014] = "WALLET_ORDER_ALREADY_CLAIMED";
        /**
         * A group of withdrawal operations (typically for the same reserve at the same exchange) has errors and will be tried again later.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_WITHDRAWAL_GROUP_INCOMPLETE"] = 7015] = "WALLET_WITHDRAWAL_GROUP_INCOMPLETE";
        /**
         * The signature on a coin by the exchange's denomination key (obtained through the merchant via tipping) is invalid after unblinding it.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_TIPPING_COIN_SIGNATURE_INVALID"] = 7016] = "WALLET_TIPPING_COIN_SIGNATURE_INVALID";
        /**
         * The wallet does not implement a version of the bank integration API that is compatible with the version offered by the bank.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_BANK_INTEGRATION_PROTOCOL_VERSION_INCOMPATIBLE"] = 7017] = "WALLET_BANK_INTEGRATION_PROTOCOL_VERSION_INCOMPATIBLE";
        /**
         * The wallet processed a taler://pay URI, but the merchant base URL in the downloaded contract terms does not match the merchant base URL derived from the URI.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_CONTRACT_TERMS_BASE_URL_MISMATCH"] = 7018] = "WALLET_CONTRACT_TERMS_BASE_URL_MISMATCH";
        /**
         * The merchant's signature on the contract terms is invalid.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["WALLET_CONTRACT_TERMS_SIGNATURE_INVALID"] = 7019] = "WALLET_CONTRACT_TERMS_SIGNATURE_INVALID";
        /**
         * The given action is invalid for the current state of the reducer.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["ANASTASIS_REDUCER_ACTION_INVALID"] = 8000] = "ANASTASIS_REDUCER_ACTION_INVALID";
        /**
         * End of error code range.
         * Returned with an HTTP status code of #MHD_HTTP_UNINITIALIZED (0).
         * (A value of 0 indicates that the error is generated client-side).
         */
        TalerErrorCode[TalerErrorCode["END"] = 9999] = "END";
    })(TalerErrorCode || (TalerErrorCode = {}));

    /*
     This file is part of GNU Taler
     (C) 2019 GNUnet e.V.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var NotificationType;
    (function (NotificationType) {
        NotificationType["CoinWithdrawn"] = "coin-withdrawn";
        NotificationType["ProposalAccepted"] = "proposal-accepted";
        NotificationType["ProposalDownloaded"] = "proposal-downloaded";
        NotificationType["RefundsSubmitted"] = "refunds-submitted";
        NotificationType["RecoupStarted"] = "recoup-started";
        NotificationType["RecoupFinished"] = "recoup-finished";
        NotificationType["RefreshRevealed"] = "refresh-revealed";
        NotificationType["RefreshMelted"] = "refresh-melted";
        NotificationType["RefreshStarted"] = "refresh-started";
        NotificationType["RefreshUnwarranted"] = "refresh-unwarranted";
        NotificationType["ReserveUpdated"] = "reserve-updated";
        NotificationType["ReserveConfirmed"] = "reserve-confirmed";
        NotificationType["ReserveCreated"] = "reserve-created";
        NotificationType["WithdrawGroupCreated"] = "withdraw-group-created";
        NotificationType["WithdrawGroupFinished"] = "withdraw-group-finished";
        NotificationType["WaitingForRetry"] = "waiting-for-retry";
        NotificationType["RefundStarted"] = "refund-started";
        NotificationType["RefundQueried"] = "refund-queried";
        NotificationType["RefundFinished"] = "refund-finished";
        NotificationType["ExchangeOperationError"] = "exchange-operation-error";
        NotificationType["RefreshOperationError"] = "refresh-operation-error";
        NotificationType["RecoupOperationError"] = "recoup-operation-error";
        NotificationType["RefundApplyOperationError"] = "refund-apply-error";
        NotificationType["RefundStatusOperationError"] = "refund-status-error";
        NotificationType["ProposalOperationError"] = "proposal-error";
        NotificationType["TipOperationError"] = "tip-error";
        NotificationType["PayOperationError"] = "pay-error";
        NotificationType["PayOperationSuccess"] = "pay-operation-success";
        NotificationType["WithdrawOperationError"] = "withdraw-error";
        NotificationType["ReserveNotYetFound"] = "reserve-not-yet-found";
        NotificationType["ReserveOperationError"] = "reserve-error";
        NotificationType["InternalError"] = "internal-error";
        NotificationType["PendingOperationProcessed"] = "pending-operation-processed";
        NotificationType["ProposalRefused"] = "proposal-refused";
        NotificationType["ReserveRegisteredWithBank"] = "reserve-registered-with-bank";
    })(NotificationType || (NotificationType = {}));

    // Ported in 2014 by Dmitry Chestnykh and Devi Mandiri.
    // TypeScript port in 2019 by Florian Dold.
    // Public domain.
    //
    // Implementation derived from TweetNaCl version 20140427.
    // See for details: http://tweetnacl.cr.yp.to/
    const gf = function (init = []) {
        const r = new Float64Array(16);
        if (init)
            for (let i = 0; i < init.length; i++)
                r[i] = init[i];
        return r;
    };
    //  Pluggable, initialized in high-level API below.
    let randombytes = function (x, n) {
        throw new Error("no PRNG");
    };
    const _9 = new Uint8Array(32);
    _9[0] = 9;
    // prettier-ignore
    const gf0 = gf();
    const gf1 = gf([1]);
    const _121665 = gf([0xdb41, 1]);
    const D = gf([
        0x78a3,
        0x1359,
        0x4dca,
        0x75eb,
        0xd8ab,
        0x4141,
        0x0a4d,
        0x0070,
        0xe898,
        0x7779,
        0x4079,
        0x8cc7,
        0xfe73,
        0x2b6f,
        0x6cee,
        0x5203,
    ]);
    const D2 = gf([
        0xf159,
        0x26b2,
        0x9b94,
        0xebd6,
        0xb156,
        0x8283,
        0x149a,
        0x00e0,
        0xd130,
        0xeef3,
        0x80f2,
        0x198e,
        0xfce7,
        0x56df,
        0xd9dc,
        0x2406,
    ]);
    const X = gf([
        0xd51a,
        0x8f25,
        0x2d60,
        0xc956,
        0xa7b2,
        0x9525,
        0xc760,
        0x692c,
        0xdc5c,
        0xfdd6,
        0xe231,
        0xc0a4,
        0x53fe,
        0xcd6e,
        0x36d3,
        0x2169,
    ]);
    const Y = gf([
        0x6658,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
        0x6666,
    ]);
    const I = gf([
        0xa0b0,
        0x4a0e,
        0x1b27,
        0xc4ee,
        0xe478,
        0xad2f,
        0x1806,
        0x2f43,
        0xd7a7,
        0x3dfb,
        0x0099,
        0x2b4d,
        0xdf0b,
        0x4fc1,
        0x2480,
        0x2b83,
    ]);
    function ts64(x, i, h, l) {
        x[i] = (h >> 24) & 0xff;
        x[i + 1] = (h >> 16) & 0xff;
        x[i + 2] = (h >> 8) & 0xff;
        x[i + 3] = h & 0xff;
        x[i + 4] = (l >> 24) & 0xff;
        x[i + 5] = (l >> 16) & 0xff;
        x[i + 6] = (l >> 8) & 0xff;
        x[i + 7] = l & 0xff;
    }
    function vn(x, xi, y, yi, n) {
        let i, d = 0;
        for (i = 0; i < n; i++)
            d |= x[xi + i] ^ y[yi + i];
        return (1 & ((d - 1) >>> 8)) - 1;
    }
    function crypto_verify_32(x, xi, y, yi) {
        return vn(x, xi, y, yi, 32);
    }
    function set25519(r, a) {
        let i;
        for (i = 0; i < 16; i++)
            r[i] = a[i] | 0;
    }
    function car25519(o) {
        let i, v, c = 1;
        for (i = 0; i < 16; i++) {
            v = o[i] + c + 65535;
            c = Math.floor(v / 65536);
            o[i] = v - c * 65536;
        }
        o[0] += c - 1 + 37 * (c - 1);
    }
    function sel25519(p, q, b) {
        let t;
        const c = ~(b - 1);
        for (let i = 0; i < 16; i++) {
            t = c & (p[i] ^ q[i]);
            p[i] ^= t;
            q[i] ^= t;
        }
    }
    function pack25519(o, n) {
        let i, j, b;
        const m = gf(), t = gf();
        for (i = 0; i < 16; i++)
            t[i] = n[i];
        car25519(t);
        car25519(t);
        car25519(t);
        for (j = 0; j < 2; j++) {
            m[0] = t[0] - 0xffed;
            for (i = 1; i < 15; i++) {
                m[i] = t[i] - 0xffff - ((m[i - 1] >> 16) & 1);
                m[i - 1] &= 0xffff;
            }
            m[15] = t[15] - 0x7fff - ((m[14] >> 16) & 1);
            b = (m[15] >> 16) & 1;
            m[14] &= 0xffff;
            sel25519(t, m, 1 - b);
        }
        for (i = 0; i < 16; i++) {
            o[2 * i] = t[i] & 0xff;
            o[2 * i + 1] = t[i] >> 8;
        }
    }
    function neq25519(a, b) {
        const c = new Uint8Array(32), d = new Uint8Array(32);
        pack25519(c, a);
        pack25519(d, b);
        return crypto_verify_32(c, 0, d, 0);
    }
    function par25519(a) {
        const d = new Uint8Array(32);
        pack25519(d, a);
        return d[0] & 1;
    }
    function unpack25519(o, n) {
        let i;
        for (i = 0; i < 16; i++)
            o[i] = n[2 * i] + (n[2 * i + 1] << 8);
        o[15] &= 0x7fff;
    }
    function A(o, a, b) {
        for (let i = 0; i < 16; i++)
            o[i] = a[i] + b[i];
    }
    function Z(o, a, b) {
        for (let i = 0; i < 16; i++)
            o[i] = a[i] - b[i];
    }
    function M(o, a, b) {
        let v, c, t0 = 0, t1 = 0, t2 = 0, t3 = 0, t4 = 0, t5 = 0, t6 = 0, t7 = 0, t8 = 0, t9 = 0, t10 = 0, t11 = 0, t12 = 0, t13 = 0, t14 = 0, t15 = 0, t16 = 0, t17 = 0, t18 = 0, t19 = 0, t20 = 0, t21 = 0, t22 = 0, t23 = 0, t24 = 0, t25 = 0, t26 = 0, t27 = 0, t28 = 0, t29 = 0, t30 = 0;
        const b0 = b[0], b1 = b[1], b2 = b[2], b3 = b[3], b4 = b[4], b5 = b[5], b6 = b[6], b7 = b[7], b8 = b[8], b9 = b[9], b10 = b[10], b11 = b[11], b12 = b[12], b13 = b[13], b14 = b[14], b15 = b[15];
        v = a[0];
        t0 += v * b0;
        t1 += v * b1;
        t2 += v * b2;
        t3 += v * b3;
        t4 += v * b4;
        t5 += v * b5;
        t6 += v * b6;
        t7 += v * b7;
        t8 += v * b8;
        t9 += v * b9;
        t10 += v * b10;
        t11 += v * b11;
        t12 += v * b12;
        t13 += v * b13;
        t14 += v * b14;
        t15 += v * b15;
        v = a[1];
        t1 += v * b0;
        t2 += v * b1;
        t3 += v * b2;
        t4 += v * b3;
        t5 += v * b4;
        t6 += v * b5;
        t7 += v * b6;
        t8 += v * b7;
        t9 += v * b8;
        t10 += v * b9;
        t11 += v * b10;
        t12 += v * b11;
        t13 += v * b12;
        t14 += v * b13;
        t15 += v * b14;
        t16 += v * b15;
        v = a[2];
        t2 += v * b0;
        t3 += v * b1;
        t4 += v * b2;
        t5 += v * b3;
        t6 += v * b4;
        t7 += v * b5;
        t8 += v * b6;
        t9 += v * b7;
        t10 += v * b8;
        t11 += v * b9;
        t12 += v * b10;
        t13 += v * b11;
        t14 += v * b12;
        t15 += v * b13;
        t16 += v * b14;
        t17 += v * b15;
        v = a[3];
        t3 += v * b0;
        t4 += v * b1;
        t5 += v * b2;
        t6 += v * b3;
        t7 += v * b4;
        t8 += v * b5;
        t9 += v * b6;
        t10 += v * b7;
        t11 += v * b8;
        t12 += v * b9;
        t13 += v * b10;
        t14 += v * b11;
        t15 += v * b12;
        t16 += v * b13;
        t17 += v * b14;
        t18 += v * b15;
        v = a[4];
        t4 += v * b0;
        t5 += v * b1;
        t6 += v * b2;
        t7 += v * b3;
        t8 += v * b4;
        t9 += v * b5;
        t10 += v * b6;
        t11 += v * b7;
        t12 += v * b8;
        t13 += v * b9;
        t14 += v * b10;
        t15 += v * b11;
        t16 += v * b12;
        t17 += v * b13;
        t18 += v * b14;
        t19 += v * b15;
        v = a[5];
        t5 += v * b0;
        t6 += v * b1;
        t7 += v * b2;
        t8 += v * b3;
        t9 += v * b4;
        t10 += v * b5;
        t11 += v * b6;
        t12 += v * b7;
        t13 += v * b8;
        t14 += v * b9;
        t15 += v * b10;
        t16 += v * b11;
        t17 += v * b12;
        t18 += v * b13;
        t19 += v * b14;
        t20 += v * b15;
        v = a[6];
        t6 += v * b0;
        t7 += v * b1;
        t8 += v * b2;
        t9 += v * b3;
        t10 += v * b4;
        t11 += v * b5;
        t12 += v * b6;
        t13 += v * b7;
        t14 += v * b8;
        t15 += v * b9;
        t16 += v * b10;
        t17 += v * b11;
        t18 += v * b12;
        t19 += v * b13;
        t20 += v * b14;
        t21 += v * b15;
        v = a[7];
        t7 += v * b0;
        t8 += v * b1;
        t9 += v * b2;
        t10 += v * b3;
        t11 += v * b4;
        t12 += v * b5;
        t13 += v * b6;
        t14 += v * b7;
        t15 += v * b8;
        t16 += v * b9;
        t17 += v * b10;
        t18 += v * b11;
        t19 += v * b12;
        t20 += v * b13;
        t21 += v * b14;
        t22 += v * b15;
        v = a[8];
        t8 += v * b0;
        t9 += v * b1;
        t10 += v * b2;
        t11 += v * b3;
        t12 += v * b4;
        t13 += v * b5;
        t14 += v * b6;
        t15 += v * b7;
        t16 += v * b8;
        t17 += v * b9;
        t18 += v * b10;
        t19 += v * b11;
        t20 += v * b12;
        t21 += v * b13;
        t22 += v * b14;
        t23 += v * b15;
        v = a[9];
        t9 += v * b0;
        t10 += v * b1;
        t11 += v * b2;
        t12 += v * b3;
        t13 += v * b4;
        t14 += v * b5;
        t15 += v * b6;
        t16 += v * b7;
        t17 += v * b8;
        t18 += v * b9;
        t19 += v * b10;
        t20 += v * b11;
        t21 += v * b12;
        t22 += v * b13;
        t23 += v * b14;
        t24 += v * b15;
        v = a[10];
        t10 += v * b0;
        t11 += v * b1;
        t12 += v * b2;
        t13 += v * b3;
        t14 += v * b4;
        t15 += v * b5;
        t16 += v * b6;
        t17 += v * b7;
        t18 += v * b8;
        t19 += v * b9;
        t20 += v * b10;
        t21 += v * b11;
        t22 += v * b12;
        t23 += v * b13;
        t24 += v * b14;
        t25 += v * b15;
        v = a[11];
        t11 += v * b0;
        t12 += v * b1;
        t13 += v * b2;
        t14 += v * b3;
        t15 += v * b4;
        t16 += v * b5;
        t17 += v * b6;
        t18 += v * b7;
        t19 += v * b8;
        t20 += v * b9;
        t21 += v * b10;
        t22 += v * b11;
        t23 += v * b12;
        t24 += v * b13;
        t25 += v * b14;
        t26 += v * b15;
        v = a[12];
        t12 += v * b0;
        t13 += v * b1;
        t14 += v * b2;
        t15 += v * b3;
        t16 += v * b4;
        t17 += v * b5;
        t18 += v * b6;
        t19 += v * b7;
        t20 += v * b8;
        t21 += v * b9;
        t22 += v * b10;
        t23 += v * b11;
        t24 += v * b12;
        t25 += v * b13;
        t26 += v * b14;
        t27 += v * b15;
        v = a[13];
        t13 += v * b0;
        t14 += v * b1;
        t15 += v * b2;
        t16 += v * b3;
        t17 += v * b4;
        t18 += v * b5;
        t19 += v * b6;
        t20 += v * b7;
        t21 += v * b8;
        t22 += v * b9;
        t23 += v * b10;
        t24 += v * b11;
        t25 += v * b12;
        t26 += v * b13;
        t27 += v * b14;
        t28 += v * b15;
        v = a[14];
        t14 += v * b0;
        t15 += v * b1;
        t16 += v * b2;
        t17 += v * b3;
        t18 += v * b4;
        t19 += v * b5;
        t20 += v * b6;
        t21 += v * b7;
        t22 += v * b8;
        t23 += v * b9;
        t24 += v * b10;
        t25 += v * b11;
        t26 += v * b12;
        t27 += v * b13;
        t28 += v * b14;
        t29 += v * b15;
        v = a[15];
        t15 += v * b0;
        t16 += v * b1;
        t17 += v * b2;
        t18 += v * b3;
        t19 += v * b4;
        t20 += v * b5;
        t21 += v * b6;
        t22 += v * b7;
        t23 += v * b8;
        t24 += v * b9;
        t25 += v * b10;
        t26 += v * b11;
        t27 += v * b12;
        t28 += v * b13;
        t29 += v * b14;
        t30 += v * b15;
        t0 += 38 * t16;
        t1 += 38 * t17;
        t2 += 38 * t18;
        t3 += 38 * t19;
        t4 += 38 * t20;
        t5 += 38 * t21;
        t6 += 38 * t22;
        t7 += 38 * t23;
        t8 += 38 * t24;
        t9 += 38 * t25;
        t10 += 38 * t26;
        t11 += 38 * t27;
        t12 += 38 * t28;
        t13 += 38 * t29;
        t14 += 38 * t30;
        // t15 left as is
        // first car
        c = 1;
        v = t0 + c + 65535;
        c = Math.floor(v / 65536);
        t0 = v - c * 65536;
        v = t1 + c + 65535;
        c = Math.floor(v / 65536);
        t1 = v - c * 65536;
        v = t2 + c + 65535;
        c = Math.floor(v / 65536);
        t2 = v - c * 65536;
        v = t3 + c + 65535;
        c = Math.floor(v / 65536);
        t3 = v - c * 65536;
        v = t4 + c + 65535;
        c = Math.floor(v / 65536);
        t4 = v - c * 65536;
        v = t5 + c + 65535;
        c = Math.floor(v / 65536);
        t5 = v - c * 65536;
        v = t6 + c + 65535;
        c = Math.floor(v / 65536);
        t6 = v - c * 65536;
        v = t7 + c + 65535;
        c = Math.floor(v / 65536);
        t7 = v - c * 65536;
        v = t8 + c + 65535;
        c = Math.floor(v / 65536);
        t8 = v - c * 65536;
        v = t9 + c + 65535;
        c = Math.floor(v / 65536);
        t9 = v - c * 65536;
        v = t10 + c + 65535;
        c = Math.floor(v / 65536);
        t10 = v - c * 65536;
        v = t11 + c + 65535;
        c = Math.floor(v / 65536);
        t11 = v - c * 65536;
        v = t12 + c + 65535;
        c = Math.floor(v / 65536);
        t12 = v - c * 65536;
        v = t13 + c + 65535;
        c = Math.floor(v / 65536);
        t13 = v - c * 65536;
        v = t14 + c + 65535;
        c = Math.floor(v / 65536);
        t14 = v - c * 65536;
        v = t15 + c + 65535;
        c = Math.floor(v / 65536);
        t15 = v - c * 65536;
        t0 += c - 1 + 37 * (c - 1);
        // second car
        c = 1;
        v = t0 + c + 65535;
        c = Math.floor(v / 65536);
        t0 = v - c * 65536;
        v = t1 + c + 65535;
        c = Math.floor(v / 65536);
        t1 = v - c * 65536;
        v = t2 + c + 65535;
        c = Math.floor(v / 65536);
        t2 = v - c * 65536;
        v = t3 + c + 65535;
        c = Math.floor(v / 65536);
        t3 = v - c * 65536;
        v = t4 + c + 65535;
        c = Math.floor(v / 65536);
        t4 = v - c * 65536;
        v = t5 + c + 65535;
        c = Math.floor(v / 65536);
        t5 = v - c * 65536;
        v = t6 + c + 65535;
        c = Math.floor(v / 65536);
        t6 = v - c * 65536;
        v = t7 + c + 65535;
        c = Math.floor(v / 65536);
        t7 = v - c * 65536;
        v = t8 + c + 65535;
        c = Math.floor(v / 65536);
        t8 = v - c * 65536;
        v = t9 + c + 65535;
        c = Math.floor(v / 65536);
        t9 = v - c * 65536;
        v = t10 + c + 65535;
        c = Math.floor(v / 65536);
        t10 = v - c * 65536;
        v = t11 + c + 65535;
        c = Math.floor(v / 65536);
        t11 = v - c * 65536;
        v = t12 + c + 65535;
        c = Math.floor(v / 65536);
        t12 = v - c * 65536;
        v = t13 + c + 65535;
        c = Math.floor(v / 65536);
        t13 = v - c * 65536;
        v = t14 + c + 65535;
        c = Math.floor(v / 65536);
        t14 = v - c * 65536;
        v = t15 + c + 65535;
        c = Math.floor(v / 65536);
        t15 = v - c * 65536;
        t0 += c - 1 + 37 * (c - 1);
        o[0] = t0;
        o[1] = t1;
        o[2] = t2;
        o[3] = t3;
        o[4] = t4;
        o[5] = t5;
        o[6] = t6;
        o[7] = t7;
        o[8] = t8;
        o[9] = t9;
        o[10] = t10;
        o[11] = t11;
        o[12] = t12;
        o[13] = t13;
        o[14] = t14;
        o[15] = t15;
    }
    function S(o, a) {
        M(o, a, a);
    }
    function inv25519(o, i) {
        const c = gf();
        let a;
        for (a = 0; a < 16; a++)
            c[a] = i[a];
        for (a = 253; a >= 0; a--) {
            S(c, c);
            if (a !== 2 && a !== 4)
                M(c, c, i);
        }
        for (a = 0; a < 16; a++)
            o[a] = c[a];
    }
    function pow2523(o, i) {
        const c = gf();
        let a;
        for (a = 0; a < 16; a++)
            c[a] = i[a];
        for (a = 250; a >= 0; a--) {
            S(c, c);
            if (a !== 1)
                M(c, c, i);
        }
        for (a = 0; a < 16; a++)
            o[a] = c[a];
    }
    function crypto_scalarmult(q, n, p) {
        const z = new Uint8Array(32);
        const x = new Float64Array(80);
        let r;
        let i;
        const a = gf(), b = gf(), c = gf(), d = gf(), e = gf(), f = gf();
        for (i = 0; i < 31; i++)
            z[i] = n[i];
        z[31] = (n[31] & 127) | 64;
        z[0] &= 248;
        unpack25519(x, p);
        for (i = 0; i < 16; i++) {
            b[i] = x[i];
            d[i] = a[i] = c[i] = 0;
        }
        a[0] = d[0] = 1;
        for (i = 254; i >= 0; --i) {
            r = (z[i >>> 3] >>> (i & 7)) & 1;
            sel25519(a, b, r);
            sel25519(c, d, r);
            A(e, a, c);
            Z(a, a, c);
            A(c, b, d);
            Z(b, b, d);
            S(d, e);
            S(f, a);
            M(a, c, a);
            M(c, b, e);
            A(e, a, c);
            Z(a, a, c);
            S(b, a);
            Z(c, d, f);
            M(a, c, _121665);
            A(a, a, d);
            M(c, c, a);
            M(a, d, f);
            M(d, b, x);
            S(b, e);
            sel25519(a, b, r);
            sel25519(c, d, r);
        }
        for (i = 0; i < 16; i++) {
            x[i + 16] = a[i];
            x[i + 32] = c[i];
            x[i + 48] = b[i];
            x[i + 64] = d[i];
        }
        const x32 = x.subarray(32);
        const x16 = x.subarray(16);
        inv25519(x32, x32);
        M(x16, x16, x32);
        pack25519(q, x16);
        return 0;
    }
    function crypto_scalarmult_base(q, n) {
        return crypto_scalarmult(q, n, _9);
    }
    // prettier-ignore
    const K = [
        0x428a2f98, 0xd728ae22, 0x71374491, 0x23ef65cd,
        0xb5c0fbcf, 0xec4d3b2f, 0xe9b5dba5, 0x8189dbbc,
        0x3956c25b, 0xf348b538, 0x59f111f1, 0xb605d019,
        0x923f82a4, 0xaf194f9b, 0xab1c5ed5, 0xda6d8118,
        0xd807aa98, 0xa3030242, 0x12835b01, 0x45706fbe,
        0x243185be, 0x4ee4b28c, 0x550c7dc3, 0xd5ffb4e2,
        0x72be5d74, 0xf27b896f, 0x80deb1fe, 0x3b1696b1,
        0x9bdc06a7, 0x25c71235, 0xc19bf174, 0xcf692694,
        0xe49b69c1, 0x9ef14ad2, 0xefbe4786, 0x384f25e3,
        0x0fc19dc6, 0x8b8cd5b5, 0x240ca1cc, 0x77ac9c65,
        0x2de92c6f, 0x592b0275, 0x4a7484aa, 0x6ea6e483,
        0x5cb0a9dc, 0xbd41fbd4, 0x76f988da, 0x831153b5,
        0x983e5152, 0xee66dfab, 0xa831c66d, 0x2db43210,
        0xb00327c8, 0x98fb213f, 0xbf597fc7, 0xbeef0ee4,
        0xc6e00bf3, 0x3da88fc2, 0xd5a79147, 0x930aa725,
        0x06ca6351, 0xe003826f, 0x14292967, 0x0a0e6e70,
        0x27b70a85, 0x46d22ffc, 0x2e1b2138, 0x5c26c926,
        0x4d2c6dfc, 0x5ac42aed, 0x53380d13, 0x9d95b3df,
        0x650a7354, 0x8baf63de, 0x766a0abb, 0x3c77b2a8,
        0x81c2c92e, 0x47edaee6, 0x92722c85, 0x1482353b,
        0xa2bfe8a1, 0x4cf10364, 0xa81a664b, 0xbc423001,
        0xc24b8b70, 0xd0f89791, 0xc76c51a3, 0x0654be30,
        0xd192e819, 0xd6ef5218, 0xd6990624, 0x5565a910,
        0xf40e3585, 0x5771202a, 0x106aa070, 0x32bbd1b8,
        0x19a4c116, 0xb8d2d0c8, 0x1e376c08, 0x5141ab53,
        0x2748774c, 0xdf8eeb99, 0x34b0bcb5, 0xe19b48a8,
        0x391c0cb3, 0xc5c95a63, 0x4ed8aa4a, 0xe3418acb,
        0x5b9cca4f, 0x7763e373, 0x682e6ff3, 0xd6b2b8a3,
        0x748f82ee, 0x5defb2fc, 0x78a5636f, 0x43172f60,
        0x84c87814, 0xa1f0ab72, 0x8cc70208, 0x1a6439ec,
        0x90befffa, 0x23631e28, 0xa4506ceb, 0xde82bde9,
        0xbef9a3f7, 0xb2c67915, 0xc67178f2, 0xe372532b,
        0xca273ece, 0xea26619c, 0xd186b8c7, 0x21c0c207,
        0xeada7dd6, 0xcde0eb1e, 0xf57d4f7f, 0xee6ed178,
        0x06f067aa, 0x72176fba, 0x0a637dc5, 0xa2c898a6,
        0x113f9804, 0xbef90dae, 0x1b710b35, 0x131c471b,
        0x28db77f5, 0x23047d84, 0x32caab7b, 0x40c72493,
        0x3c9ebe0a, 0x15c9bebc, 0x431d67c4, 0x9c100d4c,
        0x4cc5d4be, 0xcb3e42b6, 0x597f299c, 0xfc657e2a,
        0x5fcb6fab, 0x3ad6faec, 0x6c44198c, 0x4a475817
    ];
    function crypto_hashblocks_hl(hh, hl, m, n) {
        const wh = new Int32Array(16), wl = new Int32Array(16);
        let bh0, bh1, bh2, bh3, bh4, bh5, bh6, bh7, bl0, bl1, bl2, bl3, bl4, bl5, bl6, bl7, th, tl, i, j, h, l, a, b, c, d;
        let ah0 = hh[0], ah1 = hh[1], ah2 = hh[2], ah3 = hh[3], ah4 = hh[4], ah5 = hh[5], ah6 = hh[6], ah7 = hh[7], al0 = hl[0], al1 = hl[1], al2 = hl[2], al3 = hl[3], al4 = hl[4], al5 = hl[5], al6 = hl[6], al7 = hl[7];
        let pos = 0;
        while (n >= 128) {
            for (i = 0; i < 16; i++) {
                j = 8 * i + pos;
                wh[i] = (m[j + 0] << 24) | (m[j + 1] << 16) | (m[j + 2] << 8) | m[j + 3];
                wl[i] = (m[j + 4] << 24) | (m[j + 5] << 16) | (m[j + 6] << 8) | m[j + 7];
            }
            for (i = 0; i < 80; i++) {
                bh0 = ah0;
                bh1 = ah1;
                bh2 = ah2;
                bh3 = ah3;
                bh4 = ah4;
                bh5 = ah5;
                bh6 = ah6;
                bh7 = ah7;
                bl0 = al0;
                bl1 = al1;
                bl2 = al2;
                bl3 = al3;
                bl4 = al4;
                bl5 = al5;
                bl6 = al6;
                bl7 = al7;
                // add
                h = ah7;
                l = al7;
                a = l & 0xffff;
                b = l >>> 16;
                c = h & 0xffff;
                d = h >>> 16;
                // Sigma1
                h =
                    ((ah4 >>> 14) | (al4 << (32 - 14))) ^
                        ((ah4 >>> 18) | (al4 << (32 - 18))) ^
                        ((al4 >>> (41 - 32)) | (ah4 << (32 - (41 - 32))));
                l =
                    ((al4 >>> 14) | (ah4 << (32 - 14))) ^
                        ((al4 >>> 18) | (ah4 << (32 - 18))) ^
                        ((ah4 >>> (41 - 32)) | (al4 << (32 - (41 - 32))));
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                // Ch
                h = (ah4 & ah5) ^ (~ah4 & ah6);
                l = (al4 & al5) ^ (~al4 & al6);
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                // K
                h = K[i * 2];
                l = K[i * 2 + 1];
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                // w
                h = wh[i % 16];
                l = wl[i % 16];
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                b += a >>> 16;
                c += b >>> 16;
                d += c >>> 16;
                th = (c & 0xffff) | (d << 16);
                tl = (a & 0xffff) | (b << 16);
                // add
                h = th;
                l = tl;
                a = l & 0xffff;
                b = l >>> 16;
                c = h & 0xffff;
                d = h >>> 16;
                // Sigma0
                h =
                    ((ah0 >>> 28) | (al0 << (32 - 28))) ^
                        ((al0 >>> (34 - 32)) | (ah0 << (32 - (34 - 32)))) ^
                        ((al0 >>> (39 - 32)) | (ah0 << (32 - (39 - 32))));
                l =
                    ((al0 >>> 28) | (ah0 << (32 - 28))) ^
                        ((ah0 >>> (34 - 32)) | (al0 << (32 - (34 - 32)))) ^
                        ((ah0 >>> (39 - 32)) | (al0 << (32 - (39 - 32))));
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                // Maj
                h = (ah0 & ah1) ^ (ah0 & ah2) ^ (ah1 & ah2);
                l = (al0 & al1) ^ (al0 & al2) ^ (al1 & al2);
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                b += a >>> 16;
                c += b >>> 16;
                d += c >>> 16;
                bh7 = (c & 0xffff) | (d << 16);
                bl7 = (a & 0xffff) | (b << 16);
                // add
                h = bh3;
                l = bl3;
                a = l & 0xffff;
                b = l >>> 16;
                c = h & 0xffff;
                d = h >>> 16;
                h = th;
                l = tl;
                a += l & 0xffff;
                b += l >>> 16;
                c += h & 0xffff;
                d += h >>> 16;
                b += a >>> 16;
                c += b >>> 16;
                d += c >>> 16;
                bh3 = (c & 0xffff) | (d << 16);
                bl3 = (a & 0xffff) | (b << 16);
                ah1 = bh0;
                ah2 = bh1;
                ah3 = bh2;
                ah4 = bh3;
                ah5 = bh4;
                ah6 = bh5;
                ah7 = bh6;
                ah0 = bh7;
                al1 = bl0;
                al2 = bl1;
                al3 = bl2;
                al4 = bl3;
                al5 = bl4;
                al6 = bl5;
                al7 = bl6;
                al0 = bl7;
                if (i % 16 === 15) {
                    for (j = 0; j < 16; j++) {
                        // add
                        h = wh[j];
                        l = wl[j];
                        a = l & 0xffff;
                        b = l >>> 16;
                        c = h & 0xffff;
                        d = h >>> 16;
                        h = wh[(j + 9) % 16];
                        l = wl[(j + 9) % 16];
                        a += l & 0xffff;
                        b += l >>> 16;
                        c += h & 0xffff;
                        d += h >>> 16;
                        // sigma0
                        th = wh[(j + 1) % 16];
                        tl = wl[(j + 1) % 16];
                        h =
                            ((th >>> 1) | (tl << (32 - 1))) ^
                                ((th >>> 8) | (tl << (32 - 8))) ^
                                (th >>> 7);
                        l =
                            ((tl >>> 1) | (th << (32 - 1))) ^
                                ((tl >>> 8) | (th << (32 - 8))) ^
                                ((tl >>> 7) | (th << (32 - 7)));
                        a += l & 0xffff;
                        b += l >>> 16;
                        c += h & 0xffff;
                        d += h >>> 16;
                        // sigma1
                        th = wh[(j + 14) % 16];
                        tl = wl[(j + 14) % 16];
                        h =
                            ((th >>> 19) | (tl << (32 - 19))) ^
                                ((tl >>> (61 - 32)) | (th << (32 - (61 - 32)))) ^
                                (th >>> 6);
                        l =
                            ((tl >>> 19) | (th << (32 - 19))) ^
                                ((th >>> (61 - 32)) | (tl << (32 - (61 - 32)))) ^
                                ((tl >>> 6) | (th << (32 - 6)));
                        a += l & 0xffff;
                        b += l >>> 16;
                        c += h & 0xffff;
                        d += h >>> 16;
                        b += a >>> 16;
                        c += b >>> 16;
                        d += c >>> 16;
                        wh[j] = (c & 0xffff) | (d << 16);
                        wl[j] = (a & 0xffff) | (b << 16);
                    }
                }
            }
            // add
            h = ah0;
            l = al0;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[0];
            l = hl[0];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[0] = ah0 = (c & 0xffff) | (d << 16);
            hl[0] = al0 = (a & 0xffff) | (b << 16);
            h = ah1;
            l = al1;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[1];
            l = hl[1];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[1] = ah1 = (c & 0xffff) | (d << 16);
            hl[1] = al1 = (a & 0xffff) | (b << 16);
            h = ah2;
            l = al2;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[2];
            l = hl[2];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[2] = ah2 = (c & 0xffff) | (d << 16);
            hl[2] = al2 = (a & 0xffff) | (b << 16);
            h = ah3;
            l = al3;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[3];
            l = hl[3];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[3] = ah3 = (c & 0xffff) | (d << 16);
            hl[3] = al3 = (a & 0xffff) | (b << 16);
            h = ah4;
            l = al4;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[4];
            l = hl[4];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[4] = ah4 = (c & 0xffff) | (d << 16);
            hl[4] = al4 = (a & 0xffff) | (b << 16);
            h = ah5;
            l = al5;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[5];
            l = hl[5];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[5] = ah5 = (c & 0xffff) | (d << 16);
            hl[5] = al5 = (a & 0xffff) | (b << 16);
            h = ah6;
            l = al6;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[6];
            l = hl[6];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[6] = ah6 = (c & 0xffff) | (d << 16);
            hl[6] = al6 = (a & 0xffff) | (b << 16);
            h = ah7;
            l = al7;
            a = l & 0xffff;
            b = l >>> 16;
            c = h & 0xffff;
            d = h >>> 16;
            h = hh[7];
            l = hl[7];
            a += l & 0xffff;
            b += l >>> 16;
            c += h & 0xffff;
            d += h >>> 16;
            b += a >>> 16;
            c += b >>> 16;
            d += c >>> 16;
            hh[7] = ah7 = (c & 0xffff) | (d << 16);
            hl[7] = al7 = (a & 0xffff) | (b << 16);
            pos += 128;
            n -= 128;
        }
        return n;
    }
    function crypto_hash(out, m, n) {
        const hh = new Int32Array(8);
        const hl = new Int32Array(8);
        const x = new Uint8Array(256);
        const b = n;
        hh[0] = 0x6a09e667;
        hh[1] = 0xbb67ae85;
        hh[2] = 0x3c6ef372;
        hh[3] = 0xa54ff53a;
        hh[4] = 0x510e527f;
        hh[5] = 0x9b05688c;
        hh[6] = 0x1f83d9ab;
        hh[7] = 0x5be0cd19;
        hl[0] = 0xf3bcc908;
        hl[1] = 0x84caa73b;
        hl[2] = 0xfe94f82b;
        hl[3] = 0x5f1d36f1;
        hl[4] = 0xade682d1;
        hl[5] = 0x2b3e6c1f;
        hl[6] = 0xfb41bd6b;
        hl[7] = 0x137e2179;
        crypto_hashblocks_hl(hh, hl, m, n);
        n %= 128;
        for (let i = 0; i < n; i++)
            x[i] = m[b - n + i];
        x[n] = 128;
        n = 256 - 128 * (n < 112 ? 1 : 0);
        x[n - 9] = 0;
        ts64(x, n - 8, (b / 0x20000000) | 0, b << 3);
        crypto_hashblocks_hl(hh, hl, x, n);
        for (let i = 0; i < 8; i++)
            ts64(out, 8 * i, hh[i], hl[i]);
        return 0;
    }
    /**
     * Incremental version of crypto_hash.
     */
    class HashState {
        constructor() {
            this.hh = new Int32Array(8);
            this.hl = new Int32Array(8);
            this.next = new Uint8Array(128);
            this.p = 0;
            this.total = 0;
            this.hh[0] = 0x6a09e667;
            this.hh[1] = 0xbb67ae85;
            this.hh[2] = 0x3c6ef372;
            this.hh[3] = 0xa54ff53a;
            this.hh[4] = 0x510e527f;
            this.hh[5] = 0x9b05688c;
            this.hh[6] = 0x1f83d9ab;
            this.hh[7] = 0x5be0cd19;
            this.hl[0] = 0xf3bcc908;
            this.hl[1] = 0x84caa73b;
            this.hl[2] = 0xfe94f82b;
            this.hl[3] = 0x5f1d36f1;
            this.hl[4] = 0xade682d1;
            this.hl[5] = 0x2b3e6c1f;
            this.hl[6] = 0xfb41bd6b;
            this.hl[7] = 0x137e2179;
        }
        update(data) {
            this.total += data.length;
            let i = 0;
            while (i < data.length) {
                const r = 128 - this.p;
                if (r > data.length - i) {
                    for (let j = 0; i + j < data.length; j++) {
                        this.next[this.p + j] = data[i + j];
                    }
                    this.p += data.length - i;
                    break;
                }
                else {
                    for (let j = 0; this.p + j < 128; j++) {
                        this.next[this.p + j] = data[i + j];
                    }
                    crypto_hashblocks_hl(this.hh, this.hl, this.next, 128);
                    i += 128 - this.p;
                    this.p = 0;
                }
            }
            return this;
        }
        finish() {
            const out = new Uint8Array(64);
            let n = this.p;
            const x = new Uint8Array(256);
            const b = this.total;
            for (let i = 0; i < n; i++)
                x[i] = this.next[i];
            x[n] = 128;
            n = 256 - 128 * (n < 112 ? 1 : 0);
            x[n - 9] = 0;
            ts64(x, n - 8, (b / 0x20000000) | 0, b << 3);
            crypto_hashblocks_hl(this.hh, this.hl, x, n);
            for (let i = 0; i < 8; i++)
                ts64(out, 8 * i, this.hh[i], this.hl[i]);
            return out;
        }
    }
    function add$1(p, q) {
        const a = gf(), b = gf(), c = gf(), d = gf(), e = gf(), f = gf(), g = gf(), h = gf(), t = gf();
        Z(a, p[1], p[0]);
        Z(t, q[1], q[0]);
        M(a, a, t);
        A(b, p[0], p[1]);
        A(t, q[0], q[1]);
        M(b, b, t);
        M(c, p[3], q[3]);
        M(c, c, D2);
        M(d, p[2], q[2]);
        A(d, d, d);
        Z(e, b, a);
        Z(f, d, c);
        A(g, d, c);
        A(h, b, a);
        M(p[0], e, f);
        M(p[1], h, g);
        M(p[2], g, f);
        M(p[3], e, h);
    }
    function cswap(p, q, b) {
        let i;
        for (i = 0; i < 4; i++) {
            sel25519(p[i], q[i], b);
        }
    }
    function pack(r, p) {
        const tx = gf(), ty = gf(), zi = gf();
        inv25519(zi, p[2]);
        M(tx, p[0], zi);
        M(ty, p[1], zi);
        pack25519(r, ty);
        r[31] ^= par25519(tx) << 7;
    }
    function scalarmult(p, q, s) {
        let b, i;
        set25519(p[0], gf0);
        set25519(p[1], gf1);
        set25519(p[2], gf1);
        set25519(p[3], gf0);
        for (i = 255; i >= 0; --i) {
            b = (s[(i / 8) | 0] >> (i & 7)) & 1;
            cswap(p, q, b);
            add$1(q, p);
            add$1(p, p);
            cswap(p, q, b);
        }
    }
    function scalarbase(p, s) {
        const q = [gf(), gf(), gf(), gf()];
        set25519(q[0], X);
        set25519(q[1], Y);
        set25519(q[2], gf1);
        M(q[3], X, Y);
        scalarmult(p, q, s);
    }
    function crypto_sign_keypair(pk, sk, seeded) {
        const d = new Uint8Array(64);
        const p = [gf(), gf(), gf(), gf()];
        if (!seeded)
            randombytes(sk, 32);
        crypto_hash(d, sk, 32);
        d[0] &= 248;
        d[31] &= 127;
        d[31] |= 64;
        scalarbase(p, d);
        pack(pk, p);
        for (let i = 0; i < 32; i++)
            sk[i + 32] = pk[i];
        return 0;
    }
    const L = new Float64Array([
        0xed,
        0xd3,
        0xf5,
        0x5c,
        0x1a,
        0x63,
        0x12,
        0x58,
        0xd6,
        0x9c,
        0xf7,
        0xa2,
        0xde,
        0xf9,
        0xde,
        0x14,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0x10,
    ]);
    function modL(r, x) {
        let carry, i, j, k;
        for (i = 63; i >= 32; --i) {
            carry = 0;
            for (j = i - 32, k = i - 12; j < k; ++j) {
                x[j] += carry - 16 * x[i] * L[j - (i - 32)];
                carry = (x[j] + 128) >> 8;
                x[j] -= carry * 256;
            }
            x[j] += carry;
            x[i] = 0;
        }
        carry = 0;
        for (j = 0; j < 32; j++) {
            x[j] += carry - (x[31] >> 4) * L[j];
            carry = x[j] >> 8;
            x[j] &= 255;
        }
        for (j = 0; j < 32; j++)
            x[j] -= carry * L[j];
        for (i = 0; i < 32; i++) {
            x[i + 1] += x[i] >> 8;
            r[i] = x[i] & 255;
        }
    }
    function reduce(r) {
        const x = new Float64Array(64);
        for (let i = 0; i < 64; i++)
            x[i] = r[i];
        for (let i = 0; i < 64; i++)
            r[i] = 0;
        modL(r, x);
    }
    // Note: difference from C - smlen returned, not passed as argument.
    function crypto_sign(sm, m, n, sk) {
        const d = new Uint8Array(64), h = new Uint8Array(64), r = new Uint8Array(64);
        let i, j;
        const x = new Float64Array(64);
        const p = [gf(), gf(), gf(), gf()];
        crypto_hash(d, sk, 32);
        d[0] &= 248;
        d[31] &= 127;
        d[31] |= 64;
        const smlen = n + 64;
        for (i = 0; i < n; i++)
            sm[64 + i] = m[i];
        for (i = 0; i < 32; i++)
            sm[32 + i] = d[32 + i];
        crypto_hash(r, sm.subarray(32), n + 32);
        reduce(r);
        scalarbase(p, r);
        pack(sm, p);
        for (i = 32; i < 64; i++)
            sm[i] = sk[i];
        crypto_hash(h, sm, n + 64);
        reduce(h);
        for (i = 0; i < 64; i++)
            x[i] = 0;
        for (i = 0; i < 32; i++)
            x[i] = r[i];
        for (i = 0; i < 32; i++) {
            for (j = 0; j < 32; j++) {
                x[i + j] += h[i] * d[j];
            }
        }
        modL(sm.subarray(32), x);
        return smlen;
    }
    function unpackneg(r, p) {
        const t = gf();
        const chk = gf();
        const num = gf();
        const den = gf();
        const den2 = gf();
        const den4 = gf();
        const den6 = gf();
        set25519(r[2], gf1);
        unpack25519(r[1], p);
        S(num, r[1]);
        M(den, num, D);
        Z(num, num, r[2]);
        A(den, r[2], den);
        S(den2, den);
        S(den4, den2);
        M(den6, den4, den2);
        M(t, den6, num);
        M(t, t, den);
        pow2523(t, t);
        M(t, t, num);
        M(t, t, den);
        M(t, t, den);
        M(r[0], t, den);
        S(chk, r[0]);
        M(chk, chk, den);
        if (neq25519(chk, num))
            M(r[0], r[0], I);
        S(chk, r[0]);
        M(chk, chk, den);
        if (neq25519(chk, num))
            return -1;
        if (par25519(r[0]) === p[31] >> 7)
            Z(r[0], gf0, r[0]);
        M(r[3], r[0], r[1]);
        return 0;
    }
    function crypto_sign_open(m, sm, n, pk) {
        let i, mlen;
        const t = new Uint8Array(32), h = new Uint8Array(64);
        const p = [gf(), gf(), gf(), gf()], q = [gf(), gf(), gf(), gf()];
        mlen = -1;
        if (n < 64)
            return -1;
        if (unpackneg(q, pk))
            return -1;
        for (i = 0; i < n; i++)
            m[i] = sm[i];
        for (i = 0; i < 32; i++)
            m[i + 32] = pk[i];
        crypto_hash(h, m, n);
        reduce(h);
        scalarmult(p, q, h);
        scalarbase(q, sm.subarray(32));
        add$1(p, q);
        pack(t, p);
        n -= 64;
        if (crypto_verify_32(sm, 0, t, 0)) {
            for (i = 0; i < n; i++)
                m[i] = 0;
            return -1;
        }
        for (i = 0; i < n; i++)
            m[i] = sm[i + 64];
        mlen = n;
        return mlen;
    }
    const crypto_scalarmult_BYTES = 32, crypto_scalarmult_SCALARBYTES = 32, crypto_sign_BYTES = 64, crypto_sign_PUBLICKEYBYTES = 32, crypto_sign_SECRETKEYBYTES = 64, crypto_sign_SEEDBYTES = 32, crypto_hash_BYTES = 64;
    /* High-level API */
    function checkArrayTypes(...args) {
        for (let i = 0; i < args.length; i++) {
            if (!(args[i] instanceof Uint8Array))
                throw new TypeError("unexpected type, use Uint8Array");
        }
    }
    function cleanup(arr) {
        for (let i = 0; i < arr.length; i++)
            arr[i] = 0;
    }
    function randomBytes(n) {
        const b = new Uint8Array(n);
        randombytes(b, n);
        return b;
    }
    function scalarMult(n, p) {
        checkArrayTypes(n, p);
        if (n.length !== crypto_scalarmult_SCALARBYTES)
            throw new Error("bad n size");
        if (p.length !== crypto_scalarmult_BYTES)
            throw new Error("bad p size");
        const q = new Uint8Array(crypto_scalarmult_BYTES);
        crypto_scalarmult(q, n, p);
        return q;
    }
    function scalarMult_base(n) {
        checkArrayTypes(n);
        if (n.length !== crypto_scalarmult_SCALARBYTES)
            throw new Error("bad n size");
        const q = new Uint8Array(crypto_scalarmult_BYTES);
        crypto_scalarmult_base(q, n);
        return q;
    }
    function sign(msg, secretKey) {
        checkArrayTypes(msg, secretKey);
        if (secretKey.length !== crypto_sign_SECRETKEYBYTES)
            throw new Error("bad secret key size");
        const signedMsg = new Uint8Array(crypto_sign_BYTES + msg.length);
        crypto_sign(signedMsg, msg, msg.length, secretKey);
        return signedMsg;
    }
    function sign_detached(msg, secretKey) {
        const signedMsg = sign(msg, secretKey);
        const sig = new Uint8Array(crypto_sign_BYTES);
        for (let i = 0; i < sig.length; i++)
            sig[i] = signedMsg[i];
        return sig;
    }
    function sign_detached_verify(msg, sig, publicKey) {
        checkArrayTypes(msg, sig, publicKey);
        if (sig.length !== crypto_sign_BYTES)
            throw new Error("bad signature size");
        if (publicKey.length !== crypto_sign_PUBLICKEYBYTES)
            throw new Error("bad public key size");
        const sm = new Uint8Array(crypto_sign_BYTES + msg.length);
        const m = new Uint8Array(crypto_sign_BYTES + msg.length);
        let i;
        for (i = 0; i < crypto_sign_BYTES; i++)
            sm[i] = sig[i];
        for (i = 0; i < msg.length; i++)
            sm[i + crypto_sign_BYTES] = msg[i];
        return crypto_sign_open(m, sm, sm.length, publicKey) >= 0;
    }
    function sign_keyPair_fromSeed(seed) {
        checkArrayTypes(seed);
        if (seed.length !== crypto_sign_SEEDBYTES)
            throw new Error("bad seed size");
        const pk = new Uint8Array(crypto_sign_PUBLICKEYBYTES);
        const sk = new Uint8Array(crypto_sign_SECRETKEYBYTES);
        for (let i = 0; i < 32; i++)
            sk[i] = seed[i];
        crypto_sign_keypair(pk, sk, true);
        return { publicKey: pk, secretKey: sk };
    }
    function hash(msg) {
        checkArrayTypes(msg);
        const h = new Uint8Array(crypto_hash_BYTES);
        crypto_hash(h, msg, msg.length);
        return h;
    }
    function setPRNG(fn) {
        randombytes = fn;
    }
    function sign_ed25519_pk_to_curve25519(ed25519_pk) {
        const ge_a = [gf(), gf(), gf(), gf()];
        const x = gf();
        const one_minus_y = gf();
        const x25519_pk = new Uint8Array(32);
        if (unpackneg(ge_a, ed25519_pk)) {
            throw Error("invalid public key");
        }
        set25519(one_minus_y, gf1);
        Z(one_minus_y, one_minus_y, ge_a[1]);
        set25519(x, gf1);
        A(x, x, ge_a[1]);
        inv25519(one_minus_y, one_minus_y);
        M(x, x, one_minus_y);
        pack25519(x25519_pk, x);
        return x25519_pk;
    }
    (function () {
        // Initialize PRNG if environment provides CSPRNG.
        // If not, methods calling randombytes will throw.
        // @ts-ignore-error
        const cr = typeof self !== "undefined" ? self.crypto || self.msCrypto : null;
        if (cr && cr.getRandomValues) {
            // Browsers.
            const QUOTA = 65536;
            setPRNG(function (x, n) {
                let i;
                const v = new Uint8Array(n);
                for (i = 0; i < n; i += QUOTA) {
                    cr.getRandomValues(v.subarray(i, i + Math.min(n - i, QUOTA)));
                }
                for (i = 0; i < n; i++)
                    x[i] = v[i];
                cleanup(v);
            });
        }
        else if (typeof require !== "undefined") {
            // Node.js.
            // eslint-disable-next-line @typescript-eslint/no-var-requires
            const cr = require("crypto");
            if (cr && cr.randomBytes) {
                setPRNG(function (x, n) {
                    const v = cr.randomBytes(n);
                    for (let i = 0; i < n; i++)
                        x[i] = v[i];
                    cleanup(v);
                });
            }
        }
    })();

    function createCommonjsModule(fn, basedir, module) {
    	return module = {
    		path: basedir,
    		exports: {},
    		require: function (path, base) {
    			return commonjsRequire(path, (base === undefined || base === null) ? module.path : base);
    		}
    	}, fn(module, module.exports), module.exports;
    }

    function commonjsRequire () {
    	throw new Error('Dynamic requires are not currently supported by @rollup/plugin-commonjs');
    }

    var BigInteger = createCommonjsModule(function (module) {
    var bigInt = (function (undefined$1) {

        var BASE = 1e7,
            LOG_BASE = 7,
            MAX_INT = 9007199254740992,
            MAX_INT_ARR = smallToArray(MAX_INT),
            DEFAULT_ALPHABET = "0123456789abcdefghijklmnopqrstuvwxyz";

        var supportsNativeBigInt = typeof BigInt === "function";

        function Integer(v, radix, alphabet, caseSensitive) {
            if (typeof v === "undefined") return Integer[0];
            if (typeof radix !== "undefined") return +radix === 10 && !alphabet ? parseValue(v) : parseBase(v, radix, alphabet, caseSensitive);
            return parseValue(v);
        }

        function BigInteger(value, sign) {
            this.value = value;
            this.sign = sign;
            this.isSmall = false;
        }
        BigInteger.prototype = Object.create(Integer.prototype);

        function SmallInteger(value) {
            this.value = value;
            this.sign = value < 0;
            this.isSmall = true;
        }
        SmallInteger.prototype = Object.create(Integer.prototype);

        function NativeBigInt(value) {
            this.value = value;
        }
        NativeBigInt.prototype = Object.create(Integer.prototype);

        function isPrecise(n) {
            return -MAX_INT < n && n < MAX_INT;
        }

        function smallToArray(n) { // For performance reasons doesn't reference BASE, need to change this function if BASE changes
            if (n < 1e7)
                return [n];
            if (n < 1e14)
                return [n % 1e7, Math.floor(n / 1e7)];
            return [n % 1e7, Math.floor(n / 1e7) % 1e7, Math.floor(n / 1e14)];
        }

        function arrayToSmall(arr) { // If BASE changes this function may need to change
            trim(arr);
            var length = arr.length;
            if (length < 4 && compareAbs(arr, MAX_INT_ARR) < 0) {
                switch (length) {
                    case 0: return 0;
                    case 1: return arr[0];
                    case 2: return arr[0] + arr[1] * BASE;
                    default: return arr[0] + (arr[1] + arr[2] * BASE) * BASE;
                }
            }
            return arr;
        }

        function trim(v) {
            var i = v.length;
            while (v[--i] === 0);
            v.length = i + 1;
        }

        function createArray(length) { // function shamelessly stolen from Yaffle's library https://github.com/Yaffle/BigInteger
            var x = new Array(length);
            var i = -1;
            while (++i < length) {
                x[i] = 0;
            }
            return x;
        }

        function truncate(n) {
            if (n > 0) return Math.floor(n);
            return Math.ceil(n);
        }

        function add(a, b) { // assumes a and b are arrays with a.length >= b.length
            var l_a = a.length,
                l_b = b.length,
                r = new Array(l_a),
                carry = 0,
                base = BASE,
                sum, i;
            for (i = 0; i < l_b; i++) {
                sum = a[i] + b[i] + carry;
                carry = sum >= base ? 1 : 0;
                r[i] = sum - carry * base;
            }
            while (i < l_a) {
                sum = a[i] + carry;
                carry = sum === base ? 1 : 0;
                r[i++] = sum - carry * base;
            }
            if (carry > 0) r.push(carry);
            return r;
        }

        function addAny(a, b) {
            if (a.length >= b.length) return add(a, b);
            return add(b, a);
        }

        function addSmall(a, carry) { // assumes a is array, carry is number with 0 <= carry < MAX_INT
            var l = a.length,
                r = new Array(l),
                base = BASE,
                sum, i;
            for (i = 0; i < l; i++) {
                sum = a[i] - base + carry;
                carry = Math.floor(sum / base);
                r[i] = sum - carry * base;
                carry += 1;
            }
            while (carry > 0) {
                r[i++] = carry % base;
                carry = Math.floor(carry / base);
            }
            return r;
        }

        BigInteger.prototype.add = function (v) {
            var n = parseValue(v);
            if (this.sign !== n.sign) {
                return this.subtract(n.negate());
            }
            var a = this.value, b = n.value;
            if (n.isSmall) {
                return new BigInteger(addSmall(a, Math.abs(b)), this.sign);
            }
            return new BigInteger(addAny(a, b), this.sign);
        };
        BigInteger.prototype.plus = BigInteger.prototype.add;

        SmallInteger.prototype.add = function (v) {
            var n = parseValue(v);
            var a = this.value;
            if (a < 0 !== n.sign) {
                return this.subtract(n.negate());
            }
            var b = n.value;
            if (n.isSmall) {
                if (isPrecise(a + b)) return new SmallInteger(a + b);
                b = smallToArray(Math.abs(b));
            }
            return new BigInteger(addSmall(b, Math.abs(a)), a < 0);
        };
        SmallInteger.prototype.plus = SmallInteger.prototype.add;

        NativeBigInt.prototype.add = function (v) {
            return new NativeBigInt(this.value + parseValue(v).value);
        };
        NativeBigInt.prototype.plus = NativeBigInt.prototype.add;

        function subtract(a, b) { // assumes a and b are arrays with a >= b
            var a_l = a.length,
                b_l = b.length,
                r = new Array(a_l),
                borrow = 0,
                base = BASE,
                i, difference;
            for (i = 0; i < b_l; i++) {
                difference = a[i] - borrow - b[i];
                if (difference < 0) {
                    difference += base;
                    borrow = 1;
                } else borrow = 0;
                r[i] = difference;
            }
            for (i = b_l; i < a_l; i++) {
                difference = a[i] - borrow;
                if (difference < 0) difference += base;
                else {
                    r[i++] = difference;
                    break;
                }
                r[i] = difference;
            }
            for (; i < a_l; i++) {
                r[i] = a[i];
            }
            trim(r);
            return r;
        }

        function subtractAny(a, b, sign) {
            var value;
            if (compareAbs(a, b) >= 0) {
                value = subtract(a, b);
            } else {
                value = subtract(b, a);
                sign = !sign;
            }
            value = arrayToSmall(value);
            if (typeof value === "number") {
                if (sign) value = -value;
                return new SmallInteger(value);
            }
            return new BigInteger(value, sign);
        }

        function subtractSmall(a, b, sign) { // assumes a is array, b is number with 0 <= b < MAX_INT
            var l = a.length,
                r = new Array(l),
                carry = -b,
                base = BASE,
                i, difference;
            for (i = 0; i < l; i++) {
                difference = a[i] + carry;
                carry = Math.floor(difference / base);
                difference %= base;
                r[i] = difference < 0 ? difference + base : difference;
            }
            r = arrayToSmall(r);
            if (typeof r === "number") {
                if (sign) r = -r;
                return new SmallInteger(r);
            } return new BigInteger(r, sign);
        }

        BigInteger.prototype.subtract = function (v) {
            var n = parseValue(v);
            if (this.sign !== n.sign) {
                return this.add(n.negate());
            }
            var a = this.value, b = n.value;
            if (n.isSmall)
                return subtractSmall(a, Math.abs(b), this.sign);
            return subtractAny(a, b, this.sign);
        };
        BigInteger.prototype.minus = BigInteger.prototype.subtract;

        SmallInteger.prototype.subtract = function (v) {
            var n = parseValue(v);
            var a = this.value;
            if (a < 0 !== n.sign) {
                return this.add(n.negate());
            }
            var b = n.value;
            if (n.isSmall) {
                return new SmallInteger(a - b);
            }
            return subtractSmall(b, Math.abs(a), a >= 0);
        };
        SmallInteger.prototype.minus = SmallInteger.prototype.subtract;

        NativeBigInt.prototype.subtract = function (v) {
            return new NativeBigInt(this.value - parseValue(v).value);
        };
        NativeBigInt.prototype.minus = NativeBigInt.prototype.subtract;

        BigInteger.prototype.negate = function () {
            return new BigInteger(this.value, !this.sign);
        };
        SmallInteger.prototype.negate = function () {
            var sign = this.sign;
            var small = new SmallInteger(-this.value);
            small.sign = !sign;
            return small;
        };
        NativeBigInt.prototype.negate = function () {
            return new NativeBigInt(-this.value);
        };

        BigInteger.prototype.abs = function () {
            return new BigInteger(this.value, false);
        };
        SmallInteger.prototype.abs = function () {
            return new SmallInteger(Math.abs(this.value));
        };
        NativeBigInt.prototype.abs = function () {
            return new NativeBigInt(this.value >= 0 ? this.value : -this.value);
        };


        function multiplyLong(a, b) {
            var a_l = a.length,
                b_l = b.length,
                l = a_l + b_l,
                r = createArray(l),
                base = BASE,
                product, carry, i, a_i, b_j;
            for (i = 0; i < a_l; ++i) {
                a_i = a[i];
                for (var j = 0; j < b_l; ++j) {
                    b_j = b[j];
                    product = a_i * b_j + r[i + j];
                    carry = Math.floor(product / base);
                    r[i + j] = product - carry * base;
                    r[i + j + 1] += carry;
                }
            }
            trim(r);
            return r;
        }

        function multiplySmall(a, b) { // assumes a is array, b is number with |b| < BASE
            var l = a.length,
                r = new Array(l),
                base = BASE,
                carry = 0,
                product, i;
            for (i = 0; i < l; i++) {
                product = a[i] * b + carry;
                carry = Math.floor(product / base);
                r[i] = product - carry * base;
            }
            while (carry > 0) {
                r[i++] = carry % base;
                carry = Math.floor(carry / base);
            }
            return r;
        }

        function shiftLeft(x, n) {
            var r = [];
            while (n-- > 0) r.push(0);
            return r.concat(x);
        }

        function multiplyKaratsuba(x, y) {
            var n = Math.max(x.length, y.length);

            if (n <= 30) return multiplyLong(x, y);
            n = Math.ceil(n / 2);

            var b = x.slice(n),
                a = x.slice(0, n),
                d = y.slice(n),
                c = y.slice(0, n);

            var ac = multiplyKaratsuba(a, c),
                bd = multiplyKaratsuba(b, d),
                abcd = multiplyKaratsuba(addAny(a, b), addAny(c, d));

            var product = addAny(addAny(ac, shiftLeft(subtract(subtract(abcd, ac), bd), n)), shiftLeft(bd, 2 * n));
            trim(product);
            return product;
        }

        // The following function is derived from a surface fit of a graph plotting the performance difference
        // between long multiplication and karatsuba multiplication versus the lengths of the two arrays.
        function useKaratsuba(l1, l2) {
            return -0.012 * l1 - 0.012 * l2 + 0.000015 * l1 * l2 > 0;
        }

        BigInteger.prototype.multiply = function (v) {
            var n = parseValue(v),
                a = this.value, b = n.value,
                sign = this.sign !== n.sign,
                abs;
            if (n.isSmall) {
                if (b === 0) return Integer[0];
                if (b === 1) return this;
                if (b === -1) return this.negate();
                abs = Math.abs(b);
                if (abs < BASE) {
                    return new BigInteger(multiplySmall(a, abs), sign);
                }
                b = smallToArray(abs);
            }
            if (useKaratsuba(a.length, b.length)) // Karatsuba is only faster for certain array sizes
                return new BigInteger(multiplyKaratsuba(a, b), sign);
            return new BigInteger(multiplyLong(a, b), sign);
        };

        BigInteger.prototype.times = BigInteger.prototype.multiply;

        function multiplySmallAndArray(a, b, sign) { // a >= 0
            if (a < BASE) {
                return new BigInteger(multiplySmall(b, a), sign);
            }
            return new BigInteger(multiplyLong(b, smallToArray(a)), sign);
        }
        SmallInteger.prototype._multiplyBySmall = function (a) {
            if (isPrecise(a.value * this.value)) {
                return new SmallInteger(a.value * this.value);
            }
            return multiplySmallAndArray(Math.abs(a.value), smallToArray(Math.abs(this.value)), this.sign !== a.sign);
        };
        BigInteger.prototype._multiplyBySmall = function (a) {
            if (a.value === 0) return Integer[0];
            if (a.value === 1) return this;
            if (a.value === -1) return this.negate();
            return multiplySmallAndArray(Math.abs(a.value), this.value, this.sign !== a.sign);
        };
        SmallInteger.prototype.multiply = function (v) {
            return parseValue(v)._multiplyBySmall(this);
        };
        SmallInteger.prototype.times = SmallInteger.prototype.multiply;

        NativeBigInt.prototype.multiply = function (v) {
            return new NativeBigInt(this.value * parseValue(v).value);
        };
        NativeBigInt.prototype.times = NativeBigInt.prototype.multiply;

        function square(a) {
            //console.assert(2 * BASE * BASE < MAX_INT);
            var l = a.length,
                r = createArray(l + l),
                base = BASE,
                product, carry, i, a_i, a_j;
            for (i = 0; i < l; i++) {
                a_i = a[i];
                carry = 0 - a_i * a_i;
                for (var j = i; j < l; j++) {
                    a_j = a[j];
                    product = 2 * (a_i * a_j) + r[i + j] + carry;
                    carry = Math.floor(product / base);
                    r[i + j] = product - carry * base;
                }
                r[i + l] = carry;
            }
            trim(r);
            return r;
        }

        BigInteger.prototype.square = function () {
            return new BigInteger(square(this.value), false);
        };

        SmallInteger.prototype.square = function () {
            var value = this.value * this.value;
            if (isPrecise(value)) return new SmallInteger(value);
            return new BigInteger(square(smallToArray(Math.abs(this.value))), false);
        };

        NativeBigInt.prototype.square = function (v) {
            return new NativeBigInt(this.value * this.value);
        };

        function divMod1(a, b) { // Left over from previous version. Performs faster than divMod2 on smaller input sizes.
            var a_l = a.length,
                b_l = b.length,
                base = BASE,
                result = createArray(b.length),
                divisorMostSignificantDigit = b[b_l - 1],
                // normalization
                lambda = Math.ceil(base / (2 * divisorMostSignificantDigit)),
                remainder = multiplySmall(a, lambda),
                divisor = multiplySmall(b, lambda),
                quotientDigit, shift, carry, borrow, i, l, q;
            if (remainder.length <= a_l) remainder.push(0);
            divisor.push(0);
            divisorMostSignificantDigit = divisor[b_l - 1];
            for (shift = a_l - b_l; shift >= 0; shift--) {
                quotientDigit = base - 1;
                if (remainder[shift + b_l] !== divisorMostSignificantDigit) {
                    quotientDigit = Math.floor((remainder[shift + b_l] * base + remainder[shift + b_l - 1]) / divisorMostSignificantDigit);
                }
                // quotientDigit <= base - 1
                carry = 0;
                borrow = 0;
                l = divisor.length;
                for (i = 0; i < l; i++) {
                    carry += quotientDigit * divisor[i];
                    q = Math.floor(carry / base);
                    borrow += remainder[shift + i] - (carry - q * base);
                    carry = q;
                    if (borrow < 0) {
                        remainder[shift + i] = borrow + base;
                        borrow = -1;
                    } else {
                        remainder[shift + i] = borrow;
                        borrow = 0;
                    }
                }
                while (borrow !== 0) {
                    quotientDigit -= 1;
                    carry = 0;
                    for (i = 0; i < l; i++) {
                        carry += remainder[shift + i] - base + divisor[i];
                        if (carry < 0) {
                            remainder[shift + i] = carry + base;
                            carry = 0;
                        } else {
                            remainder[shift + i] = carry;
                            carry = 1;
                        }
                    }
                    borrow += carry;
                }
                result[shift] = quotientDigit;
            }
            // denormalization
            remainder = divModSmall(remainder, lambda)[0];
            return [arrayToSmall(result), arrayToSmall(remainder)];
        }

        function divMod2(a, b) { // Implementation idea shamelessly stolen from Silent Matt's library http://silentmatt.com/biginteger/
            // Performs faster than divMod1 on larger input sizes.
            var a_l = a.length,
                b_l = b.length,
                result = [],
                part = [],
                base = BASE,
                guess, xlen, highx, highy, check;
            while (a_l) {
                part.unshift(a[--a_l]);
                trim(part);
                if (compareAbs(part, b) < 0) {
                    result.push(0);
                    continue;
                }
                xlen = part.length;
                highx = part[xlen - 1] * base + part[xlen - 2];
                highy = b[b_l - 1] * base + b[b_l - 2];
                if (xlen > b_l) {
                    highx = (highx + 1) * base;
                }
                guess = Math.ceil(highx / highy);
                do {
                    check = multiplySmall(b, guess);
                    if (compareAbs(check, part) <= 0) break;
                    guess--;
                } while (guess);
                result.push(guess);
                part = subtract(part, check);
            }
            result.reverse();
            return [arrayToSmall(result), arrayToSmall(part)];
        }

        function divModSmall(value, lambda) {
            var length = value.length,
                quotient = createArray(length),
                base = BASE,
                i, q, remainder, divisor;
            remainder = 0;
            for (i = length - 1; i >= 0; --i) {
                divisor = remainder * base + value[i];
                q = truncate(divisor / lambda);
                remainder = divisor - q * lambda;
                quotient[i] = q | 0;
            }
            return [quotient, remainder | 0];
        }

        function divModAny(self, v) {
            var value, n = parseValue(v);
            if (supportsNativeBigInt) {
                return [new NativeBigInt(self.value / n.value), new NativeBigInt(self.value % n.value)];
            }
            var a = self.value, b = n.value;
            var quotient;
            if (b === 0) throw new Error("Cannot divide by zero");
            if (self.isSmall) {
                if (n.isSmall) {
                    return [new SmallInteger(truncate(a / b)), new SmallInteger(a % b)];
                }
                return [Integer[0], self];
            }
            if (n.isSmall) {
                if (b === 1) return [self, Integer[0]];
                if (b == -1) return [self.negate(), Integer[0]];
                var abs = Math.abs(b);
                if (abs < BASE) {
                    value = divModSmall(a, abs);
                    quotient = arrayToSmall(value[0]);
                    var remainder = value[1];
                    if (self.sign) remainder = -remainder;
                    if (typeof quotient === "number") {
                        if (self.sign !== n.sign) quotient = -quotient;
                        return [new SmallInteger(quotient), new SmallInteger(remainder)];
                    }
                    return [new BigInteger(quotient, self.sign !== n.sign), new SmallInteger(remainder)];
                }
                b = smallToArray(abs);
            }
            var comparison = compareAbs(a, b);
            if (comparison === -1) return [Integer[0], self];
            if (comparison === 0) return [Integer[self.sign === n.sign ? 1 : -1], Integer[0]];

            // divMod1 is faster on smaller input sizes
            if (a.length + b.length <= 200)
                value = divMod1(a, b);
            else value = divMod2(a, b);

            quotient = value[0];
            var qSign = self.sign !== n.sign,
                mod = value[1],
                mSign = self.sign;
            if (typeof quotient === "number") {
                if (qSign) quotient = -quotient;
                quotient = new SmallInteger(quotient);
            } else quotient = new BigInteger(quotient, qSign);
            if (typeof mod === "number") {
                if (mSign) mod = -mod;
                mod = new SmallInteger(mod);
            } else mod = new BigInteger(mod, mSign);
            return [quotient, mod];
        }

        BigInteger.prototype.divmod = function (v) {
            var result = divModAny(this, v);
            return {
                quotient: result[0],
                remainder: result[1]
            };
        };
        NativeBigInt.prototype.divmod = SmallInteger.prototype.divmod = BigInteger.prototype.divmod;


        BigInteger.prototype.divide = function (v) {
            return divModAny(this, v)[0];
        };
        NativeBigInt.prototype.over = NativeBigInt.prototype.divide = function (v) {
            return new NativeBigInt(this.value / parseValue(v).value);
        };
        SmallInteger.prototype.over = SmallInteger.prototype.divide = BigInteger.prototype.over = BigInteger.prototype.divide;

        BigInteger.prototype.mod = function (v) {
            return divModAny(this, v)[1];
        };
        NativeBigInt.prototype.mod = NativeBigInt.prototype.remainder = function (v) {
            return new NativeBigInt(this.value % parseValue(v).value);
        };
        SmallInteger.prototype.remainder = SmallInteger.prototype.mod = BigInteger.prototype.remainder = BigInteger.prototype.mod;

        BigInteger.prototype.pow = function (v) {
            var n = parseValue(v),
                a = this.value,
                b = n.value,
                value, x, y;
            if (b === 0) return Integer[1];
            if (a === 0) return Integer[0];
            if (a === 1) return Integer[1];
            if (a === -1) return n.isEven() ? Integer[1] : Integer[-1];
            if (n.sign) {
                return Integer[0];
            }
            if (!n.isSmall) throw new Error("The exponent " + n.toString() + " is too large.");
            if (this.isSmall) {
                if (isPrecise(value = Math.pow(a, b)))
                    return new SmallInteger(truncate(value));
            }
            x = this;
            y = Integer[1];
            while (true) {
                if (b & 1 === 1) {
                    y = y.times(x);
                    --b;
                }
                if (b === 0) break;
                b /= 2;
                x = x.square();
            }
            return y;
        };
        SmallInteger.prototype.pow = BigInteger.prototype.pow;

        NativeBigInt.prototype.pow = function (v) {
            var n = parseValue(v);
            var a = this.value, b = n.value;
            var _0 = BigInt(0), _1 = BigInt(1), _2 = BigInt(2);
            if (b === _0) return Integer[1];
            if (a === _0) return Integer[0];
            if (a === _1) return Integer[1];
            if (a === BigInt(-1)) return n.isEven() ? Integer[1] : Integer[-1];
            if (n.isNegative()) return new NativeBigInt(_0);
            var x = this;
            var y = Integer[1];
            while (true) {
                if ((b & _1) === _1) {
                    y = y.times(x);
                    --b;
                }
                if (b === _0) break;
                b /= _2;
                x = x.square();
            }
            return y;
        };

        BigInteger.prototype.modPow = function (exp, mod) {
            exp = parseValue(exp);
            mod = parseValue(mod);
            if (mod.isZero()) throw new Error("Cannot take modPow with modulus 0");
            var r = Integer[1],
                base = this.mod(mod);
            if (exp.isNegative()) {
                exp = exp.multiply(Integer[-1]);
                base = base.modInv(mod);
            }
            while (exp.isPositive()) {
                if (base.isZero()) return Integer[0];
                if (exp.isOdd()) r = r.multiply(base).mod(mod);
                exp = exp.divide(2);
                base = base.square().mod(mod);
            }
            return r;
        };
        NativeBigInt.prototype.modPow = SmallInteger.prototype.modPow = BigInteger.prototype.modPow;

        function compareAbs(a, b) {
            if (a.length !== b.length) {
                return a.length > b.length ? 1 : -1;
            }
            for (var i = a.length - 1; i >= 0; i--) {
                if (a[i] !== b[i]) return a[i] > b[i] ? 1 : -1;
            }
            return 0;
        }

        BigInteger.prototype.compareAbs = function (v) {
            var n = parseValue(v),
                a = this.value,
                b = n.value;
            if (n.isSmall) return 1;
            return compareAbs(a, b);
        };
        SmallInteger.prototype.compareAbs = function (v) {
            var n = parseValue(v),
                a = Math.abs(this.value),
                b = n.value;
            if (n.isSmall) {
                b = Math.abs(b);
                return a === b ? 0 : a > b ? 1 : -1;
            }
            return -1;
        };
        NativeBigInt.prototype.compareAbs = function (v) {
            var a = this.value;
            var b = parseValue(v).value;
            a = a >= 0 ? a : -a;
            b = b >= 0 ? b : -b;
            return a === b ? 0 : a > b ? 1 : -1;
        };

        BigInteger.prototype.compare = function (v) {
            // See discussion about comparison with Infinity:
            // https://github.com/peterolson/BigInteger.js/issues/61
            if (v === Infinity) {
                return -1;
            }
            if (v === -Infinity) {
                return 1;
            }

            var n = parseValue(v),
                a = this.value,
                b = n.value;
            if (this.sign !== n.sign) {
                return n.sign ? 1 : -1;
            }
            if (n.isSmall) {
                return this.sign ? -1 : 1;
            }
            return compareAbs(a, b) * (this.sign ? -1 : 1);
        };
        BigInteger.prototype.compareTo = BigInteger.prototype.compare;

        SmallInteger.prototype.compare = function (v) {
            if (v === Infinity) {
                return -1;
            }
            if (v === -Infinity) {
                return 1;
            }

            var n = parseValue(v),
                a = this.value,
                b = n.value;
            if (n.isSmall) {
                return a == b ? 0 : a > b ? 1 : -1;
            }
            if (a < 0 !== n.sign) {
                return a < 0 ? -1 : 1;
            }
            return a < 0 ? 1 : -1;
        };
        SmallInteger.prototype.compareTo = SmallInteger.prototype.compare;

        NativeBigInt.prototype.compare = function (v) {
            if (v === Infinity) {
                return -1;
            }
            if (v === -Infinity) {
                return 1;
            }
            var a = this.value;
            var b = parseValue(v).value;
            return a === b ? 0 : a > b ? 1 : -1;
        };
        NativeBigInt.prototype.compareTo = NativeBigInt.prototype.compare;

        BigInteger.prototype.equals = function (v) {
            return this.compare(v) === 0;
        };
        NativeBigInt.prototype.eq = NativeBigInt.prototype.equals = SmallInteger.prototype.eq = SmallInteger.prototype.equals = BigInteger.prototype.eq = BigInteger.prototype.equals;

        BigInteger.prototype.notEquals = function (v) {
            return this.compare(v) !== 0;
        };
        NativeBigInt.prototype.neq = NativeBigInt.prototype.notEquals = SmallInteger.prototype.neq = SmallInteger.prototype.notEquals = BigInteger.prototype.neq = BigInteger.prototype.notEquals;

        BigInteger.prototype.greater = function (v) {
            return this.compare(v) > 0;
        };
        NativeBigInt.prototype.gt = NativeBigInt.prototype.greater = SmallInteger.prototype.gt = SmallInteger.prototype.greater = BigInteger.prototype.gt = BigInteger.prototype.greater;

        BigInteger.prototype.lesser = function (v) {
            return this.compare(v) < 0;
        };
        NativeBigInt.prototype.lt = NativeBigInt.prototype.lesser = SmallInteger.prototype.lt = SmallInteger.prototype.lesser = BigInteger.prototype.lt = BigInteger.prototype.lesser;

        BigInteger.prototype.greaterOrEquals = function (v) {
            return this.compare(v) >= 0;
        };
        NativeBigInt.prototype.geq = NativeBigInt.prototype.greaterOrEquals = SmallInteger.prototype.geq = SmallInteger.prototype.greaterOrEquals = BigInteger.prototype.geq = BigInteger.prototype.greaterOrEquals;

        BigInteger.prototype.lesserOrEquals = function (v) {
            return this.compare(v) <= 0;
        };
        NativeBigInt.prototype.leq = NativeBigInt.prototype.lesserOrEquals = SmallInteger.prototype.leq = SmallInteger.prototype.lesserOrEquals = BigInteger.prototype.leq = BigInteger.prototype.lesserOrEquals;

        BigInteger.prototype.isEven = function () {
            return (this.value[0] & 1) === 0;
        };
        SmallInteger.prototype.isEven = function () {
            return (this.value & 1) === 0;
        };
        NativeBigInt.prototype.isEven = function () {
            return (this.value & BigInt(1)) === BigInt(0);
        };

        BigInteger.prototype.isOdd = function () {
            return (this.value[0] & 1) === 1;
        };
        SmallInteger.prototype.isOdd = function () {
            return (this.value & 1) === 1;
        };
        NativeBigInt.prototype.isOdd = function () {
            return (this.value & BigInt(1)) === BigInt(1);
        };

        BigInteger.prototype.isPositive = function () {
            return !this.sign;
        };
        SmallInteger.prototype.isPositive = function () {
            return this.value > 0;
        };
        NativeBigInt.prototype.isPositive = SmallInteger.prototype.isPositive;

        BigInteger.prototype.isNegative = function () {
            return this.sign;
        };
        SmallInteger.prototype.isNegative = function () {
            return this.value < 0;
        };
        NativeBigInt.prototype.isNegative = SmallInteger.prototype.isNegative;

        BigInteger.prototype.isUnit = function () {
            return false;
        };
        SmallInteger.prototype.isUnit = function () {
            return Math.abs(this.value) === 1;
        };
        NativeBigInt.prototype.isUnit = function () {
            return this.abs().value === BigInt(1);
        };

        BigInteger.prototype.isZero = function () {
            return false;
        };
        SmallInteger.prototype.isZero = function () {
            return this.value === 0;
        };
        NativeBigInt.prototype.isZero = function () {
            return this.value === BigInt(0);
        };

        BigInteger.prototype.isDivisibleBy = function (v) {
            var n = parseValue(v);
            if (n.isZero()) return false;
            if (n.isUnit()) return true;
            if (n.compareAbs(2) === 0) return this.isEven();
            return this.mod(n).isZero();
        };
        NativeBigInt.prototype.isDivisibleBy = SmallInteger.prototype.isDivisibleBy = BigInteger.prototype.isDivisibleBy;

        function isBasicPrime(v) {
            var n = v.abs();
            if (n.isUnit()) return false;
            if (n.equals(2) || n.equals(3) || n.equals(5)) return true;
            if (n.isEven() || n.isDivisibleBy(3) || n.isDivisibleBy(5)) return false;
            if (n.lesser(49)) return true;
            // we don't know if it's prime: let the other functions figure it out
        }

        function millerRabinTest(n, a) {
            var nPrev = n.prev(),
                b = nPrev,
                r = 0,
                d, i, x;
            while (b.isEven()) b = b.divide(2), r++;
            next: for (i = 0; i < a.length; i++) {
                if (n.lesser(a[i])) continue;
                x = bigInt(a[i]).modPow(b, n);
                if (x.isUnit() || x.equals(nPrev)) continue;
                for (d = r - 1; d != 0; d--) {
                    x = x.square().mod(n);
                    if (x.isUnit()) return false;
                    if (x.equals(nPrev)) continue next;
                }
                return false;
            }
            return true;
        }

        // Set "strict" to true to force GRH-supported lower bound of 2*log(N)^2
        BigInteger.prototype.isPrime = function (strict) {
            var isPrime = isBasicPrime(this);
            if (isPrime !== undefined$1) return isPrime;
            var n = this.abs();
            var bits = n.bitLength();
            if (bits <= 64)
                return millerRabinTest(n, [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37]);
            var logN = Math.log(2) * bits.toJSNumber();
            var t = Math.ceil((strict === true) ? (2 * Math.pow(logN, 2)) : logN);
            for (var a = [], i = 0; i < t; i++) {
                a.push(bigInt(i + 2));
            }
            return millerRabinTest(n, a);
        };
        NativeBigInt.prototype.isPrime = SmallInteger.prototype.isPrime = BigInteger.prototype.isPrime;

        BigInteger.prototype.isProbablePrime = function (iterations, rng) {
            var isPrime = isBasicPrime(this);
            if (isPrime !== undefined$1) return isPrime;
            var n = this.abs();
            var t = iterations === undefined$1 ? 5 : iterations;
            for (var a = [], i = 0; i < t; i++) {
                a.push(bigInt.randBetween(2, n.minus(2), rng));
            }
            return millerRabinTest(n, a);
        };
        NativeBigInt.prototype.isProbablePrime = SmallInteger.prototype.isProbablePrime = BigInteger.prototype.isProbablePrime;

        BigInteger.prototype.modInv = function (n) {
            var t = bigInt.zero, newT = bigInt.one, r = parseValue(n), newR = this.abs(), q, lastT, lastR;
            while (!newR.isZero()) {
                q = r.divide(newR);
                lastT = t;
                lastR = r;
                t = newT;
                r = newR;
                newT = lastT.subtract(q.multiply(newT));
                newR = lastR.subtract(q.multiply(newR));
            }
            if (!r.isUnit()) throw new Error(this.toString() + " and " + n.toString() + " are not co-prime");
            if (t.compare(0) === -1) {
                t = t.add(n);
            }
            if (this.isNegative()) {
                return t.negate();
            }
            return t;
        };

        NativeBigInt.prototype.modInv = SmallInteger.prototype.modInv = BigInteger.prototype.modInv;

        BigInteger.prototype.next = function () {
            var value = this.value;
            if (this.sign) {
                return subtractSmall(value, 1, this.sign);
            }
            return new BigInteger(addSmall(value, 1), this.sign);
        };
        SmallInteger.prototype.next = function () {
            var value = this.value;
            if (value + 1 < MAX_INT) return new SmallInteger(value + 1);
            return new BigInteger(MAX_INT_ARR, false);
        };
        NativeBigInt.prototype.next = function () {
            return new NativeBigInt(this.value + BigInt(1));
        };

        BigInteger.prototype.prev = function () {
            var value = this.value;
            if (this.sign) {
                return new BigInteger(addSmall(value, 1), true);
            }
            return subtractSmall(value, 1, this.sign);
        };
        SmallInteger.prototype.prev = function () {
            var value = this.value;
            if (value - 1 > -MAX_INT) return new SmallInteger(value - 1);
            return new BigInteger(MAX_INT_ARR, true);
        };
        NativeBigInt.prototype.prev = function () {
            return new NativeBigInt(this.value - BigInt(1));
        };

        var powersOfTwo = [1];
        while (2 * powersOfTwo[powersOfTwo.length - 1] <= BASE) powersOfTwo.push(2 * powersOfTwo[powersOfTwo.length - 1]);
        var powers2Length = powersOfTwo.length, highestPower2 = powersOfTwo[powers2Length - 1];

        function shift_isSmall(n) {
            return Math.abs(n) <= BASE;
        }

        BigInteger.prototype.shiftLeft = function (v) {
            var n = parseValue(v).toJSNumber();
            if (!shift_isSmall(n)) {
                throw new Error(String(n) + " is too large for shifting.");
            }
            if (n < 0) return this.shiftRight(-n);
            var result = this;
            if (result.isZero()) return result;
            while (n >= powers2Length) {
                result = result.multiply(highestPower2);
                n -= powers2Length - 1;
            }
            return result.multiply(powersOfTwo[n]);
        };
        NativeBigInt.prototype.shiftLeft = SmallInteger.prototype.shiftLeft = BigInteger.prototype.shiftLeft;

        BigInteger.prototype.shiftRight = function (v) {
            var remQuo;
            var n = parseValue(v).toJSNumber();
            if (!shift_isSmall(n)) {
                throw new Error(String(n) + " is too large for shifting.");
            }
            if (n < 0) return this.shiftLeft(-n);
            var result = this;
            while (n >= powers2Length) {
                if (result.isZero() || (result.isNegative() && result.isUnit())) return result;
                remQuo = divModAny(result, highestPower2);
                result = remQuo[1].isNegative() ? remQuo[0].prev() : remQuo[0];
                n -= powers2Length - 1;
            }
            remQuo = divModAny(result, powersOfTwo[n]);
            return remQuo[1].isNegative() ? remQuo[0].prev() : remQuo[0];
        };
        NativeBigInt.prototype.shiftRight = SmallInteger.prototype.shiftRight = BigInteger.prototype.shiftRight;

        function bitwise(x, y, fn) {
            y = parseValue(y);
            var xSign = x.isNegative(), ySign = y.isNegative();
            var xRem = xSign ? x.not() : x,
                yRem = ySign ? y.not() : y;
            var xDigit = 0, yDigit = 0;
            var xDivMod = null, yDivMod = null;
            var result = [];
            while (!xRem.isZero() || !yRem.isZero()) {
                xDivMod = divModAny(xRem, highestPower2);
                xDigit = xDivMod[1].toJSNumber();
                if (xSign) {
                    xDigit = highestPower2 - 1 - xDigit; // two's complement for negative numbers
                }

                yDivMod = divModAny(yRem, highestPower2);
                yDigit = yDivMod[1].toJSNumber();
                if (ySign) {
                    yDigit = highestPower2 - 1 - yDigit; // two's complement for negative numbers
                }

                xRem = xDivMod[0];
                yRem = yDivMod[0];
                result.push(fn(xDigit, yDigit));
            }
            var sum = fn(xSign ? 1 : 0, ySign ? 1 : 0) !== 0 ? bigInt(-1) : bigInt(0);
            for (var i = result.length - 1; i >= 0; i -= 1) {
                sum = sum.multiply(highestPower2).add(bigInt(result[i]));
            }
            return sum;
        }

        BigInteger.prototype.not = function () {
            return this.negate().prev();
        };
        NativeBigInt.prototype.not = SmallInteger.prototype.not = BigInteger.prototype.not;

        BigInteger.prototype.and = function (n) {
            return bitwise(this, n, function (a, b) { return a & b; });
        };
        NativeBigInt.prototype.and = SmallInteger.prototype.and = BigInteger.prototype.and;

        BigInteger.prototype.or = function (n) {
            return bitwise(this, n, function (a, b) { return a | b; });
        };
        NativeBigInt.prototype.or = SmallInteger.prototype.or = BigInteger.prototype.or;

        BigInteger.prototype.xor = function (n) {
            return bitwise(this, n, function (a, b) { return a ^ b; });
        };
        NativeBigInt.prototype.xor = SmallInteger.prototype.xor = BigInteger.prototype.xor;

        var LOBMASK_I = 1 << 30, LOBMASK_BI = (BASE & -BASE) * (BASE & -BASE) | LOBMASK_I;
        function roughLOB(n) { // get lowestOneBit (rough)
            // SmallInteger: return Min(lowestOneBit(n), 1 << 30)
            // BigInteger: return Min(lowestOneBit(n), 1 << 14) [BASE=1e7]
            var v = n.value,
                x = typeof v === "number" ? v | LOBMASK_I :
                    typeof v === "bigint" ? v | BigInt(LOBMASK_I) :
                        v[0] + v[1] * BASE | LOBMASK_BI;
            return x & -x;
        }

        function integerLogarithm(value, base) {
            if (base.compareTo(value) <= 0) {
                var tmp = integerLogarithm(value, base.square(base));
                var p = tmp.p;
                var e = tmp.e;
                var t = p.multiply(base);
                return t.compareTo(value) <= 0 ? { p: t, e: e * 2 + 1 } : { p: p, e: e * 2 };
            }
            return { p: bigInt(1), e: 0 };
        }

        BigInteger.prototype.bitLength = function () {
            var n = this;
            if (n.compareTo(bigInt(0)) < 0) {
                n = n.negate().subtract(bigInt(1));
            }
            if (n.compareTo(bigInt(0)) === 0) {
                return bigInt(0);
            }
            return bigInt(integerLogarithm(n, bigInt(2)).e).add(bigInt(1));
        };
        NativeBigInt.prototype.bitLength = SmallInteger.prototype.bitLength = BigInteger.prototype.bitLength;

        function max(a, b) {
            a = parseValue(a);
            b = parseValue(b);
            return a.greater(b) ? a : b;
        }
        function min(a, b) {
            a = parseValue(a);
            b = parseValue(b);
            return a.lesser(b) ? a : b;
        }
        function gcd(a, b) {
            a = parseValue(a).abs();
            b = parseValue(b).abs();
            if (a.equals(b)) return a;
            if (a.isZero()) return b;
            if (b.isZero()) return a;
            var c = Integer[1], d, t;
            while (a.isEven() && b.isEven()) {
                d = min(roughLOB(a), roughLOB(b));
                a = a.divide(d);
                b = b.divide(d);
                c = c.multiply(d);
            }
            while (a.isEven()) {
                a = a.divide(roughLOB(a));
            }
            do {
                while (b.isEven()) {
                    b = b.divide(roughLOB(b));
                }
                if (a.greater(b)) {
                    t = b; b = a; a = t;
                }
                b = b.subtract(a);
            } while (!b.isZero());
            return c.isUnit() ? a : a.multiply(c);
        }
        function lcm(a, b) {
            a = parseValue(a).abs();
            b = parseValue(b).abs();
            return a.divide(gcd(a, b)).multiply(b);
        }
        function randBetween(a, b, rng) {
            a = parseValue(a);
            b = parseValue(b);
            var usedRNG = rng || Math.random;
            var low = min(a, b), high = max(a, b);
            var range = high.subtract(low).add(1);
            if (range.isSmall) return low.add(Math.floor(usedRNG() * range));
            var digits = toBase(range, BASE).value;
            var result = [], restricted = true;
            for (var i = 0; i < digits.length; i++) {
                var top = restricted ? digits[i] : BASE;
                var digit = truncate(usedRNG() * top);
                result.push(digit);
                if (digit < top) restricted = false;
            }
            return low.add(Integer.fromArray(result, BASE, false));
        }

        var parseBase = function (text, base, alphabet, caseSensitive) {
            alphabet = alphabet || DEFAULT_ALPHABET;
            text = String(text);
            if (!caseSensitive) {
                text = text.toLowerCase();
                alphabet = alphabet.toLowerCase();
            }
            var length = text.length;
            var i;
            var absBase = Math.abs(base);
            var alphabetValues = {};
            for (i = 0; i < alphabet.length; i++) {
                alphabetValues[alphabet[i]] = i;
            }
            for (i = 0; i < length; i++) {
                var c = text[i];
                if (c === "-") continue;
                if (c in alphabetValues) {
                    if (alphabetValues[c] >= absBase) {
                        if (c === "1" && absBase === 1) continue;
                        throw new Error(c + " is not a valid digit in base " + base + ".");
                    }
                }
            }
            base = parseValue(base);
            var digits = [];
            var isNegative = text[0] === "-";
            for (i = isNegative ? 1 : 0; i < text.length; i++) {
                var c = text[i];
                if (c in alphabetValues) digits.push(parseValue(alphabetValues[c]));
                else if (c === "<") {
                    var start = i;
                    do { i++; } while (text[i] !== ">" && i < text.length);
                    digits.push(parseValue(text.slice(start + 1, i)));
                }
                else throw new Error(c + " is not a valid character");
            }
            return parseBaseFromArray(digits, base, isNegative);
        };

        function parseBaseFromArray(digits, base, isNegative) {
            var val = Integer[0], pow = Integer[1], i;
            for (i = digits.length - 1; i >= 0; i--) {
                val = val.add(digits[i].times(pow));
                pow = pow.times(base);
            }
            return isNegative ? val.negate() : val;
        }

        function stringify(digit, alphabet) {
            alphabet = alphabet || DEFAULT_ALPHABET;
            if (digit < alphabet.length) {
                return alphabet[digit];
            }
            return "<" + digit + ">";
        }

        function toBase(n, base) {
            base = bigInt(base);
            if (base.isZero()) {
                if (n.isZero()) return { value: [0], isNegative: false };
                throw new Error("Cannot convert nonzero numbers to base 0.");
            }
            if (base.equals(-1)) {
                if (n.isZero()) return { value: [0], isNegative: false };
                if (n.isNegative())
                    return {
                        value: [].concat.apply([], Array.apply(null, Array(-n.toJSNumber()))
                            .map(Array.prototype.valueOf, [1, 0])
                        ),
                        isNegative: false
                    };

                var arr = Array.apply(null, Array(n.toJSNumber() - 1))
                    .map(Array.prototype.valueOf, [0, 1]);
                arr.unshift([1]);
                return {
                    value: [].concat.apply([], arr),
                    isNegative: false
                };
            }

            var neg = false;
            if (n.isNegative() && base.isPositive()) {
                neg = true;
                n = n.abs();
            }
            if (base.isUnit()) {
                if (n.isZero()) return { value: [0], isNegative: false };

                return {
                    value: Array.apply(null, Array(n.toJSNumber()))
                        .map(Number.prototype.valueOf, 1),
                    isNegative: neg
                };
            }
            var out = [];
            var left = n, divmod;
            while (left.isNegative() || left.compareAbs(base) >= 0) {
                divmod = left.divmod(base);
                left = divmod.quotient;
                var digit = divmod.remainder;
                if (digit.isNegative()) {
                    digit = base.minus(digit).abs();
                    left = left.next();
                }
                out.push(digit.toJSNumber());
            }
            out.push(left.toJSNumber());
            return { value: out.reverse(), isNegative: neg };
        }

        function toBaseString(n, base, alphabet) {
            var arr = toBase(n, base);
            return (arr.isNegative ? "-" : "") + arr.value.map(function (x) {
                return stringify(x, alphabet);
            }).join('');
        }

        BigInteger.prototype.toArray = function (radix) {
            return toBase(this, radix);
        };

        SmallInteger.prototype.toArray = function (radix) {
            return toBase(this, radix);
        };

        NativeBigInt.prototype.toArray = function (radix) {
            return toBase(this, radix);
        };

        BigInteger.prototype.toString = function (radix, alphabet) {
            if (radix === undefined$1) radix = 10;
            if (radix !== 10) return toBaseString(this, radix, alphabet);
            var v = this.value, l = v.length, str = String(v[--l]), zeros = "0000000", digit;
            while (--l >= 0) {
                digit = String(v[l]);
                str += zeros.slice(digit.length) + digit;
            }
            var sign = this.sign ? "-" : "";
            return sign + str;
        };

        SmallInteger.prototype.toString = function (radix, alphabet) {
            if (radix === undefined$1) radix = 10;
            if (radix != 10) return toBaseString(this, radix, alphabet);
            return String(this.value);
        };

        NativeBigInt.prototype.toString = SmallInteger.prototype.toString;

        NativeBigInt.prototype.toJSON = BigInteger.prototype.toJSON = SmallInteger.prototype.toJSON = function () { return this.toString(); };

        BigInteger.prototype.valueOf = function () {
            return parseInt(this.toString(), 10);
        };
        BigInteger.prototype.toJSNumber = BigInteger.prototype.valueOf;

        SmallInteger.prototype.valueOf = function () {
            return this.value;
        };
        SmallInteger.prototype.toJSNumber = SmallInteger.prototype.valueOf;
        NativeBigInt.prototype.valueOf = NativeBigInt.prototype.toJSNumber = function () {
            return parseInt(this.toString(), 10);
        };

        function parseStringValue(v) {
            if (isPrecise(+v)) {
                var x = +v;
                if (x === truncate(x))
                    return supportsNativeBigInt ? new NativeBigInt(BigInt(x)) : new SmallInteger(x);
                throw new Error("Invalid integer: " + v);
            }
            var sign = v[0] === "-";
            if (sign) v = v.slice(1);
            var split = v.split(/e/i);
            if (split.length > 2) throw new Error("Invalid integer: " + split.join("e"));
            if (split.length === 2) {
                var exp = split[1];
                if (exp[0] === "+") exp = exp.slice(1);
                exp = +exp;
                if (exp !== truncate(exp) || !isPrecise(exp)) throw new Error("Invalid integer: " + exp + " is not a valid exponent.");
                var text = split[0];
                var decimalPlace = text.indexOf(".");
                if (decimalPlace >= 0) {
                    exp -= text.length - decimalPlace - 1;
                    text = text.slice(0, decimalPlace) + text.slice(decimalPlace + 1);
                }
                if (exp < 0) throw new Error("Cannot include negative exponent part for integers");
                text += (new Array(exp + 1)).join("0");
                v = text;
            }
            var isValid = /^([0-9][0-9]*)$/.test(v);
            if (!isValid) throw new Error("Invalid integer: " + v);
            if (supportsNativeBigInt) {
                return new NativeBigInt(BigInt(sign ? "-" + v : v));
            }
            var r = [], max = v.length, l = LOG_BASE, min = max - l;
            while (max > 0) {
                r.push(+v.slice(min, max));
                min -= l;
                if (min < 0) min = 0;
                max -= l;
            }
            trim(r);
            return new BigInteger(r, sign);
        }

        function parseNumberValue(v) {
            if (supportsNativeBigInt) {
                return new NativeBigInt(BigInt(v));
            }
            if (isPrecise(v)) {
                if (v !== truncate(v)) throw new Error(v + " is not an integer.");
                return new SmallInteger(v);
            }
            return parseStringValue(v.toString());
        }

        function parseValue(v) {
            if (typeof v === "number") {
                return parseNumberValue(v);
            }
            if (typeof v === "string") {
                return parseStringValue(v);
            }
            if (typeof v === "bigint") {
                return new NativeBigInt(v);
            }
            return v;
        }
        // Pre-define numbers in range [-999,999]
        for (var i = 0; i < 1000; i++) {
            Integer[i] = parseValue(i);
            if (i > 0) Integer[-i] = parseValue(-i);
        }
        // Backwards compatibility
        Integer.one = Integer[1];
        Integer.zero = Integer[0];
        Integer.minusOne = Integer[-1];
        Integer.max = max;
        Integer.min = min;
        Integer.gcd = gcd;
        Integer.lcm = lcm;
        Integer.isInstance = function (x) { return x instanceof BigInteger || x instanceof SmallInteger || x instanceof NativeBigInt; };
        Integer.randBetween = randBetween;

        Integer.fromArray = function (digits, base, isNegative) {
            return parseBaseFromArray(digits.map(parseValue), parseValue(base || 10), isNegative);
        };

        return Integer;
    })();

    // Node.js check
    if ( module.hasOwnProperty("exports")) {
        module.exports = bigInt;
    }
    });

    // SHA-256 for JavaScript.
    //
    // Written in 2014-2016 by Dmitry Chestnykh.
    // Public domain, no warranty.
    //
    // Functions (accept and return Uint8Arrays):
    //
    //   sha256(message) -> hash
    //   sha256.hmac(key, message) -> mac
    //
    //  Classes:
    //
    //   new sha256.Hash()
    const digestLength = 32;
    const blockSize = 64;
    // SHA-256 constants
    const K$1 = new Uint32Array([
        0x428a2f98,
        0x71374491,
        0xb5c0fbcf,
        0xe9b5dba5,
        0x3956c25b,
        0x59f111f1,
        0x923f82a4,
        0xab1c5ed5,
        0xd807aa98,
        0x12835b01,
        0x243185be,
        0x550c7dc3,
        0x72be5d74,
        0x80deb1fe,
        0x9bdc06a7,
        0xc19bf174,
        0xe49b69c1,
        0xefbe4786,
        0x0fc19dc6,
        0x240ca1cc,
        0x2de92c6f,
        0x4a7484aa,
        0x5cb0a9dc,
        0x76f988da,
        0x983e5152,
        0xa831c66d,
        0xb00327c8,
        0xbf597fc7,
        0xc6e00bf3,
        0xd5a79147,
        0x06ca6351,
        0x14292967,
        0x27b70a85,
        0x2e1b2138,
        0x4d2c6dfc,
        0x53380d13,
        0x650a7354,
        0x766a0abb,
        0x81c2c92e,
        0x92722c85,
        0xa2bfe8a1,
        0xa81a664b,
        0xc24b8b70,
        0xc76c51a3,
        0xd192e819,
        0xd6990624,
        0xf40e3585,
        0x106aa070,
        0x19a4c116,
        0x1e376c08,
        0x2748774c,
        0x34b0bcb5,
        0x391c0cb3,
        0x4ed8aa4a,
        0x5b9cca4f,
        0x682e6ff3,
        0x748f82ee,
        0x78a5636f,
        0x84c87814,
        0x8cc70208,
        0x90befffa,
        0xa4506ceb,
        0xbef9a3f7,
        0xc67178f2,
    ]);
    function hashBlocks(w, v, p, pos, len) {
        let a, b, c, d, e, f, g, h, u, i, j, t1, t2;
        while (len >= 64) {
            a = v[0];
            b = v[1];
            c = v[2];
            d = v[3];
            e = v[4];
            f = v[5];
            g = v[6];
            h = v[7];
            for (i = 0; i < 16; i++) {
                j = pos + i * 4;
                w[i] =
                    ((p[j] & 0xff) << 24) |
                        ((p[j + 1] & 0xff) << 16) |
                        ((p[j + 2] & 0xff) << 8) |
                        (p[j + 3] & 0xff);
            }
            for (i = 16; i < 64; i++) {
                u = w[i - 2];
                t1 =
                    ((u >>> 17) | (u << (32 - 17))) ^
                        ((u >>> 19) | (u << (32 - 19))) ^
                        (u >>> 10);
                u = w[i - 15];
                t2 =
                    ((u >>> 7) | (u << (32 - 7))) ^
                        ((u >>> 18) | (u << (32 - 18))) ^
                        (u >>> 3);
                w[i] = ((t1 + w[i - 7]) | 0) + ((t2 + w[i - 16]) | 0);
            }
            for (i = 0; i < 64; i++) {
                t1 =
                    ((((((e >>> 6) | (e << (32 - 6))) ^
                        ((e >>> 11) | (e << (32 - 11))) ^
                        ((e >>> 25) | (e << (32 - 25)))) +
                        ((e & f) ^ (~e & g))) |
                        0) +
                        ((h + ((K$1[i] + w[i]) | 0)) | 0)) |
                        0;
                t2 =
                    ((((a >>> 2) | (a << (32 - 2))) ^
                        ((a >>> 13) | (a << (32 - 13))) ^
                        ((a >>> 22) | (a << (32 - 22)))) +
                        ((a & b) ^ (a & c) ^ (b & c))) |
                        0;
                h = g;
                g = f;
                f = e;
                e = (d + t1) | 0;
                d = c;
                c = b;
                b = a;
                a = (t1 + t2) | 0;
            }
            v[0] += a;
            v[1] += b;
            v[2] += c;
            v[3] += d;
            v[4] += e;
            v[5] += f;
            v[6] += g;
            v[7] += h;
            pos += 64;
            len -= 64;
        }
        return pos;
    }
    // Hash implements SHA256 hash algorithm.
    class HashSha256 {
        constructor() {
            this.digestLength = digestLength;
            this.blockSize = blockSize;
            // Note: Int32Array is used instead of Uint32Array for performance reasons.
            this.state = new Int32Array(8); // hash state
            this.temp = new Int32Array(64); // temporary state
            this.buffer = new Uint8Array(128); // buffer for data to hash
            this.bufferLength = 0; // number of bytes in buffer
            this.bytesHashed = 0; // number of total bytes hashed
            this.finished = false; // indicates whether the hash was finalized
            this.reset();
        }
        // Resets hash state making it possible
        // to re-use this instance to hash other data.
        reset() {
            this.state[0] = 0x6a09e667;
            this.state[1] = 0xbb67ae85;
            this.state[2] = 0x3c6ef372;
            this.state[3] = 0xa54ff53a;
            this.state[4] = 0x510e527f;
            this.state[5] = 0x9b05688c;
            this.state[6] = 0x1f83d9ab;
            this.state[7] = 0x5be0cd19;
            this.bufferLength = 0;
            this.bytesHashed = 0;
            this.finished = false;
            return this;
        }
        // Cleans internal buffers and re-initializes hash state.
        clean() {
            for (let i = 0; i < this.buffer.length; i++) {
                this.buffer[i] = 0;
            }
            for (let i = 0; i < this.temp.length; i++) {
                this.temp[i] = 0;
            }
            this.reset();
        }
        // Updates hash state with the given data.
        //
        // Optionally, length of the data can be specified to hash
        // fewer bytes than data.length.
        //
        // Throws error when trying to update already finalized hash:
        // instance must be reset to use it again.
        update(data, dataLength = data.length) {
            if (this.finished) {
                throw new Error("SHA256: can't update because hash was finished.");
            }
            let dataPos = 0;
            this.bytesHashed += dataLength;
            if (this.bufferLength > 0) {
                while (this.bufferLength < 64 && dataLength > 0) {
                    this.buffer[this.bufferLength++] = data[dataPos++];
                    dataLength--;
                }
                if (this.bufferLength === 64) {
                    hashBlocks(this.temp, this.state, this.buffer, 0, 64);
                    this.bufferLength = 0;
                }
            }
            if (dataLength >= 64) {
                dataPos = hashBlocks(this.temp, this.state, data, dataPos, dataLength);
                dataLength %= 64;
            }
            while (dataLength > 0) {
                this.buffer[this.bufferLength++] = data[dataPos++];
                dataLength--;
            }
            return this;
        }
        // Finalizes hash state and puts hash into out.
        //
        // If hash was already finalized, puts the same value.
        finish(out) {
            if (!this.finished) {
                const bytesHashed = this.bytesHashed;
                const left = this.bufferLength;
                const bitLenHi = (bytesHashed / 0x20000000) | 0;
                const bitLenLo = bytesHashed << 3;
                const padLength = bytesHashed % 64 < 56 ? 64 : 128;
                this.buffer[left] = 0x80;
                for (let i = left + 1; i < padLength - 8; i++) {
                    this.buffer[i] = 0;
                }
                this.buffer[padLength - 8] = (bitLenHi >>> 24) & 0xff;
                this.buffer[padLength - 7] = (bitLenHi >>> 16) & 0xff;
                this.buffer[padLength - 6] = (bitLenHi >>> 8) & 0xff;
                this.buffer[padLength - 5] = (bitLenHi >>> 0) & 0xff;
                this.buffer[padLength - 4] = (bitLenLo >>> 24) & 0xff;
                this.buffer[padLength - 3] = (bitLenLo >>> 16) & 0xff;
                this.buffer[padLength - 2] = (bitLenLo >>> 8) & 0xff;
                this.buffer[padLength - 1] = (bitLenLo >>> 0) & 0xff;
                hashBlocks(this.temp, this.state, this.buffer, 0, padLength);
                this.finished = true;
            }
            for (let i = 0; i < 8; i++) {
                out[i * 4 + 0] = (this.state[i] >>> 24) & 0xff;
                out[i * 4 + 1] = (this.state[i] >>> 16) & 0xff;
                out[i * 4 + 2] = (this.state[i] >>> 8) & 0xff;
                out[i * 4 + 3] = (this.state[i] >>> 0) & 0xff;
            }
            return this;
        }
        // Returns the final hash digest.
        digest() {
            const out = new Uint8Array(this.digestLength);
            this.finish(out);
            return out;
        }
        // Internal function for use in HMAC for optimization.
        _saveState(out) {
            for (let i = 0; i < this.state.length; i++) {
                out[i] = this.state[i];
            }
        }
        // Internal function for use in HMAC for optimization.
        _restoreState(from, bytesHashed) {
            for (let i = 0; i < this.state.length; i++) {
                this.state[i] = from[i];
            }
            this.bytesHashed = bytesHashed;
            this.finished = false;
            this.bufferLength = 0;
        }
    }
    // Returns SHA256 hash of data.
    function sha256(data) {
        const h = new HashSha256().update(data);
        const digest = h.digest();
        h.clean();
        return digest;
    }

    /*
     This file is part of GNU Taler
     (C) 2019 GNUnet e.V.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    function sha512(data) {
        return hash(data);
    }
    function hmac(digest, blockSize, key, message) {
        if (key.byteLength > blockSize) {
            key = digest(key);
        }
        if (key.byteLength < blockSize) {
            const k = key;
            key = new Uint8Array(blockSize);
            key.set(k, 0);
        }
        const okp = new Uint8Array(blockSize);
        const ikp = new Uint8Array(blockSize);
        for (let i = 0; i < blockSize; i++) {
            ikp[i] = key[i] ^ 0x36;
            okp[i] = key[i] ^ 0x5c;
        }
        const b1 = new Uint8Array(blockSize + message.byteLength);
        b1.set(ikp, 0);
        b1.set(message, blockSize);
        const h0 = digest(b1);
        const b2 = new Uint8Array(blockSize + h0.length);
        b2.set(okp, 0);
        b2.set(h0, blockSize);
        return digest(b2);
    }
    function hmacSha512(key, message) {
        return hmac(sha512, 128, key, message);
    }
    function hmacSha256(key, message) {
        return hmac(sha256, 64, key, message);
    }
    function kdf(outputLength, ikm, salt, info) {
        // extract
        const prk = hmacSha512(salt, ikm);
        // expand
        const N = Math.ceil(outputLength / 32);
        const output = new Uint8Array(N * 32);
        for (let i = 0; i < N; i++) {
            let buf;
            if (i == 0) {
                buf = new Uint8Array(info.byteLength + 1);
                buf.set(info, 0);
            }
            else {
                buf = new Uint8Array(info.byteLength + 1 + 32);
                for (let j = 0; j < 32; j++) {
                    buf[j] = output[(i - 1) * 32 + j];
                }
                buf.set(info, 32);
            }
            buf[buf.length - 1] = i + 1;
            const chunk = hmacSha256(prk, buf);
            output.set(chunk, i * 32);
        }
        return output.slice(0, outputLength);
    }

    /*
     This file is part of GNU Taler
     (C) 2019 GNUnet e.V.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    // @ts-ignore
    const decoder = new TextDecoder();
    if (typeof decoder !== "object") {
        throw Error("FATAL: TextDecoder not available");
    }
    // @ts-ignore
    const encoder = new TextEncoder();
    if (typeof encoder !== "object") {
        throw Error("FATAL: TextEncoder not available");
    }
    const encTable = "0123456789ABCDEFGHJKMNPQRSTVWXYZ";
    class EncodingError extends Error {
        constructor() {
            super("Encoding error");
            Object.setPrototypeOf(this, EncodingError.prototype);
        }
    }
    function getValue(chr) {
        let a = chr;
        switch (chr) {
            case "O":
            case "o":
                a = "0;";
                break;
            case "i":
            case "I":
            case "l":
            case "L":
                a = "1";
                break;
            case "u":
            case "U":
                a = "V";
        }
        if (a >= "0" && a <= "9") {
            return a.charCodeAt(0) - "0".charCodeAt(0);
        }
        if (a >= "a" && a <= "z")
            a = a.toUpperCase();
        let dec = 0;
        if (a >= "A" && a <= "Z") {
            if ("I" < a)
                dec++;
            if ("L" < a)
                dec++;
            if ("O" < a)
                dec++;
            if ("U" < a)
                dec++;
            return a.charCodeAt(0) - "A".charCodeAt(0) + 10 - dec;
        }
        throw new EncodingError();
    }
    function encodeCrock(data) {
        const dataBytes = new Uint8Array(data);
        let sb = "";
        const size = data.byteLength;
        let bitBuf = 0;
        let numBits = 0;
        let pos = 0;
        while (pos < size || numBits > 0) {
            if (pos < size && numBits < 5) {
                const d = dataBytes[pos++];
                bitBuf = (bitBuf << 8) | d;
                numBits += 8;
            }
            if (numBits < 5) {
                // zero-padding
                bitBuf = bitBuf << (5 - numBits);
                numBits = 5;
            }
            const v = (bitBuf >>> (numBits - 5)) & 31;
            sb += encTable[v];
            numBits -= 5;
        }
        return sb;
    }
    function decodeCrock(encoded) {
        const size = encoded.length;
        let bitpos = 0;
        let bitbuf = 0;
        let readPosition = 0;
        const outLen = Math.floor((size * 5) / 8);
        const out = new Uint8Array(outLen);
        let outPos = 0;
        while (readPosition < size || bitpos > 0) {
            if (readPosition < size) {
                const v = getValue(encoded[readPosition++]);
                bitbuf = (bitbuf << 5) | v;
                bitpos += 5;
            }
            while (bitpos >= 8) {
                const d = (bitbuf >>> (bitpos - 8)) & 0xff;
                out[outPos++] = d;
                bitpos -= 8;
            }
            if (readPosition == size && bitpos > 0) {
                bitbuf = (bitbuf << (8 - bitpos)) & 0xff;
                bitpos = bitbuf == 0 ? 0 : 8;
            }
        }
        return out;
    }
    function eddsaGetPublic(eddsaPriv) {
        const pair = sign_keyPair_fromSeed(eddsaPriv);
        return pair.publicKey;
    }
    function ecdheGetPublic(ecdhePriv) {
        return scalarMult_base(ecdhePriv);
    }
    function keyExchangeEcdheEddsa(ecdhePriv, eddsaPub) {
        const curve25519Pub = sign_ed25519_pk_to_curve25519(eddsaPub);
        const x = scalarMult(ecdhePriv, curve25519Pub);
        return hash(x);
    }
    /**
     * KDF modulo a big integer.
     */
    function kdfMod(n, ikm, salt, info) {
        const nbits = n.bitLength().toJSNumber();
        const buflen = Math.floor((nbits - 1) / 8 + 1);
        const mask = (1 << (8 - (buflen * 8 - nbits))) - 1;
        let counter = 0;
        while (true) {
            const ctx = new Uint8Array(info.byteLength + 2);
            ctx.set(info, 0);
            ctx[ctx.length - 2] = (counter >>> 8) & 0xff;
            ctx[ctx.length - 1] = counter & 0xff;
            const buf = kdf(buflen, ikm, salt, ctx);
            const arr = Array.from(buf);
            arr[0] = arr[0] & mask;
            const r = BigInteger.fromArray(arr, 256, false);
            if (r.lt(n)) {
                return r;
            }
            counter++;
        }
    }
    function stringToBytes(s) {
        return encoder.encode(s);
    }
    function loadBigInt(arr) {
        return BigInteger.fromArray(Array.from(arr), 256, false);
    }
    function rsaBlindingKeyDerive(rsaPub, bks) {
        const salt = stringToBytes("Blinding KDF extractor HMAC key");
        const info = stringToBytes("Blinding KDF");
        return kdfMod(rsaPub.N, bks, salt, info);
    }
    /*
     * Test for malicious RSA key.
     *
     * Assuming n is an RSA modulous and r is generated using a call to
     * GNUNET_CRYPTO_kdf_mod_mpi, if gcd(r,n) != 1 then n must be a
     * malicious RSA key designed to deanomize the user.
     *
     * @param r KDF result
     * @param n RSA modulus of the public key
     */
    function rsaGcdValidate(r, n) {
        const t = BigInteger.gcd(r, n);
        if (!t.equals(BigInteger.one)) {
            throw Error("malicious RSA public key");
        }
    }
    function rsaFullDomainHash(hm, rsaPub) {
        const info = stringToBytes("RSA-FDA FTpsW!");
        const salt = rsaPubEncode(rsaPub);
        const r = kdfMod(rsaPub.N, hm, salt, info);
        rsaGcdValidate(r, rsaPub.N);
        return r;
    }
    function rsaPubDecode(rsaPub) {
        const modulusLength = (rsaPub[0] << 8) | rsaPub[1];
        const exponentLength = (rsaPub[2] << 8) | rsaPub[3];
        if (4 + exponentLength + modulusLength != rsaPub.length) {
            throw Error("invalid RSA public key (format wrong)");
        }
        const modulus = rsaPub.slice(4, 4 + modulusLength);
        const exponent = rsaPub.slice(4 + modulusLength, 4 + modulusLength + exponentLength);
        const res = {
            N: loadBigInt(modulus),
            e: loadBigInt(exponent),
        };
        return res;
    }
    function rsaPubEncode(rsaPub) {
        const mb = rsaPub.N.toArray(256).value;
        const eb = rsaPub.e.toArray(256).value;
        const out = new Uint8Array(4 + mb.length + eb.length);
        out[0] = (mb.length >>> 8) & 0xff;
        out[1] = mb.length & 0xff;
        out[2] = (eb.length >>> 8) & 0xff;
        out[3] = eb.length & 0xff;
        out.set(mb, 4);
        out.set(eb, 4 + mb.length);
        return out;
    }
    function rsaBlind(hm, bks, rsaPubEnc) {
        const rsaPub = rsaPubDecode(rsaPubEnc);
        const data = rsaFullDomainHash(hm, rsaPub);
        const r = rsaBlindingKeyDerive(rsaPub, bks);
        const r_e = r.modPow(rsaPub.e, rsaPub.N);
        const bm = r_e.multiply(data).mod(rsaPub.N);
        return new Uint8Array(bm.toArray(256).value);
    }
    function rsaUnblind(sig, rsaPubEnc, bks) {
        const rsaPub = rsaPubDecode(rsaPubEnc);
        const blinded_s = loadBigInt(sig);
        const r = rsaBlindingKeyDerive(rsaPub, bks);
        const r_inv = r.modInv(rsaPub.N);
        const s = blinded_s.multiply(r_inv).mod(rsaPub.N);
        return new Uint8Array(s.toArray(256).value);
    }
    function rsaVerify(hm, rsaSig, rsaPubEnc) {
        const rsaPub = rsaPubDecode(rsaPubEnc);
        const d = rsaFullDomainHash(hm, rsaPub);
        const sig = loadBigInt(rsaSig);
        const sig_e = sig.modPow(rsaPub.e, rsaPub.N);
        return sig_e.equals(d);
    }
    function createEddsaKeyPair() {
        const eddsaPriv = randomBytes(32);
        const eddsaPub = eddsaGetPublic(eddsaPriv);
        return { eddsaPriv, eddsaPub };
    }
    function hash$1(d) {
        return hash(d);
    }
    function eddsaSign(msg, eddsaPriv) {
        const pair = sign_keyPair_fromSeed(eddsaPriv);
        return sign_detached(msg, pair.secretKey);
    }
    function eddsaVerify(msg, sig, eddsaPub) {
        return sign_detached_verify(msg, sig, eddsaPub);
    }
    function createHashContext() {
        return new HashState();
    }
    function setupRefreshPlanchet(secretSeed, coinNumber) {
        const info = stringToBytes("taler-coin-derivation");
        const saltArrBuf = new ArrayBuffer(4);
        const salt = new Uint8Array(saltArrBuf);
        const saltDataView = new DataView(saltArrBuf);
        saltDataView.setUint32(0, coinNumber);
        const out = kdf(64, secretSeed, salt, info);
        const coinPriv = out.slice(0, 32);
        const bks = out.slice(32, 64);
        return {
            bks,
            coinPriv,
            coinPub: eddsaGetPublic(coinPriv),
        };
    }
    function setupWithdrawPlanchet(secretSeed, coinNumber) {
        const info = stringToBytes("taler-withdrawal-coin-derivation");
        const saltArrBuf = new ArrayBuffer(4);
        const salt = new Uint8Array(saltArrBuf);
        const saltDataView = new DataView(saltArrBuf);
        saltDataView.setUint32(0, coinNumber);
        const out = kdf(64, secretSeed, salt, info);
        const coinPriv = out.slice(0, 32);
        const bks = out.slice(32, 64);
        return {
            bks,
            coinPriv,
            coinPub: eddsaGetPublic(coinPriv),
        };
    }
    function setupTipPlanchet(secretSeed, coinNumber) {
        const info = stringToBytes("taler-tip-coin-derivation");
        const saltArrBuf = new ArrayBuffer(4);
        const salt = new Uint8Array(saltArrBuf);
        const saltDataView = new DataView(saltArrBuf);
        saltDataView.setUint32(0, coinNumber);
        const out = kdf(64, secretSeed, salt, info);
        const coinPriv = out.slice(0, 32);
        const bks = out.slice(32, 64);
        return {
            bks,
            coinPriv,
            coinPub: eddsaGetPublic(coinPriv),
        };
    }
    function setupRefreshTransferPub(secretSeed, transferPubIndex) {
        const info = stringToBytes("taler-transfer-pub-derivation");
        const saltArrBuf = new ArrayBuffer(4);
        const salt = new Uint8Array(saltArrBuf);
        const saltDataView = new DataView(saltArrBuf);
        saltDataView.setUint32(0, transferPubIndex);
        const out = kdf(32, secretSeed, salt, info);
        return {
            ecdhePriv: out,
            ecdhePub: ecdheGetPublic(out),
        };
    }

    /*
     This file is part of GNU Taler
     (C) 2019 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var ReserveTransactionType;
    (function (ReserveTransactionType) {
        ReserveTransactionType["Withdraw"] = "WITHDRAW";
        ReserveTransactionType["Credit"] = "CREDIT";
        ReserveTransactionType["Recoup"] = "RECOUP";
        ReserveTransactionType["Closing"] = "CLOSING";
    })(ReserveTransactionType || (ReserveTransactionType = {}));

    /*
     This file is part of TALER
     (C) 2016 GNUnet e.V.

     TALER is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     TALER is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var HttpResponseStatus;
    (function (HttpResponseStatus) {
        HttpResponseStatus[HttpResponseStatus["Ok"] = 200] = "Ok";
        HttpResponseStatus[HttpResponseStatus["Gone"] = 210] = "Gone";
        HttpResponseStatus[HttpResponseStatus["PaymentRequired"] = 402] = "PaymentRequired";
    })(HttpResponseStatus || (HttpResponseStatus = {}));

    /*
     This file is part of GNU Taler
     (C) 2017-2019 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    /**
     * Get a performance counter in milliseconds.
     */
    const performanceNow = (() => {
        // @ts-ignore
        if (typeof process !== "undefined" && process.hrtime) {
            return () => {
                const t = process.hrtime();
                return t[0] * 1e9 + t[1];
            };
        }
        // @ts-ignore
        if (typeof performance !== "undefined") {
            // @ts-ignore
            return () => performance.now();
        }
        return () => 0;
    })();

    /*
     This file is part of GNU Taler
     (C) 2019 GNUnet e.V.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    var PendingOperationType;
    (function (PendingOperationType) {
        PendingOperationType["Bug"] = "bug";
        PendingOperationType["ExchangeUpdate"] = "exchange-update";
        PendingOperationType["ExchangeCheckRefresh"] = "exchange-check-refresh";
        PendingOperationType["Pay"] = "pay";
        PendingOperationType["ProposalChoice"] = "proposal-choice";
        PendingOperationType["ProposalDownload"] = "proposal-download";
        PendingOperationType["Refresh"] = "refresh";
        PendingOperationType["Reserve"] = "reserve";
        PendingOperationType["Recoup"] = "recoup";
        PendingOperationType["RefundQuery"] = "refund-query";
        PendingOperationType["TipChoice"] = "tip-choice";
        PendingOperationType["TipPickup"] = "tip-pickup";
        PendingOperationType["Withdraw"] = "withdraw";
    })(PendingOperationType || (PendingOperationType = {}));
    /**
     * Current state of an exchange update operation.
     */
    var ExchangeUpdateOperationStage;
    (function (ExchangeUpdateOperationStage) {
        ExchangeUpdateOperationStage["FetchKeys"] = "fetch-keys";
        ExchangeUpdateOperationStage["FetchWire"] = "fetch-wire";
        ExchangeUpdateOperationStage["FinalizeUpdate"] = "finalize-update";
    })(ExchangeUpdateOperationStage || (ExchangeUpdateOperationStage = {}));
    var ReserveType;
    (function (ReserveType) {
        /**
         * Manually created.
         */
        ReserveType["Manual"] = "manual";
        /**
         * Withdrawn from a bank that has "tight" Taler integration
         */
        ReserveType["TalerBankWithdraw"] = "taler-bank-withdraw";
    })(ReserveType || (ReserveType = {}));

    /*
     This file is part of GNU Taler
     (C) 2020 Taler Systems S.A.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    /**
     * Types of coin sources.
     */
    var BackupCoinSourceType;
    (function (BackupCoinSourceType) {
        BackupCoinSourceType["Withdraw"] = "withdraw";
        BackupCoinSourceType["Refresh"] = "refresh";
        BackupCoinSourceType["Tip"] = "tip";
    })(BackupCoinSourceType || (BackupCoinSourceType = {}));
    /**
     * Reasons for why a coin is being refreshed.
     */
    var BackupRefreshReason;
    (function (BackupRefreshReason) {
        BackupRefreshReason["Manual"] = "manual";
        BackupRefreshReason["Pay"] = "pay";
        BackupRefreshReason["Refund"] = "refund";
        BackupRefreshReason["AbortPay"] = "abort-pay";
        BackupRefreshReason["Recoup"] = "recoup";
        BackupRefreshReason["BackupRestored"] = "backup-restored";
        BackupRefreshReason["Scheduled"] = "scheduled";
    })(BackupRefreshReason || (BackupRefreshReason = {}));
    var BackupRefundState;
    (function (BackupRefundState) {
        BackupRefundState["Failed"] = "failed";
        BackupRefundState["Applied"] = "applied";
        BackupRefundState["Pending"] = "pending";
    })(BackupRefundState || (BackupRefundState = {}));
    var BackupProposalStatus;
    (function (BackupProposalStatus) {
        /**
         * Proposed (and either downloaded or not,
         * depending on whether contract terms are present),
         * but the user needs to accept/reject it.
         */
        BackupProposalStatus["Proposed"] = "proposed";
        /**
         * The user has rejected the proposal.
         */
        BackupProposalStatus["Refused"] = "refused";
        /**
         * Downloading or processing the proposal has failed permanently.
         *
         * FIXME:  Should this be modeled as a "misbehavior report" instead?
         */
        BackupProposalStatus["PermanentlyFailed"] = "permanently-failed";
        /**
         * Downloaded proposal was detected as a re-purchase.
         */
        BackupProposalStatus["Repurchase"] = "repurchase";
    })(BackupProposalStatus || (BackupProposalStatus = {}));

    /*
     This file is part of GNU Taler
     (C) 2015-2019 GNUnet e.V.

     GNU Taler is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     GNU Taler is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     GNU Taler; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    const builtinCurrencies = [
        {
            auditors: [
                {
                    auditorPub: "BW9DC48PHQY4NH011SHHX36DZZ3Q22Y6X7FZ1VD1CMZ2PTFZ6PN0",
                    baseUrl: "https://auditor.demo.taler.net/",
                    expirationStamp: new Date(2027, 1).getTime(),
                },
            ],
            exchanges: [],
            fractionalDigits: 2,
            name: "KUDOS",
        },
    ];

    var bind = function bind(fn, thisArg) {
      return function wrap() {
        var args = new Array(arguments.length);
        for (var i = 0; i < args.length; i++) {
          args[i] = arguments[i];
        }
        return fn.apply(thisArg, args);
      };
    };

    /*global toString:true*/

    // utils is a library of generic helper functions non-specific to axios

    var toString = Object.prototype.toString;

    /**
     * Determine if a value is an Array
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an Array, otherwise false
     */
    function isArray(val) {
      return toString.call(val) === '[object Array]';
    }

    /**
     * Determine if a value is undefined
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if the value is undefined, otherwise false
     */
    function isUndefined(val) {
      return typeof val === 'undefined';
    }

    /**
     * Determine if a value is a Buffer
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Buffer, otherwise false
     */
    function isBuffer(val) {
      return val !== null && !isUndefined(val) && val.constructor !== null && !isUndefined(val.constructor)
        && typeof val.constructor.isBuffer === 'function' && val.constructor.isBuffer(val);
    }

    /**
     * Determine if a value is an ArrayBuffer
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an ArrayBuffer, otherwise false
     */
    function isArrayBuffer(val) {
      return toString.call(val) === '[object ArrayBuffer]';
    }

    /**
     * Determine if a value is a FormData
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an FormData, otherwise false
     */
    function isFormData(val) {
      return (typeof FormData !== 'undefined') && (val instanceof FormData);
    }

    /**
     * Determine if a value is a view on an ArrayBuffer
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a view on an ArrayBuffer, otherwise false
     */
    function isArrayBufferView(val) {
      var result;
      if ((typeof ArrayBuffer !== 'undefined') && (ArrayBuffer.isView)) {
        result = ArrayBuffer.isView(val);
      } else {
        result = (val) && (val.buffer) && (val.buffer instanceof ArrayBuffer);
      }
      return result;
    }

    /**
     * Determine if a value is a String
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a String, otherwise false
     */
    function isString(val) {
      return typeof val === 'string';
    }

    /**
     * Determine if a value is a Number
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Number, otherwise false
     */
    function isNumber(val) {
      return typeof val === 'number';
    }

    /**
     * Determine if a value is an Object
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is an Object, otherwise false
     */
    function isObject(val) {
      return val !== null && typeof val === 'object';
    }

    /**
     * Determine if a value is a plain Object
     *
     * @param {Object} val The value to test
     * @return {boolean} True if value is a plain Object, otherwise false
     */
    function isPlainObject(val) {
      if (toString.call(val) !== '[object Object]') {
        return false;
      }

      var prototype = Object.getPrototypeOf(val);
      return prototype === null || prototype === Object.prototype;
    }

    /**
     * Determine if a value is a Date
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Date, otherwise false
     */
    function isDate(val) {
      return toString.call(val) === '[object Date]';
    }

    /**
     * Determine if a value is a File
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a File, otherwise false
     */
    function isFile(val) {
      return toString.call(val) === '[object File]';
    }

    /**
     * Determine if a value is a Blob
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Blob, otherwise false
     */
    function isBlob(val) {
      return toString.call(val) === '[object Blob]';
    }

    /**
     * Determine if a value is a Function
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Function, otherwise false
     */
    function isFunction(val) {
      return toString.call(val) === '[object Function]';
    }

    /**
     * Determine if a value is a Stream
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a Stream, otherwise false
     */
    function isStream(val) {
      return isObject(val) && isFunction(val.pipe);
    }

    /**
     * Determine if a value is a URLSearchParams object
     *
     * @param {Object} val The value to test
     * @returns {boolean} True if value is a URLSearchParams object, otherwise false
     */
    function isURLSearchParams(val) {
      return typeof URLSearchParams !== 'undefined' && val instanceof URLSearchParams;
    }

    /**
     * Trim excess whitespace off the beginning and end of a string
     *
     * @param {String} str The String to trim
     * @returns {String} The String freed of excess whitespace
     */
    function trim(str) {
      return str.replace(/^\s*/, '').replace(/\s*$/, '');
    }

    /**
     * Determine if we're running in a standard browser environment
     *
     * This allows axios to run in a web worker, and react-native.
     * Both environments support XMLHttpRequest, but not fully standard globals.
     *
     * web workers:
     *  typeof window -> undefined
     *  typeof document -> undefined
     *
     * react-native:
     *  navigator.product -> 'ReactNative'
     * nativescript
     *  navigator.product -> 'NativeScript' or 'NS'
     */
    function isStandardBrowserEnv() {
      if (typeof navigator !== 'undefined' && (navigator.product === 'ReactNative' ||
                                               navigator.product === 'NativeScript' ||
                                               navigator.product === 'NS')) {
        return false;
      }
      return (
        typeof window !== 'undefined' &&
        typeof document !== 'undefined'
      );
    }

    /**
     * Iterate over an Array or an Object invoking a function for each item.
     *
     * If `obj` is an Array callback will be called passing
     * the value, index, and complete array for each item.
     *
     * If 'obj' is an Object callback will be called passing
     * the value, key, and complete object for each property.
     *
     * @param {Object|Array} obj The object to iterate
     * @param {Function} fn The callback to invoke for each item
     */
    function forEach(obj, fn) {
      // Don't bother if no value provided
      if (obj === null || typeof obj === 'undefined') {
        return;
      }

      // Force an array if not already something iterable
      if (typeof obj !== 'object') {
        /*eslint no-param-reassign:0*/
        obj = [obj];
      }

      if (isArray(obj)) {
        // Iterate over array values
        for (var i = 0, l = obj.length; i < l; i++) {
          fn.call(null, obj[i], i, obj);
        }
      } else {
        // Iterate over object keys
        for (var key in obj) {
          if (Object.prototype.hasOwnProperty.call(obj, key)) {
            fn.call(null, obj[key], key, obj);
          }
        }
      }
    }

    /**
     * Accepts varargs expecting each argument to be an object, then
     * immutably merges the properties of each object and returns result.
     *
     * When multiple objects contain the same key the later object in
     * the arguments list will take precedence.
     *
     * Example:
     *
     * ```js
     * var result = merge({foo: 123}, {foo: 456});
     * console.log(result.foo); // outputs 456
     * ```
     *
     * @param {Object} obj1 Object to merge
     * @returns {Object} Result of all merge properties
     */
    function merge(/* obj1, obj2, obj3, ... */) {
      var result = {};
      function assignValue(val, key) {
        if (isPlainObject(result[key]) && isPlainObject(val)) {
          result[key] = merge(result[key], val);
        } else if (isPlainObject(val)) {
          result[key] = merge({}, val);
        } else if (isArray(val)) {
          result[key] = val.slice();
        } else {
          result[key] = val;
        }
      }

      for (var i = 0, l = arguments.length; i < l; i++) {
        forEach(arguments[i], assignValue);
      }
      return result;
    }

    /**
     * Extends object a by mutably adding to it the properties of object b.
     *
     * @param {Object} a The object to be extended
     * @param {Object} b The object to copy properties from
     * @param {Object} thisArg The object to bind function to
     * @return {Object} The resulting value of object a
     */
    function extend(a, b, thisArg) {
      forEach(b, function assignValue(val, key) {
        if (thisArg && typeof val === 'function') {
          a[key] = bind(val, thisArg);
        } else {
          a[key] = val;
        }
      });
      return a;
    }

    /**
     * Remove byte order marker. This catches EF BB BF (the UTF-8 BOM)
     *
     * @param {string} content with BOM
     * @return {string} content value without BOM
     */
    function stripBOM(content) {
      if (content.charCodeAt(0) === 0xFEFF) {
        content = content.slice(1);
      }
      return content;
    }

    var utils = {
      isArray: isArray,
      isArrayBuffer: isArrayBuffer,
      isBuffer: isBuffer,
      isFormData: isFormData,
      isArrayBufferView: isArrayBufferView,
      isString: isString,
      isNumber: isNumber,
      isObject: isObject,
      isPlainObject: isPlainObject,
      isUndefined: isUndefined,
      isDate: isDate,
      isFile: isFile,
      isBlob: isBlob,
      isFunction: isFunction,
      isStream: isStream,
      isURLSearchParams: isURLSearchParams,
      isStandardBrowserEnv: isStandardBrowserEnv,
      forEach: forEach,
      merge: merge,
      extend: extend,
      trim: trim,
      stripBOM: stripBOM
    };

    function encode(val) {
      return encodeURIComponent(val).
        replace(/%3A/gi, ':').
        replace(/%24/g, '$').
        replace(/%2C/gi, ',').
        replace(/%20/g, '+').
        replace(/%5B/gi, '[').
        replace(/%5D/gi, ']');
    }

    /**
     * Build a URL by appending params to the end
     *
     * @param {string} url The base of the url (e.g., http://www.google.com)
     * @param {object} [params] The params to be appended
     * @returns {string} The formatted url
     */
    var buildURL = function buildURL(url, params, paramsSerializer) {
      /*eslint no-param-reassign:0*/
      if (!params) {
        return url;
      }

      var serializedParams;
      if (paramsSerializer) {
        serializedParams = paramsSerializer(params);
      } else if (utils.isURLSearchParams(params)) {
        serializedParams = params.toString();
      } else {
        var parts = [];

        utils.forEach(params, function serialize(val, key) {
          if (val === null || typeof val === 'undefined') {
            return;
          }

          if (utils.isArray(val)) {
            key = key + '[]';
          } else {
            val = [val];
          }

          utils.forEach(val, function parseValue(v) {
            if (utils.isDate(v)) {
              v = v.toISOString();
            } else if (utils.isObject(v)) {
              v = JSON.stringify(v);
            }
            parts.push(encode(key) + '=' + encode(v));
          });
        });

        serializedParams = parts.join('&');
      }

      if (serializedParams) {
        var hashmarkIndex = url.indexOf('#');
        if (hashmarkIndex !== -1) {
          url = url.slice(0, hashmarkIndex);
        }

        url += (url.indexOf('?') === -1 ? '?' : '&') + serializedParams;
      }

      return url;
    };

    function InterceptorManager() {
      this.handlers = [];
    }

    /**
     * Add a new interceptor to the stack
     *
     * @param {Function} fulfilled The function to handle `then` for a `Promise`
     * @param {Function} rejected The function to handle `reject` for a `Promise`
     *
     * @return {Number} An ID used to remove interceptor later
     */
    InterceptorManager.prototype.use = function use(fulfilled, rejected) {
      this.handlers.push({
        fulfilled: fulfilled,
        rejected: rejected
      });
      return this.handlers.length - 1;
    };

    /**
     * Remove an interceptor from the stack
     *
     * @param {Number} id The ID that was returned by `use`
     */
    InterceptorManager.prototype.eject = function eject(id) {
      if (this.handlers[id]) {
        this.handlers[id] = null;
      }
    };

    /**
     * Iterate over all the registered interceptors
     *
     * This method is particularly useful for skipping over any
     * interceptors that may have become `null` calling `eject`.
     *
     * @param {Function} fn The function to call for each interceptor
     */
    InterceptorManager.prototype.forEach = function forEach(fn) {
      utils.forEach(this.handlers, function forEachHandler(h) {
        if (h !== null) {
          fn(h);
        }
      });
    };

    var InterceptorManager_1 = InterceptorManager;

    /**
     * Transform the data for a request or a response
     *
     * @param {Object|String} data The data to be transformed
     * @param {Array} headers The headers for the request or response
     * @param {Array|Function} fns A single function or Array of functions
     * @returns {*} The resulting transformed data
     */
    var transformData = function transformData(data, headers, fns) {
      /*eslint no-param-reassign:0*/
      utils.forEach(fns, function transform(fn) {
        data = fn(data, headers);
      });

      return data;
    };

    var isCancel = function isCancel(value) {
      return !!(value && value.__CANCEL__);
    };

    var normalizeHeaderName = function normalizeHeaderName(headers, normalizedName) {
      utils.forEach(headers, function processHeader(value, name) {
        if (name !== normalizedName && name.toUpperCase() === normalizedName.toUpperCase()) {
          headers[normalizedName] = value;
          delete headers[name];
        }
      });
    };

    /**
     * Update an Error with the specified config, error code, and response.
     *
     * @param {Error} error The error to update.
     * @param {Object} config The config.
     * @param {string} [code] The error code (for example, 'ECONNABORTED').
     * @param {Object} [request] The request.
     * @param {Object} [response] The response.
     * @returns {Error} The error.
     */
    var enhanceError = function enhanceError(error, config, code, request, response) {
      error.config = config;
      if (code) {
        error.code = code;
      }

      error.request = request;
      error.response = response;
      error.isAxiosError = true;

      error.toJSON = function toJSON() {
        return {
          // Standard
          message: this.message,
          name: this.name,
          // Microsoft
          description: this.description,
          number: this.number,
          // Mozilla
          fileName: this.fileName,
          lineNumber: this.lineNumber,
          columnNumber: this.columnNumber,
          stack: this.stack,
          // Axios
          config: this.config,
          code: this.code
        };
      };
      return error;
    };

    /**
     * Create an Error with the specified message, config, error code, request and response.
     *
     * @param {string} message The error message.
     * @param {Object} config The config.
     * @param {string} [code] The error code (for example, 'ECONNABORTED').
     * @param {Object} [request] The request.
     * @param {Object} [response] The response.
     * @returns {Error} The created error.
     */
    var createError = function createError(message, config, code, request, response) {
      var error = new Error(message);
      return enhanceError(error, config, code, request, response);
    };

    /**
     * Resolve or reject a Promise based on response status.
     *
     * @param {Function} resolve A function that resolves the promise.
     * @param {Function} reject A function that rejects the promise.
     * @param {object} response The response.
     */
    var settle = function settle(resolve, reject, response) {
      var validateStatus = response.config.validateStatus;
      if (!response.status || !validateStatus || validateStatus(response.status)) {
        resolve(response);
      } else {
        reject(createError(
          'Request failed with status code ' + response.status,
          response.config,
          null,
          response.request,
          response
        ));
      }
    };

    var cookies = (
      utils.isStandardBrowserEnv() ?

      // Standard browser envs support document.cookie
        (function standardBrowserEnv() {
          return {
            write: function write(name, value, expires, path, domain, secure) {
              var cookie = [];
              cookie.push(name + '=' + encodeURIComponent(value));

              if (utils.isNumber(expires)) {
                cookie.push('expires=' + new Date(expires).toGMTString());
              }

              if (utils.isString(path)) {
                cookie.push('path=' + path);
              }

              if (utils.isString(domain)) {
                cookie.push('domain=' + domain);
              }

              if (secure === true) {
                cookie.push('secure');
              }

              document.cookie = cookie.join('; ');
            },

            read: function read(name) {
              var match = document.cookie.match(new RegExp('(^|;\\s*)(' + name + ')=([^;]*)'));
              return (match ? decodeURIComponent(match[3]) : null);
            },

            remove: function remove(name) {
              this.write(name, '', Date.now() - 86400000);
            }
          };
        })() :

      // Non standard browser env (web workers, react-native) lack needed support.
        (function nonStandardBrowserEnv() {
          return {
            write: function write() {},
            read: function read() { return null; },
            remove: function remove() {}
          };
        })()
    );

    /**
     * Determines whether the specified URL is absolute
     *
     * @param {string} url The URL to test
     * @returns {boolean} True if the specified URL is absolute, otherwise false
     */
    var isAbsoluteURL = function isAbsoluteURL(url) {
      // A URL is considered absolute if it begins with "<scheme>://" or "//" (protocol-relative URL).
      // RFC 3986 defines scheme name as a sequence of characters beginning with a letter and followed
      // by any combination of letters, digits, plus, period, or hyphen.
      return /^([a-z][a-z\d\+\-\.]*:)?\/\//i.test(url);
    };

    /**
     * Creates a new URL by combining the specified URLs
     *
     * @param {string} baseURL The base URL
     * @param {string} relativeURL The relative URL
     * @returns {string} The combined URL
     */
    var combineURLs = function combineURLs(baseURL, relativeURL) {
      return relativeURL
        ? baseURL.replace(/\/+$/, '') + '/' + relativeURL.replace(/^\/+/, '')
        : baseURL;
    };

    /**
     * Creates a new URL by combining the baseURL with the requestedURL,
     * only when the requestedURL is not already an absolute URL.
     * If the requestURL is absolute, this function returns the requestedURL untouched.
     *
     * @param {string} baseURL The base URL
     * @param {string} requestedURL Absolute or relative URL to combine
     * @returns {string} The combined full path
     */
    var buildFullPath = function buildFullPath(baseURL, requestedURL) {
      if (baseURL && !isAbsoluteURL(requestedURL)) {
        return combineURLs(baseURL, requestedURL);
      }
      return requestedURL;
    };

    // Headers whose duplicates are ignored by node
    // c.f. https://nodejs.org/api/http.html#http_message_headers
    var ignoreDuplicateOf = [
      'age', 'authorization', 'content-length', 'content-type', 'etag',
      'expires', 'from', 'host', 'if-modified-since', 'if-unmodified-since',
      'last-modified', 'location', 'max-forwards', 'proxy-authorization',
      'referer', 'retry-after', 'user-agent'
    ];

    /**
     * Parse headers into an object
     *
     * ```
     * Date: Wed, 27 Aug 2014 08:58:49 GMT
     * Content-Type: application/json
     * Connection: keep-alive
     * Transfer-Encoding: chunked
     * ```
     *
     * @param {String} headers Headers needing to be parsed
     * @returns {Object} Headers parsed into an object
     */
    var parseHeaders = function parseHeaders(headers) {
      var parsed = {};
      var key;
      var val;
      var i;

      if (!headers) { return parsed; }

      utils.forEach(headers.split('\n'), function parser(line) {
        i = line.indexOf(':');
        key = utils.trim(line.substr(0, i)).toLowerCase();
        val = utils.trim(line.substr(i + 1));

        if (key) {
          if (parsed[key] && ignoreDuplicateOf.indexOf(key) >= 0) {
            return;
          }
          if (key === 'set-cookie') {
            parsed[key] = (parsed[key] ? parsed[key] : []).concat([val]);
          } else {
            parsed[key] = parsed[key] ? parsed[key] + ', ' + val : val;
          }
        }
      });

      return parsed;
    };

    var isURLSameOrigin = (
      utils.isStandardBrowserEnv() ?

      // Standard browser envs have full support of the APIs needed to test
      // whether the request URL is of the same origin as current location.
        (function standardBrowserEnv() {
          var msie = /(msie|trident)/i.test(navigator.userAgent);
          var urlParsingNode = document.createElement('a');
          var originURL;

          /**
        * Parse a URL to discover it's components
        *
        * @param {String} url The URL to be parsed
        * @returns {Object}
        */
          function resolveURL(url) {
            var href = url;

            if (msie) {
            // IE needs attribute set twice to normalize properties
              urlParsingNode.setAttribute('href', href);
              href = urlParsingNode.href;
            }

            urlParsingNode.setAttribute('href', href);

            // urlParsingNode provides the UrlUtils interface - http://url.spec.whatwg.org/#urlutils
            return {
              href: urlParsingNode.href,
              protocol: urlParsingNode.protocol ? urlParsingNode.protocol.replace(/:$/, '') : '',
              host: urlParsingNode.host,
              search: urlParsingNode.search ? urlParsingNode.search.replace(/^\?/, '') : '',
              hash: urlParsingNode.hash ? urlParsingNode.hash.replace(/^#/, '') : '',
              hostname: urlParsingNode.hostname,
              port: urlParsingNode.port,
              pathname: (urlParsingNode.pathname.charAt(0) === '/') ?
                urlParsingNode.pathname :
                '/' + urlParsingNode.pathname
            };
          }

          originURL = resolveURL(window.location.href);

          /**
        * Determine if a URL shares the same origin as the current location
        *
        * @param {String} requestURL The URL to test
        * @returns {boolean} True if URL shares the same origin, otherwise false
        */
          return function isURLSameOrigin(requestURL) {
            var parsed = (utils.isString(requestURL)) ? resolveURL(requestURL) : requestURL;
            return (parsed.protocol === originURL.protocol &&
                parsed.host === originURL.host);
          };
        })() :

      // Non standard browser envs (web workers, react-native) lack needed support.
        (function nonStandardBrowserEnv() {
          return function isURLSameOrigin() {
            return true;
          };
        })()
    );

    var xhr = function xhrAdapter(config) {
      return new Promise(function dispatchXhrRequest(resolve, reject) {
        var requestData = config.data;
        var requestHeaders = config.headers;

        if (utils.isFormData(requestData)) {
          delete requestHeaders['Content-Type']; // Let the browser set it
        }

        var request = new XMLHttpRequest();

        // HTTP basic authentication
        if (config.auth) {
          var username = config.auth.username || '';
          var password = config.auth.password ? unescape(encodeURIComponent(config.auth.password)) : '';
          requestHeaders.Authorization = 'Basic ' + btoa(username + ':' + password);
        }

        var fullPath = buildFullPath(config.baseURL, config.url);
        request.open(config.method.toUpperCase(), buildURL(fullPath, config.params, config.paramsSerializer), true);

        // Set the request timeout in MS
        request.timeout = config.timeout;

        // Listen for ready state
        request.onreadystatechange = function handleLoad() {
          if (!request || request.readyState !== 4) {
            return;
          }

          // The request errored out and we didn't get a response, this will be
          // handled by onerror instead
          // With one exception: request that using file: protocol, most browsers
          // will return status as 0 even though it's a successful request
          if (request.status === 0 && !(request.responseURL && request.responseURL.indexOf('file:') === 0)) {
            return;
          }

          // Prepare the response
          var responseHeaders = 'getAllResponseHeaders' in request ? parseHeaders(request.getAllResponseHeaders()) : null;
          var responseData = !config.responseType || config.responseType === 'text' ? request.responseText : request.response;
          var response = {
            data: responseData,
            status: request.status,
            statusText: request.statusText,
            headers: responseHeaders,
            config: config,
            request: request
          };

          settle(resolve, reject, response);

          // Clean up request
          request = null;
        };

        // Handle browser request cancellation (as opposed to a manual cancellation)
        request.onabort = function handleAbort() {
          if (!request) {
            return;
          }

          reject(createError('Request aborted', config, 'ECONNABORTED', request));

          // Clean up request
          request = null;
        };

        // Handle low level network errors
        request.onerror = function handleError() {
          // Real errors are hidden from us by the browser
          // onerror should only fire if it's a network error
          reject(createError('Network Error', config, null, request));

          // Clean up request
          request = null;
        };

        // Handle timeout
        request.ontimeout = function handleTimeout() {
          var timeoutErrorMessage = 'timeout of ' + config.timeout + 'ms exceeded';
          if (config.timeoutErrorMessage) {
            timeoutErrorMessage = config.timeoutErrorMessage;
          }
          reject(createError(timeoutErrorMessage, config, 'ECONNABORTED',
            request));

          // Clean up request
          request = null;
        };

        // Add xsrf header
        // This is only done if running in a standard browser environment.
        // Specifically not if we're in a web worker, or react-native.
        if (utils.isStandardBrowserEnv()) {
          // Add xsrf header
          var xsrfValue = (config.withCredentials || isURLSameOrigin(fullPath)) && config.xsrfCookieName ?
            cookies.read(config.xsrfCookieName) :
            undefined;

          if (xsrfValue) {
            requestHeaders[config.xsrfHeaderName] = xsrfValue;
          }
        }

        // Add headers to the request
        if ('setRequestHeader' in request) {
          utils.forEach(requestHeaders, function setRequestHeader(val, key) {
            if (typeof requestData === 'undefined' && key.toLowerCase() === 'content-type') {
              // Remove Content-Type if data is undefined
              delete requestHeaders[key];
            } else {
              // Otherwise add header to the request
              request.setRequestHeader(key, val);
            }
          });
        }

        // Add withCredentials to request if needed
        if (!utils.isUndefined(config.withCredentials)) {
          request.withCredentials = !!config.withCredentials;
        }

        // Add responseType to request if needed
        if (config.responseType) {
          try {
            request.responseType = config.responseType;
          } catch (e) {
            // Expected DOMException thrown by browsers not compatible XMLHttpRequest Level 2.
            // But, this can be suppressed for 'json' type as it can be parsed by default 'transformResponse' function.
            if (config.responseType !== 'json') {
              throw e;
            }
          }
        }

        // Handle progress if needed
        if (typeof config.onDownloadProgress === 'function') {
          request.addEventListener('progress', config.onDownloadProgress);
        }

        // Not all browsers support upload events
        if (typeof config.onUploadProgress === 'function' && request.upload) {
          request.upload.addEventListener('progress', config.onUploadProgress);
        }

        if (config.cancelToken) {
          // Handle cancellation
          config.cancelToken.promise.then(function onCanceled(cancel) {
            if (!request) {
              return;
            }

            request.abort();
            reject(cancel);
            // Clean up request
            request = null;
          });
        }

        if (!requestData) {
          requestData = null;
        }

        // Send the request
        request.send(requestData);
      });
    };

    var DEFAULT_CONTENT_TYPE = {
      'Content-Type': 'application/x-www-form-urlencoded'
    };

    function setContentTypeIfUnset(headers, value) {
      if (!utils.isUndefined(headers) && utils.isUndefined(headers['Content-Type'])) {
        headers['Content-Type'] = value;
      }
    }

    function getDefaultAdapter() {
      var adapter;
      if (typeof XMLHttpRequest !== 'undefined') {
        // For browsers use XHR adapter
        adapter = xhr;
      } else if (typeof process !== 'undefined' && Object.prototype.toString.call(process) === '[object process]') {
        // For node use HTTP adapter
        adapter = xhr;
      }
      return adapter;
    }

    var defaults = {
      adapter: getDefaultAdapter(),

      transformRequest: [function transformRequest(data, headers) {
        normalizeHeaderName(headers, 'Accept');
        normalizeHeaderName(headers, 'Content-Type');
        if (utils.isFormData(data) ||
          utils.isArrayBuffer(data) ||
          utils.isBuffer(data) ||
          utils.isStream(data) ||
          utils.isFile(data) ||
          utils.isBlob(data)
        ) {
          return data;
        }
        if (utils.isArrayBufferView(data)) {
          return data.buffer;
        }
        if (utils.isURLSearchParams(data)) {
          setContentTypeIfUnset(headers, 'application/x-www-form-urlencoded;charset=utf-8');
          return data.toString();
        }
        if (utils.isObject(data)) {
          setContentTypeIfUnset(headers, 'application/json;charset=utf-8');
          return JSON.stringify(data);
        }
        return data;
      }],

      transformResponse: [function transformResponse(data) {
        /*eslint no-param-reassign:0*/
        if (typeof data === 'string') {
          try {
            data = JSON.parse(data);
          } catch (e) { /* Ignore */ }
        }
        return data;
      }],

      /**
       * A timeout in milliseconds to abort a request. If set to 0 (default) a
       * timeout is not created.
       */
      timeout: 0,

      xsrfCookieName: 'XSRF-TOKEN',
      xsrfHeaderName: 'X-XSRF-TOKEN',

      maxContentLength: -1,
      maxBodyLength: -1,

      validateStatus: function validateStatus(status) {
        return status >= 200 && status < 300;
      }
    };

    defaults.headers = {
      common: {
        'Accept': 'application/json, text/plain, */*'
      }
    };

    utils.forEach(['delete', 'get', 'head'], function forEachMethodNoData(method) {
      defaults.headers[method] = {};
    });

    utils.forEach(['post', 'put', 'patch'], function forEachMethodWithData(method) {
      defaults.headers[method] = utils.merge(DEFAULT_CONTENT_TYPE);
    });

    var defaults_1 = defaults;

    /**
     * Throws a `Cancel` if cancellation has been requested.
     */
    function throwIfCancellationRequested(config) {
      if (config.cancelToken) {
        config.cancelToken.throwIfRequested();
      }
    }

    /**
     * Dispatch a request to the server using the configured adapter.
     *
     * @param {object} config The config that is to be used for the request
     * @returns {Promise} The Promise to be fulfilled
     */
    var dispatchRequest = function dispatchRequest(config) {
      throwIfCancellationRequested(config);

      // Ensure headers exist
      config.headers = config.headers || {};

      // Transform request data
      config.data = transformData(
        config.data,
        config.headers,
        config.transformRequest
      );

      // Flatten headers
      config.headers = utils.merge(
        config.headers.common || {},
        config.headers[config.method] || {},
        config.headers
      );

      utils.forEach(
        ['delete', 'get', 'head', 'post', 'put', 'patch', 'common'],
        function cleanHeaderConfig(method) {
          delete config.headers[method];
        }
      );

      var adapter = config.adapter || defaults_1.adapter;

      return adapter(config).then(function onAdapterResolution(response) {
        throwIfCancellationRequested(config);

        // Transform response data
        response.data = transformData(
          response.data,
          response.headers,
          config.transformResponse
        );

        return response;
      }, function onAdapterRejection(reason) {
        if (!isCancel(reason)) {
          throwIfCancellationRequested(config);

          // Transform response data
          if (reason && reason.response) {
            reason.response.data = transformData(
              reason.response.data,
              reason.response.headers,
              config.transformResponse
            );
          }
        }

        return Promise.reject(reason);
      });
    };

    /**
     * Config-specific merge-function which creates a new config-object
     * by merging two configuration objects together.
     *
     * @param {Object} config1
     * @param {Object} config2
     * @returns {Object} New object resulting from merging config2 to config1
     */
    var mergeConfig = function mergeConfig(config1, config2) {
      // eslint-disable-next-line no-param-reassign
      config2 = config2 || {};
      var config = {};

      var valueFromConfig2Keys = ['url', 'method', 'data'];
      var mergeDeepPropertiesKeys = ['headers', 'auth', 'proxy', 'params'];
      var defaultToConfig2Keys = [
        'baseURL', 'transformRequest', 'transformResponse', 'paramsSerializer',
        'timeout', 'timeoutMessage', 'withCredentials', 'adapter', 'responseType', 'xsrfCookieName',
        'xsrfHeaderName', 'onUploadProgress', 'onDownloadProgress', 'decompress',
        'maxContentLength', 'maxBodyLength', 'maxRedirects', 'transport', 'httpAgent',
        'httpsAgent', 'cancelToken', 'socketPath', 'responseEncoding'
      ];
      var directMergeKeys = ['validateStatus'];

      function getMergedValue(target, source) {
        if (utils.isPlainObject(target) && utils.isPlainObject(source)) {
          return utils.merge(target, source);
        } else if (utils.isPlainObject(source)) {
          return utils.merge({}, source);
        } else if (utils.isArray(source)) {
          return source.slice();
        }
        return source;
      }

      function mergeDeepProperties(prop) {
        if (!utils.isUndefined(config2[prop])) {
          config[prop] = getMergedValue(config1[prop], config2[prop]);
        } else if (!utils.isUndefined(config1[prop])) {
          config[prop] = getMergedValue(undefined, config1[prop]);
        }
      }

      utils.forEach(valueFromConfig2Keys, function valueFromConfig2(prop) {
        if (!utils.isUndefined(config2[prop])) {
          config[prop] = getMergedValue(undefined, config2[prop]);
        }
      });

      utils.forEach(mergeDeepPropertiesKeys, mergeDeepProperties);

      utils.forEach(defaultToConfig2Keys, function defaultToConfig2(prop) {
        if (!utils.isUndefined(config2[prop])) {
          config[prop] = getMergedValue(undefined, config2[prop]);
        } else if (!utils.isUndefined(config1[prop])) {
          config[prop] = getMergedValue(undefined, config1[prop]);
        }
      });

      utils.forEach(directMergeKeys, function merge(prop) {
        if (prop in config2) {
          config[prop] = getMergedValue(config1[prop], config2[prop]);
        } else if (prop in config1) {
          config[prop] = getMergedValue(undefined, config1[prop]);
        }
      });

      var axiosKeys = valueFromConfig2Keys
        .concat(mergeDeepPropertiesKeys)
        .concat(defaultToConfig2Keys)
        .concat(directMergeKeys);

      var otherKeys = Object
        .keys(config1)
        .concat(Object.keys(config2))
        .filter(function filterAxiosKeys(key) {
          return axiosKeys.indexOf(key) === -1;
        });

      utils.forEach(otherKeys, mergeDeepProperties);

      return config;
    };

    /**
     * Create a new instance of Axios
     *
     * @param {Object} instanceConfig The default config for the instance
     */
    function Axios(instanceConfig) {
      this.defaults = instanceConfig;
      this.interceptors = {
        request: new InterceptorManager_1(),
        response: new InterceptorManager_1()
      };
    }

    /**
     * Dispatch a request
     *
     * @param {Object} config The config specific for this request (merged with this.defaults)
     */
    Axios.prototype.request = function request(config) {
      /*eslint no-param-reassign:0*/
      // Allow for axios('example/url'[, config]) a la fetch API
      if (typeof config === 'string') {
        config = arguments[1] || {};
        config.url = arguments[0];
      } else {
        config = config || {};
      }

      config = mergeConfig(this.defaults, config);

      // Set config.method
      if (config.method) {
        config.method = config.method.toLowerCase();
      } else if (this.defaults.method) {
        config.method = this.defaults.method.toLowerCase();
      } else {
        config.method = 'get';
      }

      // Hook up interceptors middleware
      var chain = [dispatchRequest, undefined];
      var promise = Promise.resolve(config);

      this.interceptors.request.forEach(function unshiftRequestInterceptors(interceptor) {
        chain.unshift(interceptor.fulfilled, interceptor.rejected);
      });

      this.interceptors.response.forEach(function pushResponseInterceptors(interceptor) {
        chain.push(interceptor.fulfilled, interceptor.rejected);
      });

      while (chain.length) {
        promise = promise.then(chain.shift(), chain.shift());
      }

      return promise;
    };

    Axios.prototype.getUri = function getUri(config) {
      config = mergeConfig(this.defaults, config);
      return buildURL(config.url, config.params, config.paramsSerializer).replace(/^\?/, '');
    };

    // Provide aliases for supported request methods
    utils.forEach(['delete', 'get', 'head', 'options'], function forEachMethodNoData(method) {
      /*eslint func-names:0*/
      Axios.prototype[method] = function(url, config) {
        return this.request(mergeConfig(config || {}, {
          method: method,
          url: url,
          data: (config || {}).data
        }));
      };
    });

    utils.forEach(['post', 'put', 'patch'], function forEachMethodWithData(method) {
      /*eslint func-names:0*/
      Axios.prototype[method] = function(url, data, config) {
        return this.request(mergeConfig(config || {}, {
          method: method,
          url: url,
          data: data
        }));
      };
    });

    var Axios_1 = Axios;

    /**
     * A `Cancel` is an object that is thrown when an operation is canceled.
     *
     * @class
     * @param {string=} message The message.
     */
    function Cancel(message) {
      this.message = message;
    }

    Cancel.prototype.toString = function toString() {
      return 'Cancel' + (this.message ? ': ' + this.message : '');
    };

    Cancel.prototype.__CANCEL__ = true;

    var Cancel_1 = Cancel;

    /**
     * A `CancelToken` is an object that can be used to request cancellation of an operation.
     *
     * @class
     * @param {Function} executor The executor function.
     */
    function CancelToken(executor) {
      if (typeof executor !== 'function') {
        throw new TypeError('executor must be a function.');
      }

      var resolvePromise;
      this.promise = new Promise(function promiseExecutor(resolve) {
        resolvePromise = resolve;
      });

      var token = this;
      executor(function cancel(message) {
        if (token.reason) {
          // Cancellation has already been requested
          return;
        }

        token.reason = new Cancel_1(message);
        resolvePromise(token.reason);
      });
    }

    /**
     * Throws a `Cancel` if cancellation has been requested.
     */
    CancelToken.prototype.throwIfRequested = function throwIfRequested() {
      if (this.reason) {
        throw this.reason;
      }
    };

    /**
     * Returns an object that contains a new `CancelToken` and a function that, when called,
     * cancels the `CancelToken`.
     */
    CancelToken.source = function source() {
      var cancel;
      var token = new CancelToken(function executor(c) {
        cancel = c;
      });
      return {
        token: token,
        cancel: cancel
      };
    };

    var CancelToken_1 = CancelToken;

    /**
     * Syntactic sugar for invoking a function and expanding an array for arguments.
     *
     * Common use case would be to use `Function.prototype.apply`.
     *
     *  ```js
     *  function f(x, y, z) {}
     *  var args = [1, 2, 3];
     *  f.apply(null, args);
     *  ```
     *
     * With `spread` this example can be re-written.
     *
     *  ```js
     *  spread(function(x, y, z) {})([1, 2, 3]);
     *  ```
     *
     * @param {Function} callback
     * @returns {Function}
     */
    var spread = function spread(callback) {
      return function wrap(arr) {
        return callback.apply(null, arr);
      };
    };

    /**
     * Create an instance of Axios
     *
     * @param {Object} defaultConfig The default config for the instance
     * @return {Axios} A new instance of Axios
     */
    function createInstance(defaultConfig) {
      var context = new Axios_1(defaultConfig);
      var instance = bind(Axios_1.prototype.request, context);

      // Copy axios.prototype to instance
      utils.extend(instance, Axios_1.prototype, context);

      // Copy context to instance
      utils.extend(instance, context);

      return instance;
    }

    // Create the default instance to be exported
    var axios = createInstance(defaults_1);

    // Expose Axios class to allow class inheritance
    axios.Axios = Axios_1;

    // Factory for creating new instances
    axios.create = function create(instanceConfig) {
      return createInstance(mergeConfig(axios.defaults, instanceConfig));
    };

    // Expose Cancel & CancelToken
    axios.Cancel = Cancel_1;
    axios.CancelToken = CancelToken_1;
    axios.isCancel = isCancel;

    // Expose all/spread
    axios.all = function all(promises) {
      return Promise.all(promises);
    };
    axios.spread = spread;

    var axios_1 = axios;

    // Allow use of default import syntax in TypeScript
    var _default = axios;
    axios_1.default = _default;

    /*
     Copyright 2019 Florian Dold

     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
     You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

     Unless required by applicable law or agreed to in writing, software
     distributed under the License is distributed on an "AS IS" BASIS,
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
     or implied. See the License for the specific language governing
     permissions and limitations under the License.
     */
    /** @public */
    var ResultLevel;
    (function (ResultLevel) {
        ResultLevel[ResultLevel["OnlyCount"] = 0] = "OnlyCount";
        ResultLevel[ResultLevel["OnlyKeys"] = 1] = "OnlyKeys";
        ResultLevel[ResultLevel["Full"] = 2] = "Full";
    })(ResultLevel || (ResultLevel = {}));
    /** @public */
    var StoreLevel;
    (function (StoreLevel) {
        StoreLevel[StoreLevel["NoOverwrite"] = 0] = "NoOverwrite";
        StoreLevel[StoreLevel["AllowOverwrite"] = 1] = "AllowOverwrite";
        StoreLevel[StoreLevel["UpdateExisting"] = 2] = "UpdateExisting";
    })(StoreLevel || (StoreLevel = {}));

    /*
    Copyright (c) 2018 David Piepgrass

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

    SPDX-License-Identifier: MIT
    */
    // Informative microbenchmarks & stuff:
    // http://www.jayconrod.com/posts/52/a-tour-of-v8-object-representation (very educational)
    // https://blog.mozilla.org/luke/2012/10/02/optimizing-javascript-variable-access/ (local vars are faster than properties)
    // http://benediktmeurer.de/2017/12/13/an-introduction-to-speculative-optimization-in-v8/ (other stuff)
    // https://jsperf.com/js-in-operator-vs-alternatives (avoid 'in' operator; `.p!==undefined` faster than `hasOwnProperty('p')` in all browsers)
    // https://jsperf.com/instanceof-vs-typeof-vs-constructor-vs-member (speed of type tests varies wildly across browsers)
    // https://jsperf.com/detecting-arrays-new (a.constructor===Array is best across browsers, assuming a is an object)
    // https://jsperf.com/shallow-cloning-methods (a constructor is faster than Object.create; hand-written clone faster than Object.assign)
    // https://jsperf.com/ways-to-fill-an-array (slice-and-replace is fastest)
    // https://jsperf.com/math-min-max-vs-ternary-vs-if (Math.min/max is slow on Edge)
    // https://jsperf.com/array-vs-property-access-speed (v.x/v.y is faster than a[0]/a[1] in major browsers IF hidden class is constant)
    // https://jsperf.com/detect-not-null-or-undefined (`x==null` slightly slower than `x===null||x===undefined` on all browsers)
    // Overall, microbenchmarks suggest Firefox is the fastest browser for JavaScript and Edge is the slowest.
    // Lessons from https://v8project.blogspot.com/2017/09/elements-kinds-in-v8.html:
    //   - Avoid holes in arrays. Avoid `new Array(N)`, it will be "holey" permanently.
    //   - Don't read outside bounds of an array (it scans prototype chain).
    //   - Small integer arrays are stored differently from doubles
    //   - Adding non-numbers to an array deoptimizes it permanently into a general array
    //   - Objects can be used like arrays (e.g. have length property) but are slower
    //   - V8 source (NewElementsCapacity in src/objects.h): arrays grow by 50% + 16 elements
    /** Compares two numbers, strings, arrays of numbers/strings, Dates,
     *  or objects that have a valueOf() method returning a number or string.
     *  Optimized for numbers. Returns 1 if a>b, -1 if a<b, and 0 if a===b.
     */
    function defaultComparator(a, b) {
        var c = a - b;
        if (c === c)
            return c; // a & b are number
        // General case (c is NaN): string / arrays / Date / incomparable things
        if (a)
            a = a.valueOf();
        if (b)
            b = b.valueOf();
        return a < b ? -1 : a > b ? 1 : a == b ? 0 : c;
    }
    /**
     * A reasonably fast collection of key-value pairs with a powerful API.
     * Largely compatible with the standard Map. BTree is a B+ tree data structure,
     * so the collection is sorted by key.
     *
     * B+ trees tend to use memory more efficiently than hashtables such as the
     * standard Map, especially when the collection contains a large number of
     * items. However, maintaining the sort order makes them modestly slower:
     * O(log size) rather than O(1). This B+ tree implementation supports O(1)
     * fast cloning. It also supports freeze(), which can be used to ensure that
     * a BTree is not changed accidentally.
     *
     * Confusingly, the ES6 Map.forEach(c) method calls c(value,key) instead of
     * c(key,value), in contrast to other methods such as set() and entries()
     * which put the key first. I can only assume that the order was reversed on
     * the theory that users would usually want to examine values and ignore keys.
     * BTree's forEach() therefore works the same way, but a second method
     * `.forEachPair((key,value)=>{...})` is provided which sends you the key
     * first and the value second; this method is slightly faster because it is
     * the "native" for-each method for this class.
     *
     * Out of the box, BTree supports keys that are numbers, strings, arrays of
     * numbers/strings, Date, and objects that have a valueOf() method returning a
     * number or string. Other data types, such as arrays of Date or custom
     * objects, require a custom comparator, which you must pass as the second
     * argument to the constructor (the first argument is an optional list of
     * initial items). Symbols cannot be used as keys because they are unordered
     * (one Symbol is never "greater" or "less" than another).
     *
     * @example
     * Given a {name: string, age: number} object, you can create a tree sorted by
     * name and then by age like this:
     *
     *     var tree = new BTree(undefined, (a, b) => {
     *       if (a.name > b.name)
     *         return 1; // Return a number >0 when a > b
     *       else if (a.name < b.name)
     *         return -1; // Return a number <0 when a < b
     *       else // names are equal (or incomparable)
     *         return a.age - b.age; // Return >0 when a.age > b.age
     *     });
     *
     *     tree.set({name:"Bill", age:17}, "happy");
     *     tree.set({name:"Fran", age:40}, "busy & stressed");
     *     tree.set({name:"Bill", age:55}, "recently laid off");
     *     tree.forEachPair((k, v) => {
     *       console.log(`Name: ${k.name} Age: ${k.age} Status: ${v}`);
     *     });
     *
     * @description
     * The "range" methods (`forEach, forRange, editRange`) will return the number
     * of elements that were scanned. In addition, the callback can return {break:R}
     * to stop early and return R from the outer function.
     *
     * - TODO: Test performance of preallocating values array at max size
     * - TODO: Add fast initialization when a sorted array is provided to constructor
     *
     * For more documentation see https://github.com/qwertie/btree-typescript
     *
     * Are you a C# developer? You might like the similar data structures I made for C#:
     * BDictionary, BList, etc. See http://core.loyc.net/collections/
     *
     * @author David Piepgrass
     */
    class BTree {
        /**
         * Initializes an empty B+ tree.
         * @param compare Custom function to compare pairs of elements in the tree.
         *   This is not required for numbers, strings and arrays of numbers/strings.
         * @param entries A set of key-value pairs to initialize the tree
         * @param maxNodeSize Branching factor (maximum items or children per node)
         *   Must be in range 4..256. If undefined or <4 then default is used; if >256 then 256.
         */
        constructor(entries, compare, maxNodeSize) {
            this._root = EmptyLeaf;
            this._size = 0;
            this._maxNodeSize = maxNodeSize >= 4 ? Math.min(maxNodeSize, 256) : 32;
            this._compare = compare || defaultComparator;
            if (entries)
                this.setPairs(entries);
        }
        // ES6 Map<K,V> methods ///////////////////////////////////////////////////
        /** Gets the number of key-value pairs in the tree. */
        get size() {
            return this._size;
        }
        /** Gets the number of key-value pairs in the tree. */
        get length() {
            return this._size;
        }
        /** Returns true iff the tree contains no key-value pairs. */
        get isEmpty() {
            return this._size === 0;
        }
        /** Releases the tree so that its size is 0. */
        clear() {
            this._root = EmptyLeaf;
            this._size = 0;
        }
        /** Runs a function for each key-value pair, in order from smallest to
         *  largest key. For compatibility with ES6 Map, the argument order to
         *  the callback is backwards: value first, then key. Call forEachPair
         *  instead to receive the key as the first argument.
         * @param thisArg If provided, this parameter is assigned as the `this`
         *        value for each callback.
         * @returns the number of values that were sent to the callback,
         *        or the R value if the callback returned {break:R}. */
        forEach(callback, thisArg) {
            if (thisArg !== undefined)
                callback = callback.bind(thisArg);
            return this.forEachPair((k, v) => callback(v, k, this));
        }
        /** Runs a function for each key-value pair, in order from smallest to
         *  largest key. The callback can return {break:R} (where R is any value
         *  except undefined) to stop immediately and return R from forEachPair.
         * @param onFound A function that is called for each key-value pair. This
         *        function can return {break:R} to stop early with result R.
         *        The reason that you must return {break:R} instead of simply R
         *        itself is for consistency with editRange(), which allows
         *        multiple actions, not just breaking.
         * @param initialCounter This is the value of the third argument of
         *        `onFound` the first time it is called. The counter increases
         *        by one each time `onFound` is called. Default value: 0
         * @returns the number of pairs sent to the callback (plus initialCounter,
         *        if you provided one). If the callback returned {break:R} then
         *        the R value is returned instead. */
        forEachPair(callback, initialCounter) {
            var low = this.minKey(), high = this.maxKey();
            return this.forRange(low, high, true, callback, initialCounter);
        }
        /**
         * Finds a pair in the tree and returns the associated value.
         * @param defaultValue a value to return if the key was not found.
         * @returns the value, or defaultValue if the key was not found.
         * @description Computational complexity: O(log size)
         */
        get(key, defaultValue) {
            return this._root.get(key, defaultValue, this);
        }
        /**
         * Adds or overwrites a key-value pair in the B+ tree.
         * @param key the key is used to determine the sort order of
         *        data in the tree.
         * @param value data to associate with the key (optional)
         * @param overwrite Whether to overwrite an existing key-value pair
         *        (default: true). If this is false and there is an existing
         *        key-value pair then this method has no effect.
         * @returns true if a new key-value pair was added.
         * @description Computational complexity: O(log size)
         * Note: when overwriting a previous entry, the key is updated
         * as well as the value. This has no effect unless the new key
         * has data that does not affect its sort order.
         */
        set(key, value, overwrite) {
            if (this._root.isShared)
                this._root = this._root.clone();
            var result = this._root.set(key, value, overwrite, this);
            if (result === true || result === false)
                return result;
            // Root node has split, so create a new root node.
            this._root = new BNodeInternal([this._root, result]);
            return true;
        }
        /**
         * Returns true if the key exists in the B+ tree, false if not.
         * Use get() for best performance; use has() if you need to
         * distinguish between "undefined value" and "key not present".
         * @param key Key to detect
         * @description Computational complexity: O(log size)
         */
        has(key) {
            return this.forRange(key, key, true, undefined) !== 0;
        }
        /**
         * Removes a single key-value pair from the B+ tree.
         * @param key Key to find
         * @returns true if a pair was found and removed, false otherwise.
         * @description Computational complexity: O(log size)
         */
        delete(key) {
            return this.editRange(key, key, true, DeleteRange) !== 0;
        }
        with(key, value, overwrite) {
            let nu = this.clone();
            return nu.set(key, value, overwrite) || overwrite ? nu : this;
        }
        /** Returns a copy of the tree with the specified key-value pairs set. */
        withPairs(pairs, overwrite) {
            let nu = this.clone();
            return nu.setPairs(pairs, overwrite) !== 0 || overwrite ? nu : this;
        }
        /** Returns a copy of the tree with the specified keys present.
         *  @param keys The keys to add. If a key is already present in the tree,
         *         neither the existing key nor the existing value is modified.
         *  @param returnThisIfUnchanged if true, returns this if all keys already
         *  existed. Performance note: due to the architecture of this class, all
         *  node(s) leading to existing keys are cloned even if the collection is
         *  ultimately unchanged.
         */
        withKeys(keys, returnThisIfUnchanged) {
            let nu = this.clone(), changed = false;
            for (var i = 0; i < keys.length; i++)
                changed = nu.set(keys[i], undefined, false) || changed;
            return returnThisIfUnchanged && !changed ? this : nu;
        }
        /** Returns a copy of the tree with the specified key removed.
         * @param returnThisIfUnchanged if true, returns this if the key didn't exist.
         *  Performance note: due to the architecture of this class, node(s) leading
         *  to where the key would have been stored are cloned even when the key
         *  turns out not to exist and the collection is unchanged.
         */
        without(key, returnThisIfUnchanged) {
            return this.withoutRange(key, key, true, returnThisIfUnchanged);
        }
        /** Returns a copy of the tree with the specified keys removed.
         * @param returnThisIfUnchanged if true, returns this if none of the keys
         *  existed. Performance note: due to the architecture of this class,
         *  node(s) leading to where the key would have been stored are cloned
         *  even when the key turns out not to exist.
         */
        withoutKeys(keys, returnThisIfUnchanged) {
            let nu = this.clone();
            return nu.deleteKeys(keys) || !returnThisIfUnchanged ? nu : this;
        }
        /** Returns a copy of the tree with the specified range of keys removed. */
        withoutRange(low, high, includeHigh, returnThisIfUnchanged) {
            let nu = this.clone();
            if (nu.deleteRange(low, high, includeHigh) === 0 && returnThisIfUnchanged)
                return this;
            return nu;
        }
        /** Returns a copy of the tree with pairs removed whenever the callback
         *  function returns false. `where()` is a synonym for this method. */
        filter(callback, returnThisIfUnchanged) {
            var nu = this.greedyClone();
            var del;
            nu.editAll((k, v, i) => {
                if (!callback(k, v, i))
                    return (del = Delete);
            });
            if (!del && returnThisIfUnchanged)
                return this;
            return nu;
        }
        /** Returns a copy of the tree with all values altered by a callback function. */
        mapValues(callback) {
            var tmp = {};
            var nu = this.greedyClone();
            nu.editAll((k, v, i) => {
                return (tmp.value = callback(v, k, i)), tmp;
            });
            return nu;
        }
        reduce(callback, initialValue) {
            let i = 0, p = initialValue;
            var it = this.entries(this.minKey(), ReusedArray), next;
            while (!(next = it.next()).done)
                p = callback(p, next.value, i++, this);
            return p;
        }
        // Iterator methods ///////////////////////////////////////////////////////
        /** Returns an iterator that provides items in order (ascending order if
         *  the collection's comparator uses ascending order, as is the default.)
         *  @param lowestKey First key to be iterated, or undefined to start at
         *         minKey(). If the specified key doesn't exist then iteration
         *         starts at the next higher key (according to the comparator).
         *  @param reusedArray Optional array used repeatedly to store key-value
         *         pairs, to avoid creating a new array on every iteration.
         */
        entries(lowestKey, reusedArray) {
            var info = this.findPath(lowestKey);
            if (info === undefined)
                return iterator();
            var { nodequeue, nodeindex, leaf } = info;
            var state = reusedArray !== undefined ? 1 : 0;
            var i = lowestKey === undefined
                ? -1
                : leaf.indexOf(lowestKey, 0, this._compare) - 1;
            return iterator(() => {
                jump: for (;;) {
                    switch (state) {
                        case 0:
                            if (++i < leaf.keys.length)
                                return { done: false, value: [leaf.keys[i], leaf.values[i]] };
                            state = 2;
                            continue;
                        case 1:
                            if (++i < leaf.keys.length) {
                                (reusedArray[0] = leaf.keys[i]),
                                    (reusedArray[1] = leaf.values[i]);
                                return { done: false, value: reusedArray };
                            }
                            state = 2;
                        case 2:
                            // Advance to the next leaf node
                            for (var level = -1;;) {
                                if (++level >= nodequeue.length) {
                                    state = 3;
                                    continue jump;
                                }
                                if (++nodeindex[level] < nodequeue[level].length)
                                    break;
                            }
                            for (; level > 0; level--) {
                                nodequeue[level - 1] = nodequeue[level][nodeindex[level]].children;
                                nodeindex[level - 1] = 0;
                            }
                            leaf = nodequeue[0][nodeindex[0]];
                            i = -1;
                            state = reusedArray !== undefined ? 1 : 0;
                            continue;
                        case 3:
                            return { done: true, value: undefined };
                    }
                }
            });
        }
        /** Returns an iterator that provides items in reversed order.
         *  @param highestKey Key at which to start iterating, or undefined to
         *         start at minKey(). If the specified key doesn't exist then iteration
         *         starts at the next lower key (according to the comparator).
         *  @param reusedArray Optional array used repeatedly to store key-value
         *         pairs, to avoid creating a new array on every iteration.
         *  @param skipHighest Iff this flag is true and the highestKey exists in the
         *         collection, the pair matching highestKey is skipped, not iterated.
         */
        entriesReversed(highestKey, reusedArray, skipHighest) {
            if ((highestKey = highestKey || this.maxKey()) === undefined)
                return iterator(); // collection is empty
            var { nodequeue, nodeindex, leaf } = this.findPath(highestKey) || this.findPath(this.maxKey());
            check$1(!nodequeue[0] || leaf === nodequeue[0][nodeindex[0]], "wat!");
            var i = leaf.indexOf(highestKey, 0, this._compare);
            if (!(skipHighest || this._compare(leaf.keys[i], highestKey) > 0))
                i++;
            var state = reusedArray !== undefined ? 1 : 0;
            return iterator(() => {
                jump: for (;;) {
                    switch (state) {
                        case 0:
                            if (--i >= 0)
                                return { done: false, value: [leaf.keys[i], leaf.values[i]] };
                            state = 2;
                            continue;
                        case 1:
                            if (--i >= 0) {
                                (reusedArray[0] = leaf.keys[i]),
                                    (reusedArray[1] = leaf.values[i]);
                                return { done: false, value: reusedArray };
                            }
                            state = 2;
                        case 2:
                            // Advance to the next leaf node
                            for (var level = -1;;) {
                                if (++level >= nodequeue.length) {
                                    state = 3;
                                    continue jump;
                                }
                                if (--nodeindex[level] >= 0)
                                    break;
                            }
                            for (; level > 0; level--) {
                                nodequeue[level - 1] = nodequeue[level][nodeindex[level]].children;
                                nodeindex[level - 1] = nodequeue[level - 1].length - 1;
                            }
                            leaf = nodequeue[0][nodeindex[0]];
                            i = leaf.keys.length;
                            state = reusedArray !== undefined ? 1 : 0;
                            continue;
                        case 3:
                            return { done: true, value: undefined };
                    }
                }
            });
        }
        /* Used by entries() and entriesReversed() to prepare to start iterating.
         * It develops a "node queue" for each non-leaf level of the tree.
         * Levels are numbered "bottom-up" so that level 0 is a list of leaf
         * nodes from a low-level non-leaf node. The queue at a given level L
         * consists of nodequeue[L] which is the children of a BNodeInternal,
         * and nodeindex[L], the current index within that child list, such
         * such that nodequeue[L-1] === nodequeue[L][nodeindex[L]].children.
         * (However inside this function the order is reversed.)
         */
        findPath(key) {
            var nextnode = this._root;
            var nodequeue, nodeindex;
            if (nextnode.isLeaf) {
                (nodequeue = EmptyArray), (nodeindex = EmptyArray); // avoid allocations
            }
            else {
                (nodequeue = []), (nodeindex = []);
                for (var d = 0; !nextnode.isLeaf; d++) {
                    nodequeue[d] = nextnode.children;
                    nodeindex[d] =
                        key === undefined ? 0 : nextnode.indexOf(key, 0, this._compare);
                    if (nodeindex[d] >= nodequeue[d].length)
                        return; // first key > maxKey()
                    nextnode = nodequeue[d][nodeindex[d]];
                }
                nodequeue.reverse();
                nodeindex.reverse();
            }
            return { nodequeue, nodeindex, leaf: nextnode };
        }
        /** Returns a new iterator for iterating the keys of each pair in ascending order.
         *  @param firstKey: Minimum key to include in the output. */
        keys(firstKey) {
            var it = this.entries(firstKey, ReusedArray);
            return iterator(() => {
                var n = it.next();
                if (n.value)
                    n.value = n.value[0];
                return n;
            });
        }
        /** Returns a new iterator for iterating the values of each pair in order by key.
         *  @param firstKey: Minimum key whose associated value is included in the output. */
        values(firstKey) {
            var it = this.entries(firstKey, ReusedArray);
            return iterator(() => {
                var n = it.next();
                if (n.value)
                    n.value = n.value[1];
                return n;
            });
        }
        // Additional methods /////////////////////////////////////////////////////
        /** Returns the maximum number of children/values before nodes will split. */
        get maxNodeSize() {
            return this._maxNodeSize;
        }
        /** Gets the lowest key in the tree. Complexity: O(log size) */
        minKey() {
            return this._root.minKey();
        }
        /** Gets the highest key in the tree. Complexity: O(1) */
        maxKey() {
            return this._root.maxKey();
        }
        /** Quickly clones the tree by marking the root node as shared.
         *  Both copies remain editable. When you modify either copy, any
         *  nodes that are shared (or potentially shared) between the two
         *  copies are cloned so that the changes do not affect other copies.
         *  This is known as copy-on-write behavior, or "lazy copying". */
        clone() {
            this._root.isShared = true;
            var result = new BTree(undefined, this._compare, this._maxNodeSize);
            result._root = this._root;
            result._size = this._size;
            return result;
        }
        /** Performs a greedy clone, immediately duplicating any nodes that are
         *  not currently marked as shared, in order to avoid marking any nodes
         *  as shared.
         *  @param force Clone all nodes, even shared ones.
         */
        greedyClone(force) {
            var result = new BTree(undefined, this._compare, this._maxNodeSize);
            result._root = this._root.greedyClone(force);
            result._size = this._size;
            return result;
        }
        /** Gets an array filled with the contents of the tree, sorted by key */
        toArray(maxLength = 0x7fffffff) {
            let min = this.minKey(), max = this.maxKey();
            if (min !== undefined)
                return this.getRange(min, max, true, maxLength);
            return [];
        }
        /** Gets an array of all keys, sorted */
        keysArray() {
            var results = [];
            this._root.forRange(this.minKey(), this.maxKey(), true, false, this, 0, (k, v) => {
                results.push(k);
            });
            return results;
        }
        /** Gets an array of all values, sorted by key */
        valuesArray() {
            var results = [];
            this._root.forRange(this.minKey(), this.maxKey(), true, false, this, 0, (k, v) => {
                results.push(v);
            });
            return results;
        }
        /** Gets a string representing the tree's data based on toArray(). */
        toString() {
            return this.toArray().toString();
        }
        /** Stores a key-value pair only if the key doesn't already exist in the tree.
         * @returns true if a new key was added
         */
        setIfNotPresent(key, value) {
            return this.set(key, value, false);
        }
        /** Returns the next pair whose key is larger than the specified key (or undefined if there is none) */
        nextHigherPair(key) {
            var it = this.entries(key, ReusedArray);
            var r = it.next();
            if (!r.done && this._compare(r.value[0], key) <= 0)
                r = it.next();
            return r.value;
        }
        /** Returns the next key larger than the specified key (or undefined if there is none) */
        nextHigherKey(key) {
            var p = this.nextHigherPair(key);
            return p ? p[0] : p;
        }
        /** Returns the next pair whose key is smaller than the specified key (or undefined if there is none) */
        nextLowerPair(key) {
            var it = this.entriesReversed(key, ReusedArray, true);
            return it.next().value;
        }
        /** Returns the next key smaller than the specified key (or undefined if there is none) */
        nextLowerKey(key) {
            var p = this.nextLowerPair(key);
            return p ? p[0] : p;
        }
        /** Edits the value associated with a key in the tree, if it already exists.
         * @returns true if the key existed, false if not.
         */
        changeIfPresent(key, value) {
            return this.editRange(key, key, true, (k, v) => ({ value })) !== 0;
        }
        /**
         * Builds an array of pairs from the specified range of keys, sorted by key.
         * Each returned pair is also an array: pair[0] is the key, pair[1] is the value.
         * @param low The first key in the array will be greater than or equal to `low`.
         * @param high This method returns when a key larger than this is reached.
         * @param includeHigh If the `high` key is present, its pair will be included
         *        in the output if and only if this parameter is true. Note: if the
         *        `low` key is present, it is always included in the output.
         * @param maxLength Length limit. getRange will stop scanning the tree when
         *                  the array reaches this size.
         * @description Computational complexity: O(result.length + log size)
         */
        getRange(low, high, includeHigh, maxLength = 0x3ffffff) {
            var results = [];
            this._root.forRange(low, high, includeHigh, false, this, 0, (k, v) => {
                results.push([k, v]);
                return results.length > maxLength ? Break : undefined;
            });
            return results;
        }
        /** Adds all pairs from a list of key-value pairs.
         * @param pairs Pairs to add to this tree. If there are duplicate keys,
         *        later pairs currently overwrite earlier ones (e.g. [[0,1],[0,7]]
         *        associates 0 with 7.)
         * @param overwrite Whether to overwrite pairs that already exist (if false,
         *        pairs[i] is ignored when the key pairs[i][0] already exists.)
         * @returns The number of pairs added to the collection.
         * @description Computational complexity: O(pairs.length * log(size + pairs.length))
         */
        setPairs(pairs, overwrite) {
            var added = 0;
            for (var i = 0; i < pairs.length; i++)
                if (this.set(pairs[i][0], pairs[i][1], overwrite))
                    added++;
            return added;
        }
        /**
         * Scans the specified range of keys, in ascending order by key.
         * Note: the callback `onFound` must not insert or remove items in the
         * collection. Doing so may cause incorrect data to be sent to the
         * callback afterward.
         * @param low The first key scanned will be greater than or equal to `low`.
         * @param high Scanning stops when a key larger than this is reached.
         * @param includeHigh If the `high` key is present, `onFound` is called for
         *        that final pair if and only if this parameter is true.
         * @param onFound A function that is called for each key-value pair. This
         *        function can return {break:R} to stop early with result R.
         * @param initialCounter Initial third argument of onFound. This value
         *        increases by one each time `onFound` is called. Default: 0
         * @returns The number of values found, or R if the callback returned
         *        `{break:R}` to stop early.
         * @description Computational complexity: O(number of items scanned + log size)
         */
        forRange(low, high, includeHigh, onFound, initialCounter) {
            var r = this._root.forRange(low, high, includeHigh, false, this, initialCounter || 0, onFound);
            return typeof r === "number" ? r : r.break;
        }
        /**
         * Scans and potentially modifies values for a subsequence of keys.
         * Note: the callback `onFound` should ideally be a pure function.
         *   Specfically, it must not insert items, call clone(), or change
         *   the collection except via return value; out-of-band editing may
         *   cause an exception or may cause incorrect data to be sent to
         *   the callback (duplicate or missed items). It must not cause a
         *   clone() of the collection, otherwise the clone could be modified
         *   by changes requested by the callback.
         * @param low The first key scanned will be greater than or equal to `low`.
         * @param high Scanning stops when a key larger than this is reached.
         * @param includeHigh If the `high` key is present, `onFound` is called for
         *        that final pair if and only if this parameter is true.
         * @param onFound A function that is called for each key-value pair. This
         *        function can return `{value:v}` to change the value associated
         *        with the current key, `{delete:true}` to delete the current pair,
         *        `{break:R}` to stop early with result R, or it can return nothing
         *        (undefined or {}) to cause no effect and continue iterating.
         *        `{break:R}` can be combined with one of the other two commands.
         *        The third argument `counter` is the number of items iterated
         *        previously; it equals 0 when `onFound` is called the first time.
         * @returns The number of values scanned, or R if the callback returned
         *        `{break:R}` to stop early.
         * @description
         *   Computational complexity: O(number of items scanned + log size)
         *   Note: if the tree has been cloned with clone(), any shared
         *   nodes are copied before `onFound` is called. This takes O(n) time
         *   where n is proportional to the amount of shared data scanned.
         */
        editRange(low, high, includeHigh, onFound, initialCounter) {
            var root = this._root;
            if (root.isShared)
                this._root = root = root.clone();
            try {
                var r = root.forRange(low, high, includeHigh, true, this, initialCounter || 0, onFound);
                return typeof r === "number" ? r : r.break;
            }
            finally {
                while (root.keys.length <= 1 && !root.isLeaf)
                    this._root = root =
                        root.keys.length === 0
                            ? EmptyLeaf
                            : root.children[0];
            }
        }
        /** Same as `editRange` except that the callback is called for all pairs. */
        editAll(onFound, initialCounter) {
            return this.editRange(this.minKey(), this.maxKey(), true, onFound, initialCounter);
        }
        /**
         * Removes a range of key-value pairs from the B+ tree.
         * @param low The first key scanned will be greater than or equal to `low`.
         * @param high Scanning stops when a key larger than this is reached.
         * @param includeHigh Specifies whether the `high` key, if present, is deleted.
         * @returns The number of key-value pairs that were deleted.
         * @description Computational complexity: O(log size + number of items deleted)
         */
        deleteRange(low, high, includeHigh) {
            return this.editRange(low, high, includeHigh, DeleteRange);
        }
        /** Deletes a series of keys from the collection. */
        deleteKeys(keys) {
            for (var i = 0, r = 0; i < keys.length; i++)
                if (this.delete(keys[i]))
                    r++;
            return r;
        }
        /** Gets the height of the tree: the number of internal nodes between the
         *  BTree object and its leaf nodes (zero if there are no internal nodes). */
        get height() {
            for (var node = this._root, h = -1; node != null; h++)
                node = node.children;
            return h;
        }
        /** Makes the object read-only to ensure it is not accidentally modified.
         *  Freezing does not have to be permanent; unfreeze() reverses the effect.
         *  This is accomplished by replacing mutator functions with a function
         *  that throws an Error. Compared to using a property (e.g. this.isFrozen)
         *  this implementation gives better performance in non-frozen BTrees.
         */
        freeze() {
            var t = this;
            // Note: all other mutators ultimately call set() or editRange()
            //       so we don't need to override those others.
            t.clear = t.set = t.editRange = function () {
                throw new Error("Attempted to modify a frozen BTree");
            };
        }
        /** Ensures mutations are allowed, reversing the effect of freeze(). */
        unfreeze() {
            // @ts-ignore
            delete this.clear;
            // @ts-ignore
            delete this.set;
            // @ts-ignore
            delete this.editRange;
        }
        /** Returns true if the tree appears to be frozen. */
        get isFrozen() {
            return this.hasOwnProperty("editRange");
        }
        /** Scans the tree for signs of serious bugs (e.g. this.size doesn't match
         *  number of elements, internal nodes not caching max element properly...)
         *  Computational complexity: O(number of nodes), i.e. O(size). This method
         *  skips the most expensive test - whether all keys are sorted - but it
         *  does check that maxKey() of the children of internal nodes are sorted. */
        checkValid() {
            var size = this._root.checkValid(0, this);
            check$1(size === this.size, "size mismatch: counted ", size, "but stored", this.size);
        }
    }
    if (Symbol && Symbol.iterator)
        // iterator is equivalent to entries()
        BTree.prototype[Symbol.iterator] = BTree.prototype.entries;
    BTree.prototype.where = BTree.prototype.filter;
    BTree.prototype.setRange = BTree.prototype.setPairs;
    BTree.prototype.add = BTree.prototype.set;
    function iterator(next = () => ({
        done: true,
        value: undefined,
    })) {
        var result = { next };
        if (Symbol && Symbol.iterator)
            result[Symbol.iterator] = function () {
                return this;
            };
        return result;
    }
    /** Leaf node / base class. **************************************************/
    class BNode {
        constructor(keys = [], values) {
            this.keys = keys;
            this.values = values || undefVals;
            this.isShared = undefined;
        }
        get isLeaf() {
            return this.children === undefined;
        }
        // Shared methods /////////////////////////////////////////////////////////
        maxKey() {
            return this.keys[this.keys.length - 1];
        }
        // If key not found, returns i^failXor where i is the insertion index.
        // Callers that don't care whether there was a match will set failXor=0.
        indexOf(key, failXor, cmp) {
            // TODO: benchmark multiple search strategies
            const keys = this.keys;
            var lo = 0, hi = keys.length, mid = hi >> 1;
            while (lo < hi) {
                var c = cmp(keys[mid], key);
                if (c < 0)
                    lo = mid + 1;
                else if (c > 0)
                    // key < keys[mid]
                    hi = mid;
                else if (c === 0)
                    return mid;
                else {
                    // c is NaN or otherwise invalid
                    if (key === key)
                        // at least the search key is not NaN
                        return keys.length;
                    else
                        throw new Error("BTree: NaN was used as a key");
                }
                mid = (lo + hi) >> 1;
            }
            return mid ^ failXor;
            // Unrolled version: benchmarks show same speed, not worth using
            /*var i = 1, c: number = 0, sum = 0;
            if (keys.length >= 4) {
              i = 3;
              if (keys.length >= 8) {
                i = 7;
                if (keys.length >= 16) {
                  i = 15;
                  if (keys.length >= 32) {
                    i = 31;
                    if (keys.length >= 64) {
                      i = 127;
                      i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 64 : -64;
                      sum += c;
                      i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 32 : -32;
                      sum += c;
                    }
                    i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 16 : -16;
                    sum += c;
                  }
                  i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 8 : -8;
                  sum += c;
                }
                i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 4 : -4;
                sum += c;
              }
              i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 2 : -2;
              sum += c;
            }
            i += (c = i < keys.length ? cmp(keys[i], key) : 1) < 0 ? 1 : -1;
            c = i < keys.length ? cmp(keys[i], key) : 1;
            sum += c;
            if (c < 0) {
              ++i;
              c = i < keys.length ? cmp(keys[i], key) : 1;
              sum += c;
            }
            if (sum !== sum) {
              if (key === key) // at least the search key is not NaN
                return keys.length ^ failXor;
              else
                throw new Error("BTree: NaN was used as a key");
            }
            return c === 0 ? i : i ^ failXor;*/
        }
        // Leaf Node: misc //////////////////////////////////////////////////////////
        minKey() {
            return this.keys[0];
        }
        clone() {
            var v = this.values;
            return new BNode(this.keys.slice(0), v === undefVals ? v : v.slice(0));
        }
        greedyClone(force) {
            return this.isShared && !force ? this : this.clone();
        }
        get(key, defaultValue, tree) {
            var i = this.indexOf(key, -1, tree._compare);
            return i < 0 ? defaultValue : this.values[i];
        }
        checkValid(depth, tree) {
            var kL = this.keys.length, vL = this.values.length;
            check$1(this.values === undefVals ? kL <= vL : kL === vL, "keys/values length mismatch: depth", depth, "with lengths", kL, vL);
            // Note: we don't check for "node too small" because sometimes a node
            // can legitimately have size 1. This occurs if there is a batch
            // deletion, leaving a node of size 1, and the siblings are full so
            // it can't be merged with adjacent nodes. However, the parent will
            // verify that the average node size is at least half of the maximum.
            check$1(depth == 0 || kL > 0, "empty leaf at depth", depth);
            return kL;
        }
        // Leaf Node: set & node splitting //////////////////////////////////////////
        set(key, value, overwrite, tree) {
            var i = this.indexOf(key, -1, tree._compare);
            if (i < 0) {
                // key does not exist yet
                i = ~i;
                tree._size++;
                if (this.keys.length < tree._maxNodeSize) {
                    return this.insertInLeaf(i, key, value, tree);
                }
                else {
                    // This leaf node is full and must split
                    var newRightSibling = this.splitOffRightSide(), target = this;
                    if (i > this.keys.length) {
                        i -= this.keys.length;
                        target = newRightSibling;
                    }
                    target.insertInLeaf(i, key, value, tree);
                    return newRightSibling;
                }
            }
            else {
                // Key already exists
                if (overwrite !== false) {
                    if (value !== undefined)
                        this.reifyValues();
                    // usually this is a no-op, but some users may wish to edit the key
                    this.keys[i] = key;
                    this.values[i] = value;
                }
                return false;
            }
        }
        reifyValues() {
            if (this.values === undefVals)
                return (this.values = this.values.slice(0, this.keys.length));
            return this.values;
        }
        insertInLeaf(i, key, value, tree) {
            this.keys.splice(i, 0, key);
            if (this.values === undefVals) {
                while (undefVals.length < tree._maxNodeSize)
                    undefVals.push(undefined);
                if (value === undefined) {
                    return true;
                }
                else {
                    this.values = undefVals.slice(0, this.keys.length - 1);
                }
            }
            this.values.splice(i, 0, value);
            return true;
        }
        takeFromRight(rhs) {
            // Reminder: parent node must update its copy of key for this node
            // assert: neither node is shared
            // assert rhs.keys.length > (maxNodeSize/2 && this.keys.length<maxNodeSize)
            var v = this.values;
            if (rhs.values === undefVals) {
                if (v !== undefVals)
                    v.push(undefined);
            }
            else {
                v = this.reifyValues();
                v.push(rhs.values.shift());
            }
            this.keys.push(rhs.keys.shift());
        }
        takeFromLeft(lhs) {
            // Reminder: parent node must update its copy of key for this node
            // assert: neither node is shared
            // assert rhs.keys.length > (maxNodeSize/2 && this.keys.length<maxNodeSize)
            var v = this.values;
            if (lhs.values === undefVals) {
                if (v !== undefVals)
                    v.unshift(undefined);
            }
            else {
                v = this.reifyValues();
                v.unshift(lhs.values.pop());
            }
            this.keys.unshift(lhs.keys.pop());
        }
        splitOffRightSide() {
            // Reminder: parent node must update its copy of key for this node
            var half = this.keys.length >> 1, keys = this.keys.splice(half);
            var values = this.values === undefVals ? undefVals : this.values.splice(half);
            return new BNode(keys, values);
        }
        // Leaf Node: scanning & deletions //////////////////////////////////////////
        forRange(low, high, includeHigh, editMode, tree, count, onFound) {
            var cmp = tree._compare;
            var iLow, iHigh;
            if (high === low) {
                if (!includeHigh)
                    return count;
                iHigh = (iLow = this.indexOf(low, -1, cmp)) + 1;
                if (iLow < 0)
                    return count;
            }
            else {
                iLow = this.indexOf(low, 0, cmp);
                iHigh = this.indexOf(high, -1, cmp);
                if (iHigh < 0)
                    iHigh = ~iHigh;
                else if (includeHigh === true)
                    iHigh++;
            }
            var keys = this.keys, values = this.values;
            if (onFound !== undefined) {
                for (var i = iLow; i < iHigh; i++) {
                    var key = keys[i];
                    var result = onFound(key, values[i], count++);
                    if (result !== undefined) {
                        if (editMode === true) {
                            if (key !== keys[i] || this.isShared === true)
                                throw new Error("BTree illegally changed or cloned in editRange");
                            if (result.delete) {
                                this.keys.splice(i, 1);
                                if (this.values !== undefVals)
                                    this.values.splice(i, 1);
                                tree._size--;
                                i--;
                                iHigh--;
                            }
                            else if (result.hasOwnProperty("value")) {
                                values[i] = result.value;
                            }
                        }
                        if (result.break !== undefined)
                            return result;
                    }
                }
            }
            else
                count += iHigh - iLow;
            return count;
        }
        /** Adds entire contents of right-hand sibling (rhs is left unchanged) */
        mergeSibling(rhs, _) {
            this.keys.push.apply(this.keys, rhs.keys);
            if (this.values === undefVals) {
                if (rhs.values === undefVals)
                    return;
                this.values = this.values.slice(0, this.keys.length);
            }
            this.values.push.apply(this.values, rhs.reifyValues());
        }
    }
    /** Internal node (non-leaf node) ********************************************/
    class BNodeInternal extends BNode {
        constructor(children, keys) {
            if (!keys) {
                keys = [];
                for (var i = 0; i < children.length; i++)
                    keys[i] = children[i].maxKey();
            }
            super(keys);
            this.children = children;
        }
        clone() {
            var children = this.children.slice(0);
            for (var i = 0; i < children.length; i++)
                children[i].isShared = true;
            return new BNodeInternal(children, this.keys.slice(0));
        }
        greedyClone(force) {
            if (this.isShared && !force)
                return this;
            var nu = new BNodeInternal(this.children.slice(0), this.keys.slice(0));
            for (var i = 0; i < nu.children.length; i++)
                nu.children[i] = nu.children[i].greedyClone();
            return nu;
        }
        minKey() {
            return this.children[0].minKey();
        }
        get(key, defaultValue, tree) {
            var i = this.indexOf(key, 0, tree._compare), children = this.children;
            return i < children.length
                ? children[i].get(key, defaultValue, tree)
                : undefined;
        }
        checkValid(depth, tree) {
            var kL = this.keys.length, cL = this.children.length;
            check$1(kL === cL, "keys/children length mismatch: depth", depth, "lengths", kL, cL);
            check$1(kL > 1, "internal node has length", kL, "at depth", depth);
            var size = 0, c = this.children, k = this.keys, childSize = 0;
            for (var i = 0; i < cL; i++) {
                size += c[i].checkValid(depth + 1, tree);
                childSize += c[i].keys.length;
                check$1(size >= childSize, "wtf"); // no way this will ever fail
                check$1(i === 0 || c[i - 1].constructor === c[i].constructor, "type mismatch");
                if (c[i].maxKey() != k[i])
                    check$1(false, "keys[", i, "] =", k[i], "is wrong, should be ", c[i].maxKey(), "at depth", depth);
                if (!(i === 0 || tree._compare(k[i - 1], k[i]) < 0))
                    check$1(false, "sort violation at depth", depth, "index", i, "keys", k[i - 1], k[i]);
            }
            var toofew = childSize < (tree.maxNodeSize >> 1) * cL;
            if (toofew || childSize > tree.maxNodeSize * cL)
                check$1(false, toofew ? "too few" : "too many", "children (", childSize, size, ") at depth", depth, ", maxNodeSize:", tree.maxNodeSize, "children.length:", cL);
            return size;
        }
        // Internal Node: set & node splitting //////////////////////////////////////
        set(key, value, overwrite, tree) {
            var c = this.children, max = tree._maxNodeSize, cmp = tree._compare;
            var i = Math.min(this.indexOf(key, 0, cmp), c.length - 1), child = c[i];
            if (child.isShared)
                c[i] = child = child.clone();
            if (child.keys.length >= max) {
                // child is full; inserting anything else will cause a split.
                // Shifting an item to the left or right sibling may avoid a split.
                // We can do a shift if the adjacent node is not full and if the
                // current key can still be placed in the same node after the shift.
                var other;
                if (i > 0 &&
                    (other = c[i - 1]).keys.length < max &&
                    cmp(child.keys[0], key) < 0) {
                    if (other.isShared)
                        c[i - 1] = other = other.clone();
                    other.takeFromRight(child);
                    this.keys[i - 1] = other.maxKey();
                }
                else if ((other = c[i + 1]) !== undefined &&
                    other.keys.length < max &&
                    cmp(child.maxKey(), key) < 0) {
                    if (other.isShared)
                        c[i + 1] = other = other.clone();
                    other.takeFromLeft(child);
                    this.keys[i] = c[i].maxKey();
                }
            }
            var result = child.set(key, value, overwrite, tree);
            if (result === false)
                return false;
            this.keys[i] = child.maxKey();
            if (result === true)
                return true;
            // The child has split and `result` is a new right child... does it fit?
            if (this.keys.length < max) {
                // yes
                this.insert(i + 1, result);
                return true;
            }
            else {
                // no, we must split also
                var newRightSibling = this.splitOffRightSide(), target = this;
                if (cmp(result.maxKey(), this.maxKey()) > 0) {
                    target = newRightSibling;
                    i -= this.keys.length;
                }
                target.insert(i + 1, result);
                return newRightSibling;
            }
        }
        insert(i, child) {
            this.children.splice(i, 0, child);
            this.keys.splice(i, 0, child.maxKey());
        }
        splitOffRightSide() {
            var half = this.children.length >> 1;
            return new BNodeInternal(this.children.splice(half), this.keys.splice(half));
        }
        takeFromRight(rhs) {
            // Reminder: parent node must update its copy of key for this node
            // assert: neither node is shared
            // assert rhs.keys.length > (maxNodeSize/2 && this.keys.length<maxNodeSize)
            this.keys.push(rhs.keys.shift());
            this.children.push(rhs.children.shift());
        }
        takeFromLeft(lhs) {
            // Reminder: parent node must update its copy of key for this node
            // assert: neither node is shared
            // assert rhs.keys.length > (maxNodeSize/2 && this.keys.length<maxNodeSize)
            this.keys.unshift(lhs.keys.pop());
            this.children.unshift(lhs.children.pop());
        }
        // Internal Node: scanning & deletions //////////////////////////////////////
        forRange(low, high, includeHigh, editMode, tree, count, onFound) {
            var cmp = tree._compare;
            var iLow = this.indexOf(low, 0, cmp), i = iLow;
            var iHigh = Math.min(high === low ? iLow : this.indexOf(high, 0, cmp), this.keys.length - 1);
            var keys = this.keys, children = this.children;
            if (!editMode) {
                // Simple case
                for (; i <= iHigh; i++) {
                    var result = children[i].forRange(low, high, includeHigh, editMode, tree, count, onFound);
                    if (typeof result !== "number")
                        return result;
                    count = result;
                }
            }
            else if (i <= iHigh) {
                try {
                    for (; i <= iHigh; i++) {
                        if (children[i].isShared)
                            children[i] = children[i].clone();
                        var result = children[i].forRange(low, high, includeHigh, editMode, tree, count, onFound);
                        keys[i] = children[i].maxKey();
                        if (typeof result !== "number")
                            return result;
                        count = result;
                    }
                }
                finally {
                    // Deletions may have occurred, so look for opportunities to merge nodes.
                    var half = tree._maxNodeSize >> 1;
                    if (iLow > 0)
                        iLow--;
                    for (i = iHigh; i >= iLow; i--) {
                        if (children[i].keys.length <= half)
                            this.tryMerge(i, tree._maxNodeSize);
                    }
                    // Are we completely empty?
                    if (children[0].keys.length === 0) {
                        check$1(children.length === 1 && keys.length === 1, "emptiness bug");
                        children.shift();
                        keys.shift();
                    }
                }
            }
            return count;
        }
        /** Merges child i with child i+1 if their combined size is not too large */
        tryMerge(i, maxSize) {
            var children = this.children;
            if (i >= 0 && i + 1 < children.length) {
                if (children[i].keys.length + children[i + 1].keys.length <= maxSize) {
                    if (children[i].isShared)
                        // cloned already UNLESS i is outside scan range
                        children[i] = children[i].clone();
                    children[i].mergeSibling(children[i + 1], maxSize);
                    children.splice(i + 1, 1);
                    this.keys.splice(i + 1, 1);
                    this.keys[i] = children[i].maxKey();
                    return true;
                }
            }
            return false;
        }
        mergeSibling(rhs, maxNodeSize) {
            // assert !this.isShared;
            var oldLength = this.keys.length;
            this.keys.push.apply(this.keys, rhs.keys);
            this.children.push.apply(this.children, rhs.children);
            // If our children are themselves almost empty due to a mass-delete,
            // they may need to be merged too (but only the oldLength-1 and its
            // right sibling should need this).
            this.tryMerge(oldLength - 1, maxNodeSize);
        }
    }
    // Optimization: this array of `undefined`s is used instead of a normal
    // array of values in nodes where `undefined` is the only value.
    // Its length is extended to max node size on first use; since it can
    // be shared between trees with different maximums, its length can only
    // increase, never decrease. Its type should be undefined[] but strangely
    // TypeScript won't allow the comparison V[] === undefined[]. To prevent
    // users from making this array too large, BTree has a maximum node size.
    var undefVals = [];
    const Delete = { delete: true }, DeleteRange = () => Delete;
    const Break = { break: true };
    const EmptyLeaf = (function () {
        var n = new BNode();
        n.isShared = true;
        return n;
    })();
    const EmptyArray = [];
    const ReusedArray = []; // assumed thread-local
    function check$1(fact, ...args) {
        if (!fact) {
            args.unshift("B+ tree "); // at beginning of message
            throw new Error(args.join(" "));
        }
    }
    /** A BTree frozen in the empty state. */
    const EmptyBTree = (() => {
        let t = new BTree();
        t.freeze();
        return t;
    })();

    /*
     Copyright 2019 Florian Dold

     Licensed under the Apache License, Version 2.0 (the "License");
     you may not use this file except in compliance with the License.
     You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

     Unless required by applicable law or agreed to in writing, software
     distributed under the License is distributed on an "AS IS" BASIS,
     WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
     or implied. See the License for the specific language governing
     permissions and limitations under the License.
     */
    var TransactionLevel;
    (function (TransactionLevel) {
        TransactionLevel[TransactionLevel["Disconnected"] = 0] = "Disconnected";
        TransactionLevel[TransactionLevel["Connected"] = 1] = "Connected";
        TransactionLevel[TransactionLevel["Read"] = 2] = "Read";
        TransactionLevel[TransactionLevel["Write"] = 3] = "Write";
        TransactionLevel[TransactionLevel["VersionChange"] = 4] = "VersionChange";
    })(TransactionLevel || (TransactionLevel = {}));

    // globalThis polyfill, see https://mathiasbynens.be/notes/globalthis
    (function () {
        if (typeof globalThis === "object")
            return;
        Object.defineProperty(Object.prototype, "__magic__", {
            get: function () {
                return this;
            },
            configurable: true,
        });
        // @ts-ignore: polyfill magic
        __magic__.globalThis = __magic__; // lolwat
        // @ts-ignore: polyfill magic
        delete Object.prototype.__magic__;
    })();

    /*
     This file is part of GNU Taler
     (C) 2019-2020 Taler Systems SA

     TALER is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     TALER is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
     */
    const logger = new Logger("cryptoImplementation.ts");
    var SignaturePurpose;
    (function (SignaturePurpose) {
        SignaturePurpose[SignaturePurpose["WALLET_RESERVE_WITHDRAW"] = 1200] = "WALLET_RESERVE_WITHDRAW";
        SignaturePurpose[SignaturePurpose["WALLET_COIN_DEPOSIT"] = 1201] = "WALLET_COIN_DEPOSIT";
        SignaturePurpose[SignaturePurpose["MASTER_DENOMINATION_KEY_VALIDITY"] = 1025] = "MASTER_DENOMINATION_KEY_VALIDITY";
        SignaturePurpose[SignaturePurpose["MASTER_WIRE_FEES"] = 1028] = "MASTER_WIRE_FEES";
        SignaturePurpose[SignaturePurpose["MASTER_WIRE_DETAILS"] = 1030] = "MASTER_WIRE_DETAILS";
        SignaturePurpose[SignaturePurpose["WALLET_COIN_MELT"] = 1202] = "WALLET_COIN_MELT";
        SignaturePurpose[SignaturePurpose["TEST"] = 4242] = "TEST";
        SignaturePurpose[SignaturePurpose["MERCHANT_PAYMENT_OK"] = 1104] = "MERCHANT_PAYMENT_OK";
        SignaturePurpose[SignaturePurpose["MERCHANT_CONTRACT"] = 1101] = "MERCHANT_CONTRACT";
        SignaturePurpose[SignaturePurpose["WALLET_COIN_RECOUP"] = 1203] = "WALLET_COIN_RECOUP";
        SignaturePurpose[SignaturePurpose["WALLET_COIN_LINK"] = 1204] = "WALLET_COIN_LINK";
        SignaturePurpose[SignaturePurpose["EXCHANGE_CONFIRM_RECOUP"] = 1039] = "EXCHANGE_CONFIRM_RECOUP";
        SignaturePurpose[SignaturePurpose["EXCHANGE_CONFIRM_RECOUP_REFRESH"] = 1041] = "EXCHANGE_CONFIRM_RECOUP_REFRESH";
        SignaturePurpose[SignaturePurpose["SYNC_BACKUP_UPLOAD"] = 1450] = "SYNC_BACKUP_UPLOAD";
    })(SignaturePurpose || (SignaturePurpose = {}));
    function amountToBuffer(amount) {
        const buffer = new ArrayBuffer(8 + 4 + 12);
        const dvbuf = new DataView(buffer);
        const u8buf = new Uint8Array(buffer);
        const curr = stringToBytes(amount.currency);
        dvbuf.setBigUint64(0, BigInt(amount.value));
        dvbuf.setUint32(8, amount.fraction);
        u8buf.set(curr, 8 + 4);
        return u8buf;
    }
    function timestampRoundedToBuffer(ts) {
        const b = new ArrayBuffer(8);
        const v = new DataView(b);
        const tsRounded = timestampTruncateToSecond(ts);
        const s = BigInt(tsRounded.t_ms) * BigInt(1000);
        v.setBigUint64(0, s);
        return new Uint8Array(b);
    }
    class SignaturePurposeBuilder {
        constructor(purposeNum) {
            this.purposeNum = purposeNum;
            this.chunks = [];
        }
        put(bytes) {
            this.chunks.push(Uint8Array.from(bytes));
            return this;
        }
        build() {
            let payloadLen = 0;
            for (const c of this.chunks) {
                payloadLen += c.byteLength;
            }
            const buf = new ArrayBuffer(4 + 4 + payloadLen);
            const u8buf = new Uint8Array(buf);
            let p = 8;
            for (const c of this.chunks) {
                u8buf.set(c, p);
                p += c.byteLength;
            }
            const dvbuf = new DataView(buf);
            dvbuf.setUint32(0, payloadLen + 4 + 4);
            dvbuf.setUint32(4, this.purposeNum);
            return u8buf;
        }
    }
    function buildSigPS(purposeNum) {
        return new SignaturePurposeBuilder(purposeNum);
    }
    class CryptoImplementation {
        /**
         * Create a pre-coin of the given denomination to be withdrawn from then given
         * reserve.
         */
        createPlanchet(req) {
            const reservePub = decodeCrock(req.reservePub);
            const reservePriv = decodeCrock(req.reservePriv);
            const denomPub = decodeCrock(req.denomPub);
            const derivedPlanchet = setupWithdrawPlanchet(decodeCrock(req.secretSeed), req.coinIndex);
            const coinPubHash = hash$1(derivedPlanchet.coinPub);
            const ev = rsaBlind(coinPubHash, derivedPlanchet.bks, denomPub);
            const amountWithFee = Amounts.add(req.value, req.feeWithdraw).amount;
            const denomPubHash = hash$1(denomPub);
            const evHash = hash$1(ev);
            const withdrawRequest = buildSigPS(SignaturePurpose.WALLET_RESERVE_WITHDRAW)
                .put(reservePub)
                .put(amountToBuffer(amountWithFee))
                .put(denomPubHash)
                .put(evHash)
                .build();
            const sig = eddsaSign(withdrawRequest, reservePriv);
            const planchet = {
                blindingKey: encodeCrock(derivedPlanchet.bks),
                coinEv: encodeCrock(ev),
                coinPriv: encodeCrock(derivedPlanchet.coinPriv),
                coinPub: encodeCrock(derivedPlanchet.coinPub),
                coinValue: req.value,
                denomPub: encodeCrock(denomPub),
                denomPubHash: encodeCrock(denomPubHash),
                reservePub: encodeCrock(reservePub),
                withdrawSig: encodeCrock(sig),
                coinEvHash: encodeCrock(evHash),
            };
            return planchet;
        }
        /**
         * Create a planchet used for tipping, including the private keys.
         */
        createTipPlanchet(req) {
            const fc = setupTipPlanchet(decodeCrock(req.secretSeed), req.planchetIndex);
            const denomPub = decodeCrock(req.denomPub);
            const coinPubHash = hash$1(fc.coinPub);
            const ev = rsaBlind(coinPubHash, fc.bks, denomPub);
            const tipPlanchet = {
                blindingKey: encodeCrock(fc.bks),
                coinEv: encodeCrock(ev),
                coinEvHash: encodeCrock(hash$1(ev)),
                coinPriv: encodeCrock(fc.coinPriv),
                coinPub: encodeCrock(fc.coinPub),
            };
            return tipPlanchet;
        }
        /**
         * Create and sign a message to recoup a coin.
         */
        createRecoupRequest(coin) {
            const p = buildSigPS(SignaturePurpose.WALLET_COIN_RECOUP)
                .put(decodeCrock(coin.coinPub))
                .put(decodeCrock(coin.denomPubHash))
                .put(decodeCrock(coin.blindingKey))
                .build();
            const coinPriv = decodeCrock(coin.coinPriv);
            const coinSig = eddsaSign(p, coinPriv);
            const paybackRequest = {
                coin_blind_key_secret: coin.blindingKey,
                coin_pub: coin.coinPub,
                coin_sig: encodeCrock(coinSig),
                denom_pub_hash: coin.denomPubHash,
                denom_sig: coin.denomSig,
                refreshed: coin.coinSource.type === CoinSourceType.Refresh,
            };
            return paybackRequest;
        }
        /**
         * Check if a payment signature is valid.
         */
        isValidPaymentSignature(sig, contractHash, merchantPub) {
            const p = buildSigPS(SignaturePurpose.MERCHANT_PAYMENT_OK)
                .put(decodeCrock(contractHash))
                .build();
            const sigBytes = decodeCrock(sig);
            const pubBytes = decodeCrock(merchantPub);
            return eddsaVerify(p, sigBytes, pubBytes);
        }
        /**
         * Check if a wire fee is correctly signed.
         */
        isValidWireFee(type, wf, masterPub) {
            const p = buildSigPS(SignaturePurpose.MASTER_WIRE_FEES)
                .put(hash$1(stringToBytes(type + "\0")))
                .put(timestampRoundedToBuffer(wf.startStamp))
                .put(timestampRoundedToBuffer(wf.endStamp))
                .put(amountToBuffer(wf.wireFee))
                .put(amountToBuffer(wf.closingFee))
                .build();
            const sig = decodeCrock(wf.sig);
            const pub = decodeCrock(masterPub);
            return eddsaVerify(p, sig, pub);
        }
        /**
         * Check if the signature of a denomination is valid.
         */
        isValidDenom(denom, masterPub) {
            const p = buildSigPS(SignaturePurpose.MASTER_DENOMINATION_KEY_VALIDITY)
                .put(decodeCrock(masterPub))
                .put(timestampRoundedToBuffer(denom.stampStart))
                .put(timestampRoundedToBuffer(denom.stampExpireWithdraw))
                .put(timestampRoundedToBuffer(denom.stampExpireDeposit))
                .put(timestampRoundedToBuffer(denom.stampExpireLegal))
                .put(amountToBuffer(denom.value))
                .put(amountToBuffer(denom.feeWithdraw))
                .put(amountToBuffer(denom.feeDeposit))
                .put(amountToBuffer(denom.feeRefresh))
                .put(amountToBuffer(denom.feeRefund))
                .put(decodeCrock(denom.denomPubHash))
                .build();
            const sig = decodeCrock(denom.masterSig);
            const pub = decodeCrock(masterPub);
            return eddsaVerify(p, sig, pub);
        }
        isValidWireAccount(paytoUri, sig, masterPub) {
            const h = kdf(64, stringToBytes("exchange-wire-signature"), stringToBytes(paytoUri + "\0"), new Uint8Array(0));
            const p = buildSigPS(SignaturePurpose.MASTER_WIRE_DETAILS).put(h).build();
            return eddsaVerify(p, decodeCrock(sig), decodeCrock(masterPub));
        }
        isValidContractTermsSignature(contractTermsHash, sig, merchantPub) {
            const cthDec = decodeCrock(contractTermsHash);
            const p = buildSigPS(SignaturePurpose.MERCHANT_CONTRACT)
                .put(cthDec)
                .build();
            return eddsaVerify(p, decodeCrock(sig), decodeCrock(merchantPub));
        }
        /**
         * Create a new EdDSA key pair.
         */
        createEddsaKeypair() {
            const pair = createEddsaKeyPair();
            return {
                priv: encodeCrock(pair.eddsaPriv),
                pub: encodeCrock(pair.eddsaPub),
            };
        }
        /**
         * Unblind a blindly signed value.
         */
        rsaUnblind(blindedSig, bk, pk) {
            const denomSig = rsaUnblind(decodeCrock(blindedSig), decodeCrock(pk), decodeCrock(bk));
            return encodeCrock(denomSig);
        }
        /**
         * Unblind a blindly signed value.
         */
        rsaVerify(hm, sig, pk) {
            return rsaVerify(hash$1(decodeCrock(hm)), decodeCrock(sig), decodeCrock(pk));
        }
        /**
         * Generate updated coins (to store in the database)
         * and deposit permissions for each given coin.
         */
        signDepositPermission(depositInfo) {
            const d = buildSigPS(SignaturePurpose.WALLET_COIN_DEPOSIT)
                .put(decodeCrock(depositInfo.contractTermsHash))
                .put(decodeCrock(depositInfo.wireInfoHash))
                .put(decodeCrock(depositInfo.denomPubHash))
                .put(timestampRoundedToBuffer(depositInfo.timestamp))
                .put(timestampRoundedToBuffer(depositInfo.refundDeadline))
                .put(amountToBuffer(depositInfo.spendAmount))
                .put(amountToBuffer(depositInfo.feeDeposit))
                .put(decodeCrock(depositInfo.merchantPub))
                .put(decodeCrock(depositInfo.coinPub))
                .build();
            const coinSig = eddsaSign(d, decodeCrock(depositInfo.coinPriv));
            const s = {
                coin_pub: depositInfo.coinPub,
                coin_sig: encodeCrock(coinSig),
                contribution: Amounts.stringify(depositInfo.spendAmount),
                h_denom: depositInfo.denomPubHash,
                exchange_url: depositInfo.exchangeBaseUrl,
                ub_sig: depositInfo.denomSig,
            };
            return s;
        }
        deriveRefreshSession(req) {
            const { newCoinDenoms, feeRefresh: meltFee, kappa, meltCoinDenomPubHash, meltCoinPriv, meltCoinPub, sessionSecretSeed: refreshSessionSecretSeed, } = req;
            const currency = newCoinDenoms[0].value.currency;
            let valueWithFee = Amounts.getZero(currency);
            for (const ncd of newCoinDenoms) {
                const t = Amounts.add(ncd.value, ncd.feeWithdraw).amount;
                valueWithFee = Amounts.add(valueWithFee, Amounts.mult(t, ncd.count).amount).amount;
            }
            // melt fee
            valueWithFee = Amounts.add(valueWithFee, meltFee).amount;
            const sessionHc = createHashContext();
            const transferPubs = [];
            const transferPrivs = [];
            const planchetsForGammas = [];
            logger.trace("starting RC computation");
            for (let i = 0; i < kappa; i++) {
                const transferKeyPair = setupRefreshTransferPub(decodeCrock(refreshSessionSecretSeed), i);
                sessionHc.update(transferKeyPair.ecdhePub);
                logger.trace(`HASH transfer_pub ${encodeCrock(transferKeyPair.ecdhePub)}`);
                transferPrivs.push(encodeCrock(transferKeyPair.ecdhePriv));
                transferPubs.push(encodeCrock(transferKeyPair.ecdhePub));
            }
            for (const denomSel of newCoinDenoms) {
                for (let i = 0; i < denomSel.count; i++) {
                    const r = decodeCrock(denomSel.denomPub);
                    sessionHc.update(r);
                    logger.trace(`HASH new_coins ${encodeCrock(r)}`);
                }
            }
            sessionHc.update(decodeCrock(meltCoinPub));
            logger.trace(`HASH coin_pub ${meltCoinPub}`);
            sessionHc.update(amountToBuffer(valueWithFee));
            logger.trace(`HASH melt_amount ${encodeCrock(amountToBuffer(valueWithFee))}`);
            for (let i = 0; i < kappa; i++) {
                const planchets = [];
                for (let j = 0; j < newCoinDenoms.length; j++) {
                    const denomSel = newCoinDenoms[j];
                    for (let k = 0; k < denomSel.count; k++) {
                        const coinNumber = planchets.length;
                        const transferPriv = decodeCrock(transferPrivs[i]);
                        const oldCoinPub = decodeCrock(meltCoinPub);
                        const transferSecret = keyExchangeEcdheEddsa(transferPriv, oldCoinPub);
                        const fresh = setupRefreshPlanchet(transferSecret, coinNumber);
                        const coinPriv = fresh.coinPriv;
                        const coinPub = fresh.coinPub;
                        const blindingFactor = fresh.bks;
                        const pubHash = hash$1(coinPub);
                        const denomPub = decodeCrock(denomSel.denomPub);
                        const ev = rsaBlind(pubHash, blindingFactor, denomPub);
                        const planchet = {
                            blindingKey: encodeCrock(blindingFactor),
                            coinEv: encodeCrock(ev),
                            privateKey: encodeCrock(coinPriv),
                            publicKey: encodeCrock(coinPub),
                            coinEvHash: encodeCrock(hash$1(ev)),
                        };
                        planchets.push(planchet);
                        logger.trace(`GENERATE i=${i} coin=${coinNumber} m=${encodeCrock(pubHash)} bf=${encodeCrock(blindingFactor)} dp=${encodeCrock(denomPub)} ev=${encodeCrock(ev)}`);
                        sessionHc.update(ev);
                        logger.trace(`HASH ev ${encodeCrock(ev)}`);
                    }
                }
                planchetsForGammas.push(planchets);
            }
            const sessionHash = sessionHc.finish();
            logger.trace(`RHASH ${encodeCrock(sessionHash)}`);
            const confirmData = buildSigPS(SignaturePurpose.WALLET_COIN_MELT)
                .put(sessionHash)
                .put(decodeCrock(meltCoinDenomPubHash))
                .put(amountToBuffer(valueWithFee))
                .put(amountToBuffer(meltFee))
                .put(decodeCrock(meltCoinPub))
                .build();
            const confirmSig = eddsaSign(confirmData, decodeCrock(meltCoinPriv));
            const refreshSession = {
                confirmSig: encodeCrock(confirmSig),
                hash: encodeCrock(sessionHash),
                meltCoinPub: meltCoinPub,
                planchetsForGammas: planchetsForGammas,
                transferPrivs,
                transferPubs,
                meltValueWithFee: valueWithFee,
            };
            return refreshSession;
        }
        /**
         * Hash a string including the zero terminator.
         */
        hashString(str) {
            const b = stringToBytes(str + "\0");
            return encodeCrock(hash$1(b));
        }
        /**
         * Hash a crockford encoded value.
         */
        hashEncoded(encodedBytes) {
            return encodeCrock(hash$1(decodeCrock(encodedBytes)));
        }
        signCoinLink(oldCoinPriv, newDenomHash, oldCoinPub, transferPub, coinEv) {
            const coinEvHash = hash$1(decodeCrock(coinEv));
            const coinLink = buildSigPS(SignaturePurpose.WALLET_COIN_LINK)
                .put(decodeCrock(newDenomHash))
                .put(decodeCrock(oldCoinPub))
                .put(decodeCrock(transferPub))
                .put(coinEvHash)
                .build();
            const coinPriv = decodeCrock(oldCoinPriv);
            const sig = eddsaSign(coinLink, coinPriv);
            return encodeCrock(sig);
        }
        benchmark(repetitions) {
            let time_hash = 0;
            for (let i = 0; i < repetitions; i++) {
                const start = performanceNow();
                this.hashString("hello world");
                time_hash += performanceNow() - start;
            }
            let time_hash_big = 0;
            for (let i = 0; i < repetitions; i++) {
                const ba = randomBytes(4096);
                const start = performanceNow();
                hash$1(ba);
                time_hash_big += performanceNow() - start;
            }
            let time_eddsa_create = 0;
            for (let i = 0; i < repetitions; i++) {
                const start = performanceNow();
                createEddsaKeyPair();
                time_eddsa_create += performanceNow() - start;
            }
            let time_eddsa_sign = 0;
            const p = randomBytes(4096);
            const pair = createEddsaKeyPair();
            for (let i = 0; i < repetitions; i++) {
                const start = performanceNow();
                eddsaSign(p, pair.eddsaPriv);
                time_eddsa_sign += performanceNow() - start;
            }
            const sig = eddsaSign(p, pair.eddsaPriv);
            let time_eddsa_verify = 0;
            for (let i = 0; i < repetitions; i++) {
                const start = performanceNow();
                eddsaVerify(p, sig, pair.eddsaPub);
                time_eddsa_verify += performanceNow() - start;
            }
            return {
                repetitions,
                time: {
                    hash_small: time_hash,
                    hash_big: time_hash_big,
                    eddsa_create: time_eddsa_create,
                    eddsa_sign: time_eddsa_sign,
                    eddsa_verify: time_eddsa_verify,
                },
            };
        }
        makeSyncSignature(req) {
            const hNew = decodeCrock(req.newHash);
            let hOld;
            if (req.oldHash) {
                hOld = decodeCrock(req.oldHash);
            }
            else {
                hOld = new Uint8Array(64);
            }
            const sigBlob = new SignaturePurposeBuilder(SignaturePurpose.SYNC_BACKUP_UPLOAD)
                .put(hOld)
                .put(hNew)
                .build();
            const uploadSig = eddsaSign(sigBlob, decodeCrock(req.accountPriv));
            return encodeCrock(uploadSig);
        }
    }
    CryptoImplementation.enableTracing = false;

    var jed = createCommonjsModule(function (module, exports) {
    /**
     * @preserve jed.js https://github.com/SlexAxton/Jed
     */
    /*
    -----------
    A gettext compatible i18n library for modern JavaScript Applications

    by Alex Sexton - AlexSexton [at] gmail - @SlexAxton

    MIT License

    A jQuery Foundation project - requires CLA to contribute -
    https://contribute.jquery.org/CLA/



    Jed offers the entire applicable GNU gettext spec'd set of
    functions, but also offers some nicer wrappers around them.
    The api for gettext was written for a language with no function
    overloading, so Jed allows a little more of that.

    Many thanks to Joshua I. Miller - unrtst@cpan.org - who wrote
    gettext.js back in 2008. I was able to vet a lot of my ideas
    against his. I also made sure Jed passed against his tests
    in order to offer easy upgrades -- jsgettext.berlios.de
    */
    (function (root, undef) {

      // Set up some underscore-style functions, if you already have
      // underscore, feel free to delete this section, and use it
      // directly, however, the amount of functions used doesn't
      // warrant having underscore as a full dependency.
      // Underscore 1.3.0 was used to port and is licensed
      // under the MIT License by Jeremy Ashkenas.
      var ArrayProto    = Array.prototype,
          ObjProto      = Object.prototype,
          slice         = ArrayProto.slice,
          hasOwnProp    = ObjProto.hasOwnProperty,
          nativeForEach = ArrayProto.forEach,
          breaker       = {};

      // We're not using the OOP style _ so we don't need the
      // extra level of indirection. This still means that you
      // sub out for real `_` though.
      var _ = {
        forEach : function( obj, iterator, context ) {
          var i, l, key;
          if ( obj === null ) {
            return;
          }

          if ( nativeForEach && obj.forEach === nativeForEach ) {
            obj.forEach( iterator, context );
          }
          else if ( obj.length === +obj.length ) {
            for ( i = 0, l = obj.length; i < l; i++ ) {
              if ( i in obj && iterator.call( context, obj[i], i, obj ) === breaker ) {
                return;
              }
            }
          }
          else {
            for ( key in obj) {
              if ( hasOwnProp.call( obj, key ) ) {
                if ( iterator.call (context, obj[key], key, obj ) === breaker ) {
                  return;
                }
              }
            }
          }
        },
        extend : function( obj ) {
          this.forEach( slice.call( arguments, 1 ), function ( source ) {
            for ( var prop in source ) {
              obj[prop] = source[prop];
            }
          });
          return obj;
        }
      };
      // END Miniature underscore impl

      // Jed is a constructor function
      var Jed = function ( options ) {
        // Some minimal defaults
        this.defaults = {
          "locale_data" : {
            "messages" : {
              "" : {
                "domain"       : "messages",
                "lang"         : "en",
                "plural_forms" : "nplurals=2; plural=(n != 1);"
              }
              // There are no default keys, though
            }
          },
          // The default domain if one is missing
          "domain" : "messages",
          // enable debug mode to log untranslated strings to the console
          "debug" : false
        };

        // Mix in the sent options with the default options
        this.options = _.extend( {}, this.defaults, options );
        this.textdomain( this.options.domain );

        if ( options.domain && ! this.options.locale_data[ this.options.domain ] ) {
          throw new Error('Text domain set to non-existent domain: `' + options.domain + '`');
        }
      };

      // The gettext spec sets this character as the default
      // delimiter for context lookups.
      // e.g.: context\u0004key
      // If your translation company uses something different,
      // just change this at any time and it will use that instead.
      Jed.context_delimiter = String.fromCharCode( 4 );

      function getPluralFormFunc ( plural_form_string ) {
        return Jed.PF.compile( plural_form_string || "nplurals=2; plural=(n != 1);");
      }

      function Chain( key, i18n ){
        this._key = key;
        this._i18n = i18n;
      }

      // Create a chainable api for adding args prettily
      _.extend( Chain.prototype, {
        onDomain : function ( domain ) {
          this._domain = domain;
          return this;
        },
        withContext : function ( context ) {
          this._context = context;
          return this;
        },
        ifPlural : function ( num, pkey ) {
          this._val = num;
          this._pkey = pkey;
          return this;
        },
        fetch : function ( sArr ) {
          if ( {}.toString.call( sArr ) != '[object Array]' ) {
            sArr = [].slice.call(arguments, 0);
          }
          return ( sArr && sArr.length ? Jed.sprintf : function(x){ return x; } )(
            this._i18n.dcnpgettext(this._domain, this._context, this._key, this._pkey, this._val),
            sArr
          );
        }
      });

      // Add functions to the Jed prototype.
      // These will be the functions on the object that's returned
      // from creating a `new Jed()`
      // These seem redundant, but they gzip pretty well.
      _.extend( Jed.prototype, {
        // The sexier api start point
        translate : function ( key ) {
          return new Chain( key, this );
        },

        textdomain : function ( domain ) {
          if ( ! domain ) {
            return this._textdomain;
          }
          this._textdomain = domain;
        },

        gettext : function ( key ) {
          return this.dcnpgettext.call( this, undef, undef, key );
        },

        dgettext : function ( domain, key ) {
         return this.dcnpgettext.call( this, domain, undef, key );
        },

        dcgettext : function ( domain , key /*, category */ ) {
          // Ignores the category anyways
          return this.dcnpgettext.call( this, domain, undef, key );
        },

        ngettext : function ( skey, pkey, val ) {
          return this.dcnpgettext.call( this, undef, undef, skey, pkey, val );
        },

        dngettext : function ( domain, skey, pkey, val ) {
          return this.dcnpgettext.call( this, domain, undef, skey, pkey, val );
        },

        dcngettext : function ( domain, skey, pkey, val/*, category */) {
          return this.dcnpgettext.call( this, domain, undef, skey, pkey, val );
        },

        pgettext : function ( context, key ) {
          return this.dcnpgettext.call( this, undef, context, key );
        },

        dpgettext : function ( domain, context, key ) {
          return this.dcnpgettext.call( this, domain, context, key );
        },

        dcpgettext : function ( domain, context, key/*, category */) {
          return this.dcnpgettext.call( this, domain, context, key );
        },

        npgettext : function ( context, skey, pkey, val ) {
          return this.dcnpgettext.call( this, undef, context, skey, pkey, val );
        },

        dnpgettext : function ( domain, context, skey, pkey, val ) {
          return this.dcnpgettext.call( this, domain, context, skey, pkey, val );
        },

        // The most fully qualified gettext function. It has every option.
        // Since it has every option, we can use it from every other method.
        // This is the bread and butter.
        // Technically there should be one more argument in this function for 'Category',
        // but since we never use it, we might as well not waste the bytes to define it.
        dcnpgettext : function ( domain, context, singular_key, plural_key, val ) {
          // Set some defaults

          plural_key = plural_key || singular_key;

          // Use the global domain default if one
          // isn't explicitly passed in
          domain = domain || this._textdomain;

          var fallback;

          // Handle special cases

          // No options found
          if ( ! this.options ) {
            // There's likely something wrong, but we'll return the correct key for english
            // We do this by instantiating a brand new Jed instance with the default set
            // for everything that could be broken.
            fallback = new Jed();
            return fallback.dcnpgettext.call( fallback, undefined, undefined, singular_key, plural_key, val );
          }

          // No translation data provided
          if ( ! this.options.locale_data ) {
            throw new Error('No locale data provided.');
          }

          if ( ! this.options.locale_data[ domain ] ) {
            throw new Error('Domain `' + domain + '` was not found.');
          }

          if ( ! this.options.locale_data[ domain ][ "" ] ) {
            throw new Error('No locale meta information provided.');
          }

          // Make sure we have a truthy key. Otherwise we might start looking
          // into the empty string key, which is the options for the locale
          // data.
          if ( ! singular_key ) {
            throw new Error('No translation key found.');
          }

          var key  = context ? context + Jed.context_delimiter + singular_key : singular_key,
              locale_data = this.options.locale_data,
              dict = locale_data[ domain ],
              defaultConf = (locale_data.messages || this.defaults.locale_data.messages)[""],
              pluralForms = dict[""].plural_forms || dict[""]["Plural-Forms"] || dict[""]["plural-forms"] || defaultConf.plural_forms || defaultConf["Plural-Forms"] || defaultConf["plural-forms"],
              val_list,
              res;

          var val_idx;
          if (val === undefined) {
            // No value passed in; assume singular key lookup.
            val_idx = 0;

          } else {
            // Value has been passed in; use plural-forms calculations.

            // Handle invalid numbers, but try casting strings for good measure
            if ( typeof val != 'number' ) {
              val = parseInt( val, 10 );

              if ( isNaN( val ) ) {
                throw new Error('The number that was passed in is not a number.');
              }
            }

            val_idx = getPluralFormFunc(pluralForms)(val);
          }

          // Throw an error if a domain isn't found
          if ( ! dict ) {
            throw new Error('No domain named `' + domain + '` could be found.');
          }

          val_list = dict[ key ];

          // If there is no match, then revert back to
          // english style singular/plural with the keys passed in.
          if ( ! val_list || val_idx > val_list.length ) {
            if (this.options.missing_key_callback) {
              this.options.missing_key_callback(key, domain);
            }
            res = [ singular_key, plural_key ];

            // collect untranslated strings
            if (this.options.debug===true) {
              console.log(res[ getPluralFormFunc(pluralForms)( val ) ]);
            }
            return res[ getPluralFormFunc()( val ) ];
          }

          res = val_list[ val_idx ];

          // This includes empty strings on purpose
          if ( ! res  ) {
            res = [ singular_key, plural_key ];
            return res[ getPluralFormFunc()( val ) ];
          }
          return res;
        }
      });


      // We add in sprintf capabilities for post translation value interolation
      // This is not internally used, so you can remove it if you have this
      // available somewhere else, or want to use a different system.

      // We _slightly_ modify the normal sprintf behavior to more gracefully handle
      // undefined values.

      /**
       sprintf() for JavaScript 0.7-beta1
       http://www.diveintojavascript.com/projects/javascript-sprintf

       Copyright (c) Alexandru Marasteanu <alexaholic [at) gmail (dot] com>
       All rights reserved.

       Redistribution and use in source and binary forms, with or without
       modification, are permitted provided that the following conditions are met:
           * Redistributions of source code must retain the above copyright
             notice, this list of conditions and the following disclaimer.
           * Redistributions in binary form must reproduce the above copyright
             notice, this list of conditions and the following disclaimer in the
             documentation and/or other materials provided with the distribution.
           * Neither the name of sprintf() for JavaScript nor the
             names of its contributors may be used to endorse or promote products
             derived from this software without specific prior written permission.

       THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
       ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
       WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
       DISCLAIMED. IN NO EVENT SHALL Alexandru Marasteanu BE LIABLE FOR ANY
       DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
       (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
       LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
       ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
       (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
       SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
      */
      var sprintf = (function() {
        function get_type(variable) {
          return Object.prototype.toString.call(variable).slice(8, -1).toLowerCase();
        }
        function str_repeat(input, multiplier) {
          for (var output = []; multiplier > 0; output[--multiplier] = input) {/* do nothing */}
          return output.join('');
        }

        var str_format = function() {
          if (!str_format.cache.hasOwnProperty(arguments[0])) {
            str_format.cache[arguments[0]] = str_format.parse(arguments[0]);
          }
          return str_format.format.call(null, str_format.cache[arguments[0]], arguments);
        };

        str_format.format = function(parse_tree, argv) {
          var cursor = 1, tree_length = parse_tree.length, node_type = '', arg, output = [], i, k, match, pad, pad_character, pad_length;
          for (i = 0; i < tree_length; i++) {
            node_type = get_type(parse_tree[i]);
            if (node_type === 'string') {
              output.push(parse_tree[i]);
            }
            else if (node_type === 'array') {
              match = parse_tree[i]; // convenience purposes only
              if (match[2]) { // keyword argument
                arg = argv[cursor];
                for (k = 0; k < match[2].length; k++) {
                  if (!arg.hasOwnProperty(match[2][k])) {
                    throw(sprintf('[sprintf] property "%s" does not exist', match[2][k]));
                  }
                  arg = arg[match[2][k]];
                }
              }
              else if (match[1]) { // positional argument (explicit)
                arg = argv[match[1]];
              }
              else { // positional argument (implicit)
                arg = argv[cursor++];
              }

              if (/[^s]/.test(match[8]) && (get_type(arg) != 'number')) {
                throw(sprintf('[sprintf] expecting number but found %s', get_type(arg)));
              }

              // Jed EDIT
              if ( typeof arg == 'undefined' || arg === null ) {
                arg = '';
              }
              // Jed EDIT

              switch (match[8]) {
                case 'b': arg = arg.toString(2); break;
                case 'c': arg = String.fromCharCode(arg); break;
                case 'd': arg = parseInt(arg, 10); break;
                case 'e': arg = match[7] ? arg.toExponential(match[7]) : arg.toExponential(); break;
                case 'f': arg = match[7] ? parseFloat(arg).toFixed(match[7]) : parseFloat(arg); break;
                case 'o': arg = arg.toString(8); break;
                case 's': arg = ((arg = String(arg)) && match[7] ? arg.substring(0, match[7]) : arg); break;
                case 'u': arg = Math.abs(arg); break;
                case 'x': arg = arg.toString(16); break;
                case 'X': arg = arg.toString(16).toUpperCase(); break;
              }
              arg = (/[def]/.test(match[8]) && match[3] && arg >= 0 ? '+'+ arg : arg);
              pad_character = match[4] ? match[4] == '0' ? '0' : match[4].charAt(1) : ' ';
              pad_length = match[6] - String(arg).length;
              pad = match[6] ? str_repeat(pad_character, pad_length) : '';
              output.push(match[5] ? arg + pad : pad + arg);
            }
          }
          return output.join('');
        };

        str_format.cache = {};

        str_format.parse = function(fmt) {
          var _fmt = fmt, match = [], parse_tree = [], arg_names = 0;
          while (_fmt) {
            if ((match = /^[^\x25]+/.exec(_fmt)) !== null) {
              parse_tree.push(match[0]);
            }
            else if ((match = /^\x25{2}/.exec(_fmt)) !== null) {
              parse_tree.push('%');
            }
            else if ((match = /^\x25(?:([1-9]\d*)\$|\(([^\)]+)\))?(\+)?(0|'[^$])?(-)?(\d+)?(?:\.(\d+))?([b-fosuxX])/.exec(_fmt)) !== null) {
              if (match[2]) {
                arg_names |= 1;
                var field_list = [], replacement_field = match[2], field_match = [];
                if ((field_match = /^([a-z_][a-z_\d]*)/i.exec(replacement_field)) !== null) {
                  field_list.push(field_match[1]);
                  while ((replacement_field = replacement_field.substring(field_match[0].length)) !== '') {
                    if ((field_match = /^\.([a-z_][a-z_\d]*)/i.exec(replacement_field)) !== null) {
                      field_list.push(field_match[1]);
                    }
                    else if ((field_match = /^\[(\d+)\]/.exec(replacement_field)) !== null) {
                      field_list.push(field_match[1]);
                    }
                    else {
                      throw('[sprintf] huh?');
                    }
                  }
                }
                else {
                  throw('[sprintf] huh?');
                }
                match[2] = field_list;
              }
              else {
                arg_names |= 2;
              }
              if (arg_names === 3) {
                throw('[sprintf] mixing positional and named placeholders is not (yet) supported');
              }
              parse_tree.push(match);
            }
            else {
              throw('[sprintf] huh?');
            }
            _fmt = _fmt.substring(match[0].length);
          }
          return parse_tree;
        };

        return str_format;
      })();

      var vsprintf = function(fmt, argv) {
        argv.unshift(fmt);
        return sprintf.apply(null, argv);
      };

      Jed.parse_plural = function ( plural_forms, n ) {
        plural_forms = plural_forms.replace(/n/g, n);
        return Jed.parse_expression(plural_forms);
      };

      Jed.sprintf = function ( fmt, args ) {
        if ( {}.toString.call( args ) == '[object Array]' ) {
          return vsprintf( fmt, [].slice.call(args) );
        }
        return sprintf.apply(this, [].slice.call(arguments) );
      };

      Jed.prototype.sprintf = function () {
        return Jed.sprintf.apply(this, arguments);
      };
      // END sprintf Implementation

      // Start the Plural forms section
      // This is a full plural form expression parser. It is used to avoid
      // running 'eval' or 'new Function' directly against the plural
      // forms.
      //
      // This can be important if you get translations done through a 3rd
      // party vendor. I encourage you to use this instead, however, I
      // also will provide a 'precompiler' that you can use at build time
      // to output valid/safe function representations of the plural form
      // expressions. This means you can build this code out for the most
      // part.
      Jed.PF = {};

      Jed.PF.parse = function ( p ) {
        var plural_str = Jed.PF.extractPluralExpr( p );
        return Jed.PF.parser.parse.call(Jed.PF.parser, plural_str);
      };

      Jed.PF.compile = function ( p ) {
        // Handle trues and falses as 0 and 1
        function imply( val ) {
          return (val === true ? 1 : val ? val : 0);
        }

        var ast = Jed.PF.parse( p );
        return function ( n ) {
          return imply( Jed.PF.interpreter( ast )( n ) );
        };
      };

      Jed.PF.interpreter = function ( ast ) {
        return function ( n ) {
          switch ( ast.type ) {
            case 'GROUP':
              return Jed.PF.interpreter( ast.expr )( n );
            case 'TERNARY':
              if ( Jed.PF.interpreter( ast.expr )( n ) ) {
                return Jed.PF.interpreter( ast.truthy )( n );
              }
              return Jed.PF.interpreter( ast.falsey )( n );
            case 'OR':
              return Jed.PF.interpreter( ast.left )( n ) || Jed.PF.interpreter( ast.right )( n );
            case 'AND':
              return Jed.PF.interpreter( ast.left )( n ) && Jed.PF.interpreter( ast.right )( n );
            case 'LT':
              return Jed.PF.interpreter( ast.left )( n ) < Jed.PF.interpreter( ast.right )( n );
            case 'GT':
              return Jed.PF.interpreter( ast.left )( n ) > Jed.PF.interpreter( ast.right )( n );
            case 'LTE':
              return Jed.PF.interpreter( ast.left )( n ) <= Jed.PF.interpreter( ast.right )( n );
            case 'GTE':
              return Jed.PF.interpreter( ast.left )( n ) >= Jed.PF.interpreter( ast.right )( n );
            case 'EQ':
              return Jed.PF.interpreter( ast.left )( n ) == Jed.PF.interpreter( ast.right )( n );
            case 'NEQ':
              return Jed.PF.interpreter( ast.left )( n ) != Jed.PF.interpreter( ast.right )( n );
            case 'MOD':
              return Jed.PF.interpreter( ast.left )( n ) % Jed.PF.interpreter( ast.right )( n );
            case 'VAR':
              return n;
            case 'NUM':
              return ast.val;
            default:
              throw new Error("Invalid Token found.");
          }
        };
      };

      Jed.PF.extractPluralExpr = function ( p ) {
        // trim first
        p = p.replace(/^\s\s*/, '').replace(/\s\s*$/, '');

        if (! /;\s*$/.test(p)) {
          p = p.concat(';');
        }

        var nplurals_re = /nplurals\=(\d+);/,
            plural_re = /plural\=(.*);/,
            nplurals_matches = p.match( nplurals_re ),
            res = {},
            plural_matches;

        // Find the nplurals number
        if ( nplurals_matches.length > 1 ) {
          res.nplurals = nplurals_matches[1];
        }
        else {
          throw new Error('nplurals not found in plural_forms string: ' + p );
        }

        // remove that data to get to the formula
        p = p.replace( nplurals_re, "" );
        plural_matches = p.match( plural_re );

        if (!( plural_matches && plural_matches.length > 1 ) ) {
          throw new Error('`plural` expression not found: ' + p);
        }
        return plural_matches[ 1 ];
      };

      /* Jison generated parser */
      Jed.PF.parser = (function(){

    var parser = {trace: function trace() { },
    yy: {},
    symbols_: {"error":2,"expressions":3,"e":4,"EOF":5,"?":6,":":7,"||":8,"&&":9,"<":10,"<=":11,">":12,">=":13,"!=":14,"==":15,"%":16,"(":17,")":18,"n":19,"NUMBER":20,"$accept":0,"$end":1},
    terminals_: {2:"error",5:"EOF",6:"?",7:":",8:"||",9:"&&",10:"<",11:"<=",12:">",13:">=",14:"!=",15:"==",16:"%",17:"(",18:")",19:"n",20:"NUMBER"},
    productions_: [0,[3,2],[4,5],[4,3],[4,3],[4,3],[4,3],[4,3],[4,3],[4,3],[4,3],[4,3],[4,3],[4,1],[4,1]],
    performAction: function anonymous(yytext,yyleng,yylineno,yy,yystate,$$,_$) {

    var $0 = $$.length - 1;
    switch (yystate) {
    case 1: return { type : 'GROUP', expr: $$[$0-1] };
    case 2:this.$ = { type: 'TERNARY', expr: $$[$0-4], truthy : $$[$0-2], falsey: $$[$0] };
    break;
    case 3:this.$ = { type: "OR", left: $$[$0-2], right: $$[$0] };
    break;
    case 4:this.$ = { type: "AND", left: $$[$0-2], right: $$[$0] };
    break;
    case 5:this.$ = { type: 'LT', left: $$[$0-2], right: $$[$0] };
    break;
    case 6:this.$ = { type: 'LTE', left: $$[$0-2], right: $$[$0] };
    break;
    case 7:this.$ = { type: 'GT', left: $$[$0-2], right: $$[$0] };
    break;
    case 8:this.$ = { type: 'GTE', left: $$[$0-2], right: $$[$0] };
    break;
    case 9:this.$ = { type: 'NEQ', left: $$[$0-2], right: $$[$0] };
    break;
    case 10:this.$ = { type: 'EQ', left: $$[$0-2], right: $$[$0] };
    break;
    case 11:this.$ = { type: 'MOD', left: $$[$0-2], right: $$[$0] };
    break;
    case 12:this.$ = { type: 'GROUP', expr: $$[$0-1] };
    break;
    case 13:this.$ = { type: 'VAR' };
    break;
    case 14:this.$ = { type: 'NUM', val: Number(yytext) };
    break;
    }
    },
    table: [{3:1,4:2,17:[1,3],19:[1,4],20:[1,5]},{1:[3]},{5:[1,6],6:[1,7],8:[1,8],9:[1,9],10:[1,10],11:[1,11],12:[1,12],13:[1,13],14:[1,14],15:[1,15],16:[1,16]},{4:17,17:[1,3],19:[1,4],20:[1,5]},{5:[2,13],6:[2,13],7:[2,13],8:[2,13],9:[2,13],10:[2,13],11:[2,13],12:[2,13],13:[2,13],14:[2,13],15:[2,13],16:[2,13],18:[2,13]},{5:[2,14],6:[2,14],7:[2,14],8:[2,14],9:[2,14],10:[2,14],11:[2,14],12:[2,14],13:[2,14],14:[2,14],15:[2,14],16:[2,14],18:[2,14]},{1:[2,1]},{4:18,17:[1,3],19:[1,4],20:[1,5]},{4:19,17:[1,3],19:[1,4],20:[1,5]},{4:20,17:[1,3],19:[1,4],20:[1,5]},{4:21,17:[1,3],19:[1,4],20:[1,5]},{4:22,17:[1,3],19:[1,4],20:[1,5]},{4:23,17:[1,3],19:[1,4],20:[1,5]},{4:24,17:[1,3],19:[1,4],20:[1,5]},{4:25,17:[1,3],19:[1,4],20:[1,5]},{4:26,17:[1,3],19:[1,4],20:[1,5]},{4:27,17:[1,3],19:[1,4],20:[1,5]},{6:[1,7],8:[1,8],9:[1,9],10:[1,10],11:[1,11],12:[1,12],13:[1,13],14:[1,14],15:[1,15],16:[1,16],18:[1,28]},{6:[1,7],7:[1,29],8:[1,8],9:[1,9],10:[1,10],11:[1,11],12:[1,12],13:[1,13],14:[1,14],15:[1,15],16:[1,16]},{5:[2,3],6:[2,3],7:[2,3],8:[2,3],9:[1,9],10:[1,10],11:[1,11],12:[1,12],13:[1,13],14:[1,14],15:[1,15],16:[1,16],18:[2,3]},{5:[2,4],6:[2,4],7:[2,4],8:[2,4],9:[2,4],10:[1,10],11:[1,11],12:[1,12],13:[1,13],14:[1,14],15:[1,15],16:[1,16],18:[2,4]},{5:[2,5],6:[2,5],7:[2,5],8:[2,5],9:[2,5],10:[2,5],11:[2,5],12:[2,5],13:[2,5],14:[2,5],15:[2,5],16:[1,16],18:[2,5]},{5:[2,6],6:[2,6],7:[2,6],8:[2,6],9:[2,6],10:[2,6],11:[2,6],12:[2,6],13:[2,6],14:[2,6],15:[2,6],16:[1,16],18:[2,6]},{5:[2,7],6:[2,7],7:[2,7],8:[2,7],9:[2,7],10:[2,7],11:[2,7],12:[2,7],13:[2,7],14:[2,7],15:[2,7],16:[1,16],18:[2,7]},{5:[2,8],6:[2,8],7:[2,8],8:[2,8],9:[2,8],10:[2,8],11:[2,8],12:[2,8],13:[2,8],14:[2,8],15:[2,8],16:[1,16],18:[2,8]},{5:[2,9],6:[2,9],7:[2,9],8:[2,9],9:[2,9],10:[2,9],11:[2,9],12:[2,9],13:[2,9],14:[2,9],15:[2,9],16:[1,16],18:[2,9]},{5:[2,10],6:[2,10],7:[2,10],8:[2,10],9:[2,10],10:[2,10],11:[2,10],12:[2,10],13:[2,10],14:[2,10],15:[2,10],16:[1,16],18:[2,10]},{5:[2,11],6:[2,11],7:[2,11],8:[2,11],9:[2,11],10:[2,11],11:[2,11],12:[2,11],13:[2,11],14:[2,11],15:[2,11],16:[2,11],18:[2,11]},{5:[2,12],6:[2,12],7:[2,12],8:[2,12],9:[2,12],10:[2,12],11:[2,12],12:[2,12],13:[2,12],14:[2,12],15:[2,12],16:[2,12],18:[2,12]},{4:30,17:[1,3],19:[1,4],20:[1,5]},{5:[2,2],6:[1,7],7:[2,2],8:[1,8],9:[1,9],10:[1,10],11:[1,11],12:[1,12],13:[1,13],14:[1,14],15:[1,15],16:[1,16],18:[2,2]}],
    defaultActions: {6:[2,1]},
    parseError: function parseError(str, hash) {
        throw new Error(str);
    },
    parse: function parse(input) {
        var self = this,
            stack = [0],
            vstack = [null], // semantic value stack
            lstack = [], // location stack
            table = this.table,
            yytext = '',
            yylineno = 0,
            yyleng = 0,
            recovering = 0,
            TERROR = 2,
            EOF = 1;

        //this.reductionCount = this.shiftCount = 0;

        this.lexer.setInput(input);
        this.lexer.yy = this.yy;
        this.yy.lexer = this.lexer;
        if (typeof this.lexer.yylloc == 'undefined')
            this.lexer.yylloc = {};
        var yyloc = this.lexer.yylloc;
        lstack.push(yyloc);

        if (typeof this.yy.parseError === 'function')
            this.parseError = this.yy.parseError;

        function popStack (n) {
            stack.length = stack.length - 2*n;
            vstack.length = vstack.length - n;
            lstack.length = lstack.length - n;
        }

        function lex() {
            var token;
            token = self.lexer.lex() || 1; // $end = 1
            // if token isn't its numeric value, convert
            if (typeof token !== 'number') {
                token = self.symbols_[token] || token;
            }
            return token;
        }

        var symbol, preErrorSymbol, state, action, r, yyval={},p,len,newState, expected;
        while (true) {
            // retreive state number from top of stack
            state = stack[stack.length-1];

            // use default actions if available
            if (this.defaultActions[state]) {
                action = this.defaultActions[state];
            } else {
                if (symbol == null)
                    symbol = lex();
                // read action for current state and first input
                action = table[state] && table[state][symbol];
            }

            // handle parse error
            
            if (typeof action === 'undefined' || !action.length || !action[0]) {

                if (!recovering) {
                    // Report error
                    expected = [];
                    for (p in table[state]) if (this.terminals_[p] && p > 2) {
                        expected.push("'"+this.terminals_[p]+"'");
                    }
                    var errStr = '';
                    if (this.lexer.showPosition) {
                        errStr = 'Parse error on line '+(yylineno+1)+":\n"+this.lexer.showPosition()+"\nExpecting "+expected.join(', ') + ", got '" + this.terminals_[symbol]+ "'";
                    } else {
                        errStr = 'Parse error on line '+(yylineno+1)+": Unexpected " +
                                      (symbol == 1 /*EOF*/ ? "end of input" :
                                                  ("'"+(this.terminals_[symbol] || symbol)+"'"));
                    }
                    this.parseError(errStr,
                        {text: this.lexer.match, token: this.terminals_[symbol] || symbol, line: this.lexer.yylineno, loc: yyloc, expected: expected});
                }

                // just recovered from another error
                if (recovering == 3) {
                    if (symbol == EOF) {
                        throw new Error(errStr || 'Parsing halted.');
                    }

                    // discard current lookahead and grab another
                    yyleng = this.lexer.yyleng;
                    yytext = this.lexer.yytext;
                    yylineno = this.lexer.yylineno;
                    yyloc = this.lexer.yylloc;
                    symbol = lex();
                }

                // try to recover from error
                while (1) {
                    // check for error recovery rule in this state
                    if ((TERROR.toString()) in table[state]) {
                        break;
                    }
                    if (state == 0) {
                        throw new Error(errStr || 'Parsing halted.');
                    }
                    popStack(1);
                    state = stack[stack.length-1];
                }

                preErrorSymbol = symbol; // save the lookahead token
                symbol = TERROR;         // insert generic error symbol as new lookahead
                state = stack[stack.length-1];
                action = table[state] && table[state][TERROR];
                recovering = 3; // allow 3 real symbols to be shifted before reporting a new error
            }

            // this shouldn't happen, unless resolve defaults are off
            if (action[0] instanceof Array && action.length > 1) {
                throw new Error('Parse Error: multiple actions possible at state: '+state+', token: '+symbol);
            }

            switch (action[0]) {

                case 1: // shift
                    //this.shiftCount++;

                    stack.push(symbol);
                    vstack.push(this.lexer.yytext);
                    lstack.push(this.lexer.yylloc);
                    stack.push(action[1]); // push state
                    symbol = null;
                    if (!preErrorSymbol) { // normal execution/no error
                        yyleng = this.lexer.yyleng;
                        yytext = this.lexer.yytext;
                        yylineno = this.lexer.yylineno;
                        yyloc = this.lexer.yylloc;
                        if (recovering > 0)
                            recovering--;
                    } else { // error just occurred, resume old lookahead f/ before error
                        symbol = preErrorSymbol;
                        preErrorSymbol = null;
                    }
                    break;

                case 2: // reduce
                    //this.reductionCount++;

                    len = this.productions_[action[1]][1];

                    // perform semantic action
                    yyval.$ = vstack[vstack.length-len]; // default to $$ = $1
                    // default location, uses first token for firsts, last for lasts
                    yyval._$ = {
                        first_line: lstack[lstack.length-(len||1)].first_line,
                        last_line: lstack[lstack.length-1].last_line,
                        first_column: lstack[lstack.length-(len||1)].first_column,
                        last_column: lstack[lstack.length-1].last_column
                    };
                    r = this.performAction.call(yyval, yytext, yyleng, yylineno, this.yy, action[1], vstack, lstack);

                    if (typeof r !== 'undefined') {
                        return r;
                    }

                    // pop off stack
                    if (len) {
                        stack = stack.slice(0,-1*len*2);
                        vstack = vstack.slice(0, -1*len);
                        lstack = lstack.slice(0, -1*len);
                    }

                    stack.push(this.productions_[action[1]][0]);    // push nonterminal (reduce)
                    vstack.push(yyval.$);
                    lstack.push(yyval._$);
                    // goto new state = table[STATE][NONTERMINAL]
                    newState = table[stack[stack.length-2]][stack[stack.length-1]];
                    stack.push(newState);
                    break;

                case 3: // accept
                    return true;
            }

        }

        return true;
    }};/* Jison generated lexer */
    var lexer = (function(){

    var lexer = ({EOF:1,
    parseError:function parseError(str, hash) {
            if (this.yy.parseError) {
                this.yy.parseError(str, hash);
            } else {
                throw new Error(str);
            }
        },
    setInput:function (input) {
            this._input = input;
            this._more = this._less = this.done = false;
            this.yylineno = this.yyleng = 0;
            this.yytext = this.matched = this.match = '';
            this.conditionStack = ['INITIAL'];
            this.yylloc = {first_line:1,first_column:0,last_line:1,last_column:0};
            return this;
        },
    input:function () {
            var ch = this._input[0];
            this.yytext+=ch;
            this.yyleng++;
            this.match+=ch;
            this.matched+=ch;
            var lines = ch.match(/\n/);
            if (lines) this.yylineno++;
            this._input = this._input.slice(1);
            return ch;
        },
    unput:function (ch) {
            this._input = ch + this._input;
            return this;
        },
    more:function () {
            this._more = true;
            return this;
        },
    pastInput:function () {
            var past = this.matched.substr(0, this.matched.length - this.match.length);
            return (past.length > 20 ? '...':'') + past.substr(-20).replace(/\n/g, "");
        },
    upcomingInput:function () {
            var next = this.match;
            if (next.length < 20) {
                next += this._input.substr(0, 20-next.length);
            }
            return (next.substr(0,20)+(next.length > 20 ? '...':'')).replace(/\n/g, "");
        },
    showPosition:function () {
            var pre = this.pastInput();
            var c = new Array(pre.length + 1).join("-");
            return pre + this.upcomingInput() + "\n" + c+"^";
        },
    next:function () {
            if (this.done) {
                return this.EOF;
            }
            if (!this._input) this.done = true;

            var token,
                match,
                lines;
            if (!this._more) {
                this.yytext = '';
                this.match = '';
            }
            var rules = this._currentRules();
            for (var i=0;i < rules.length; i++) {
                match = this._input.match(this.rules[rules[i]]);
                if (match) {
                    lines = match[0].match(/\n.*/g);
                    if (lines) this.yylineno += lines.length;
                    this.yylloc = {first_line: this.yylloc.last_line,
                                   last_line: this.yylineno+1,
                                   first_column: this.yylloc.last_column,
                                   last_column: lines ? lines[lines.length-1].length-1 : this.yylloc.last_column + match[0].length};
                    this.yytext += match[0];
                    this.match += match[0];
                    this.matches = match;
                    this.yyleng = this.yytext.length;
                    this._more = false;
                    this._input = this._input.slice(match[0].length);
                    this.matched += match[0];
                    token = this.performAction.call(this, this.yy, this, rules[i],this.conditionStack[this.conditionStack.length-1]);
                    if (token) return token;
                    else return;
                }
            }
            if (this._input === "") {
                return this.EOF;
            } else {
                this.parseError('Lexical error on line '+(this.yylineno+1)+'. Unrecognized text.\n'+this.showPosition(),
                        {text: "", token: null, line: this.yylineno});
            }
        },
    lex:function lex() {
            var r = this.next();
            if (typeof r !== 'undefined') {
                return r;
            } else {
                return this.lex();
            }
        },
    begin:function begin(condition) {
            this.conditionStack.push(condition);
        },
    popState:function popState() {
            return this.conditionStack.pop();
        },
    _currentRules:function _currentRules() {
            return this.conditions[this.conditionStack[this.conditionStack.length-1]].rules;
        },
    topState:function () {
            return this.conditionStack[this.conditionStack.length-2];
        },
    pushState:function begin(condition) {
            this.begin(condition);
        }});
    lexer.performAction = function anonymous(yy,yy_,$avoiding_name_collisions,YY_START) {
    switch($avoiding_name_collisions) {
    case 0:/* skip whitespace */
    break;
    case 1:return 20
    case 2:return 19
    case 3:return 8
    case 4:return 9
    case 5:return 6
    case 6:return 7
    case 7:return 11
    case 8:return 13
    case 9:return 10
    case 10:return 12
    case 11:return 14
    case 12:return 15
    case 13:return 16
    case 14:return 17
    case 15:return 18
    case 16:return 5
    case 17:return 'INVALID'
    }
    };
    lexer.rules = [/^\s+/,/^[0-9]+(\.[0-9]+)?\b/,/^n\b/,/^\|\|/,/^&&/,/^\?/,/^:/,/^<=/,/^>=/,/^</,/^>/,/^!=/,/^==/,/^%/,/^\(/,/^\)/,/^$/,/^./];
    lexer.conditions = {"INITIAL":{"rules":[0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17],"inclusive":true}};return lexer;})();
    parser.lexer = lexer;
    return parser;
    })();
    // End parser

      // Handle node, amd, and global systems
      {
        if ( module.exports) {
          exports = module.exports = Jed;
        }
        exports.Jed = Jed;
      }

    })();
    });

    /*
     This file is part of TALER
     (C) 2016 GNUnet e.V.

     TALER is free software; you can redistribute it and/or modify it under the
     terms of the GNU General Public License as published by the Free Software
     Foundation; either version 3, or (at your option) any later version.

     TALER is distributed in the hope that it will be useful, but WITHOUT ANY
     WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
     A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

     You should have received a copy of the GNU General Public License along with
     TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
    */
    const logger$1 = new Logger("browserWorkerEntry.ts");
    const worker = self;
    function handleRequest(operation, id, args) {
        return __awaiter(this, void 0, void 0, function* () {
            const impl = new CryptoImplementation();
            if (!(operation in impl)) {
                console.error(`crypto operation '${operation}' not found`);
                return;
            }
            try {
                const result = impl[operation](...args);
                worker.postMessage({ result, id });
            }
            catch (e) {
                logger$1.error("error during operation", e);
                return;
            }
        });
    }
    worker.onmessage = (msg) => {
        const args = msg.data.args;
        if (!Array.isArray(args)) {
            console.error("args must be array");
            return;
        }
        const id = msg.data.id;
        if (typeof id !== "number") {
            console.error("RPC id must be number");
            return;
        }
        const operation = msg.data.operation;
        if (typeof operation !== "string") {
            console.error("RPC operation must be string");
            return;
        }
        handleRequest(operation, id, args).catch((e) => {
            console.error("error in browsere worker", e);
        });
    };

}());
